
/* $Id: merge_coords.c,v 3.5 2000/06/12 21:38:33 kohl Exp $ */


/*
 *                      CUMULVS version 1.0:
 *      Computational Steering and Interactive Visualization
 *                   in Distributed Applications
 *           Oak Ridge National Laboratory, Oak Ridge TN.
 *      Authors:  James Arthur Kohl and Philip M. Papadopoulos
 *                   (C) 1996 All Rights Reserved
 *
 *                              NOTICE
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted
 * provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear
 * in supporting documentation.
 *
 * Neither the Institution, Oak Ridge National Laboratory, nor the
 * Authors make any representations about the suitability of this
 * software for any purpose.  This software is provided ``as is''
 * without express or implied warranty.
 *
 * CUMULVS was funded by the U.S. Department of Energy.
 */


/*
			Copyright (c) 1989 by
			Advanced Visual Systems Inc.
			All Rights Reserved
	
	This software comprises unpublished confidential information of
	Advanced Visual Systems Inc. and may not be used, copied or made
	available to anyone, except in accordance with the license
	under which it is furnished.
	
	This file is under sccs control at AVS in:
	/nfs/caffeine/root/src1/sccs/avs/examples/s.threshold.c
	
*/

#include <stdio.h>
#include <avs/avs.h>
#include <avs/field.h>

/*********************************************************************/


#define INPUT_FIELD_DEFN	"field 3D scalar float rectilinear"

#define COORD_FIELD_DEFN	"field 3D scalar float 3-coord irregular"


static int merge_coords();

static int mc_compute();


static int ((*mod_list[])()) = { merge_coords };
#define NMODS (sizeof(mod_list) / sizeof(char *))

void
AVSinit_modules()
{
	AVSinit_from_module_list (mod_list, NMODS);
}


/*  The routine "merge_coords" is the description routine.  */

static int merge_coords()
{
	/* Set the module name and type */
	AVSset_module_name("Merge Coords", MODULE_FILTER);

	/* Create an input port for the required field input */
	AVScreate_input_port("Input Field", INPUT_FIELD_DEFN, REQUIRED );

	/* Create an input port for the coord field */
	AVScreate_input_port("Coord Input", COORD_FIELD_DEFN, REQUIRED );

	/* Create an output port for the merged coord field */
	AVScreate_output_port("Coord Field", COORD_FIELD_DEFN );

	/* Tell avs what subroutine to call to do the compute */
	AVSset_compute_proc( mc_compute );

	return( 1 );
}


/*
 * mc_compute is the compute routine.  It is called whenever AVS
 * wants to compute new merge_coords results.  The arguments are: the
 * value of the input field, the new output field (doubly indirected),
 * the minimum parameter value and the maximum parameter value.  Note
 * the order is always inputs, outputs, parameters.  The min comes
 * before the max because in the description routine above, the min
 * is declared before the max.
 */

static int mc_compute( input, coords, output )
AVSfield_float *input, *coords, **output;
{
	register int i, j, k;

	int dims[3];

	/*
	 * We use a triply nested loop to traverse the field.  The macros
	 * MAXX, MAXY, and MAXZ determine the maximum extent of the field
	 * in each of the three dimensions.  We know this will be a
	 * 3-dimensional field because of the declaration in the description
	 * routine, so we don't need to check.  When we want to reference
	 * an element of the field we use the I3D macro which picks an
	 * element of a 3D field.  Note that the first index (i) varies the
	 * fastest in memory, so we make that the innermost loop.
	 */

	/* Allocate the Output Field Frames */

	dims[0] = MAXX(input);
	dims[1] = MAXY(input);
	dims[2] = MAXZ(input);

	if ( *output == NULL )
	{
		printf( "Allocating Output Coord Field\n" );

		*output = (AVSfield_float *) AVSdata_alloc( COORD_FIELD_DEFN,
			dims );
	}

	/* Fill In Output Fields */

	for ( k=0 ; k < MAXZ(input) ; k++ )
	{
		for ( j=0 ; j < MAXY(input) ; j++ )
		{
			for ( i=0 ; i < MAXX(input) ; i++ )
			{
				/* Set Coords */

				((*output)->points)[ ((*output)->dimensions[0] *
					((*output)->dimensions[1] * (k) + (j))) + (i) ]
						= COORD_X_3D( coords, i, j, k );

				((*output)->points)[ ((*output)->dimensions[0] *
					(((*output)->dimensions[1] *
					((*output)->dimensions[2] + (k))) + (j))) + (i) ]
						= COORD_Y_3D( coords, i, j, k );

				((*output)->points)[ ((*output)->dimensions[0] * \
					(((*output)->dimensions[1] *
					((*output)->dimensions[2]*2 + (k))) + (j))) + (i) ]
						= COORD_Z_3D( coords, i, j, k );

				/* Get Data Value */

				I3D( *output, i, j, k ) = I3D( input, i, j, k );
			}
		}
	}

	/* When we're done, we return 1 to indicate success */
	return(1);
}

