
/* $Id: const.h,v 3.1 1998/01/27 16:03:16 kohl Exp $ */


/*
 *                      CUMULVS version 1.0:
 *      Computational Steering and Interactive Visualization
 *                   in Distributed Applications
 *           Oak Ridge National Laboratory, Oak Ridge TN.
 *      Authors:  James Arthur Kohl and Philip M. Papadopoulos
 *                   (C) 1996 All Rights Reserved
 *
 *                              NOTICE
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted
 * provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear
 * in supporting documentation.
 *
 * Neither the Institution, Oak Ridge National Laboratory, nor the
 * Authors make any representations about the suitability of this
 * software for any purpose.  This software is provided ``as is''
 * without express or implied warranty.
 *
 * CUMULVS was funded by the U.S. Department of Energy.
 */


/* FlyMe Version */

#define FLYME_VERSION "1.0"


/* Define Default Location of Slicer Main Directory */

#define FLYME_DEFAULT_DIR "examples/flyme"

/* Define Default Location of STV Main Directory */

#define STV_DEFAULT_DIR "CUMULVS"

/* Winning Score */

#define WINNING_SCORE    50000


/* Heli / Bullet Update Constants */

#define NOT_UPDATED    0
#define NO_CHANGE      1
#define UPDATED        2


/* Play Mode Constants */

#define MONITORING     0
#define PLAYING        1


/* Arrow Constants */

#define ARROW_SIZE    50


/* Rotor Constants */

#define ROTOR_HT		( 3 * HELI_RADIUS / 4 )
#define ROTOR_WT		( ROTOR_HT / 5 )

#define ROTOR_HT2		( ROTOR_HT / 2 )
#define ROTOR_WT2		( ROTOR_WT / 2 )


/* Helicopter Body Constants */

#define BODY_HT			( HELI_RADIUS / 2 )
#define BODY_WT			( BODY_HT / 2 )

#define BODY_HT2		( BODY_HT / 2 )
#define BODY_WT2		( BODY_WT / 2 )


/* Helicopter Tail Constants */

#define TAIL_HT			( BODY_HT / 2 )
#define TAIL_WT			( 2 * TAIL_HT / 3 )

#define TAIL_HT2		( TAIL_HT / 2 )
#define TAIL_WT2		( TAIL_WT / 2 )


/* Bullet Constants */

#define BULLET_SIZE      4


/* Current Heli Check Macro */

#define IS_CURRENT( _heli ) \
( CURRENT_HELI != NULL && (_heli)->hid == CURRENT_HELI->hid )


/* Floating Point Equivalence Macro */

#define EQUIV_DELTA			0.000001

#define EQUIV( _d1, _d2 ) \
( ( (_d1) != 0.0 ) ? \
	( ( fabs( (_d1) - (_d2) ) / fabs( _d1 ) ) < EQUIV_DELTA ) \
	: ( fabs( (_d1) - (_d2) ) < EQUIV_DELTA ) )


/* Screen to Region Coordinate Macros */

#define SCREEN_TO_REGION( _coord, _region_min, _coord_offset, _scale ) \
( (int) ( (double) ( (_coord) - (_coord_offset) ) / (_scale) ) \
	+ (_region_min) )

#define REGION_TO_SCREEN( _coord, _region_min, _coord_offset, _scale ) \
( (int) ( (double) ( (_coord) - (_region_min) ) * (_scale) ) \
	+ (_coord_offset) )


/* TCL Globals Macros */

#define GET_TCL_GLOBAL( _itp, _name ) \
	Tcl_GetVar( _itp, _name, TCL_GLOBAL_ONLY )

#define SET_TCL_GLOBAL( _itp, _name, _str ) \
	Tcl_SetVar( _itp, _name, _str, TCL_GLOBAL_ONLY )


/* TK Drawing Macros */


#define DRAW_TEXT( _itp, _c, _x, _y, _str ) \
(	sprintf( TMP_CMD, \
		"%s create text %d %d -text {%s} -anchor nw -font %s", \
		_c, \
		( (_x) + COL_WIDTH ), ( (_y) + (2 * FRAME_BORDER) ), \
		_str, DATA_FONT ), \
	Tcl_Eval( _itp, TMP_CMD ), \
	atoi( (_itp)->result ) \
)


#define DRAW_LINE( _itp, _c, _x1, _y1, _x2, _y2 ) \
(	sprintf( TMP_CMD, "%s create line %d %d %d %d", \
		_c, _x1, _y1, _x2, _y2 ), \
	Tcl_Eval( _itp, TMP_CMD ), \
	atoi( (_itp)->result ) \
)


#define DRAW_RECT( _itp, _c, _x1, _y1, _x2, _y2, _fill, _color, _wt ) \
(	sprintf( TMP_COORD1, "-fill \"%s\"", _fill ), \
	sprintf( TMP_COORD2, "-outline \"%s\"", _color ), \
	sprintf( TMP_COORD3, "-width %d", _wt ), \
	sprintf( TMP_CMD, "%s create rectangle %d %d %d %d %s %s %s", \
		_c, _x1, _y1, _x2, _y2, TMP_COORD1, TMP_COORD2, TMP_COORD3 ), \
	Tcl_Eval( _itp, TMP_CMD ), \
	atoi( (_itp)->result ) \
)


#define RECT_COORDS( _itp, _c, _id, _x1, _y1, _x2, _y2 ) \
(	sprintf( TMP_CMD, "%s coords %d %d %d %d %d", \
		_c, _id, _x1, _y1, _x2, _y2 ), \
	Tcl_Eval( _itp, TMP_CMD ) \
)


#define DRAW_CIRCLE( _itp, _c, _x, _y, _r, _fill, _clr, _wt, _scale ) \
(	sprintf( TMP_COORD1, "-fill \"%s\"", _fill ), \
	sprintf( TMP_COORD2, "-outline \"%s\"", _clr ), \
	sprintf( TMP_COORD3, "-width %d", _wt ), \
	sprintf( TMP_CMD, "%s create oval %d %d %d %d %s %s %s", \
		_c, \
		(_x) - ( (int) ( (_r) * (_scale) ) ), \
		(_y) - ( (int) ( (_r) * (_scale) ) ), \
		(_x) + ( (int) ( (_r) * (_scale) ) ), \
		(_y) + ( (int) ( (_r) * (_scale) ) ), \
		TMP_COORD1, TMP_COORD2, TMP_COORD3 ), \
	Tcl_Eval( _itp, TMP_CMD ), \
	atoi( (_itp)->result ) \
)


#define CIRCLE_COORDS( _itp, _c, _id, _x, _y, _r, _fl, _cr, _wt, _sc ) \
(	sprintf( TMP_CMD, "%s coords %d %d %d %d %d", \
		_c, _id, \
		(_x) - ( (int) ( (_r) * (_sc) ) ), \
		(_y) - ( (int) ( (_r) * (_sc) ) ), \
		(_x) + ( (int) ( (_r) * (_sc) ) ), \
		(_y) + ( (int) ( (_r) * (_sc) ) ) ), \
	Tcl_Eval( _itp, TMP_CMD ), \
	UPDATE_POLY( _itp, _c, _id, _fl, _cr, _wt ) \
)


#define CREATE_POLY( _itp, _c, _x, _y, _smooth, _fill, _color, _wt ) \
(	sprintf( TMP_COORD1, "-smooth %s", _smooth ), \
	sprintf( TMP_COORD2, "-fill \"%s\"", _fill ), \
	sprintf( TMP_COORD3, "-outline \"%s\"", _color ), \
	sprintf( TMP_COORD4, "-width %d", _wt ), \
	sprintf( TMP_CMD, \
		"%s create polygon %d %d %d %d %d %d %s %s %s %s", \
		_c, _x, _y, _x, _y, _x, _y, \
		TMP_COORD1, TMP_COORD2, TMP_COORD3, TMP_COORD4 ), \
	Tcl_Eval( _itp, TMP_CMD ), \
	atoi( (_itp)->result ) \
)


#define UPDATE_POLY( _itp, _c, _id, _fill, _color, _wt ) \
(	sprintf( TMP_COORD1, "-fill \"%s\"", _fill ), \
	sprintf( TMP_COORD2, "-outline \"%s\"", _color ), \
	sprintf( TMP_COORD3, "-width %d", _wt ), \
	sprintf( TMP_CMD, \
		"%s itemconfigure %d %s %s %s", \
		_c, _id, TMP_COORD1, TMP_COORD2, TMP_COORD3 ), \
	Tcl_Eval( _itp, TMP_CMD ), \
	atoi( (_itp)->result ) \
)


/* Coordinate / Angle / Scale Macros */

#define X_OF_ANGLE( _x, _y, _angle, _scale ) \
	( (int) ( (_scale) * ( ( (double) (_y) * sin( _angle ) ) \
		- ( (double) (_x) * cos( _angle ) ) ) ) )

#define Y_OF_ANGLE( _x, _y, _angle, _scale ) \
	( (int) ( (_scale) * ( ( (double) (_y) * cos( _angle ) ) \
		+ ( (double) (_x) * sin( _angle ) ) ) ) )

#define ANGLE_COORDS( _x, _y, _x0, _y0, _angle, _scale ) \
	(_x0) + X_OF_ANGLE( (_x) - (_x0), (_y) - (_y0), _angle, _scale ), \
	(_y0) + Y_OF_ANGLE( (_x) - (_x0), (_y) - (_y0), _angle, _scale )


/* Arrow Drawing Macro */

#define DRAW_ARROW( _itp, _c, _id, _x, _y, _angle, _scale ) \
{ \
	sprintf( TMP_CMD, \
		"%s coords %d %d %d %d %d %d %d %d %d %d %d %d %d", \
		_c, _id,  \
		ANGLE_COORDS( (_x), (_y) + ARROW_SIZE, \
			_x, _y, _angle, _scale ),\
		ANGLE_COORDS( (_x), (_y) - ARROW_SIZE, \
			_x, _y, _angle, _scale ), \
		ANGLE_COORDS( (_x) - (ARROW_SIZE / 8), \
				(_y) - ( 7 * ARROW_SIZE / 8 ), \
			_x, _y, _angle, _scale ), \
		ANGLE_COORDS( (_x), (_y) - ARROW_SIZE, \
			_x, _y, _angle, _scale ), \
		ANGLE_COORDS( (_x) + (ARROW_SIZE / 8), \
				(_y) - ( 7 * ARROW_SIZE / 8 ), \
			_x, _y, _angle, _scale ), \
		ANGLE_COORDS( (_x), (_y) - ARROW_SIZE, \
			_x, _y, _angle, _scale ) \
	); \
\
	Tcl_Eval( _itp, TMP_CMD ); \
}


/* Helicopter Rotor Drawing Macros */

#define ROTOR( _str, _x, _y, _angle, _scale ) \
	sprintf( _str, "%d %d %d %d %d %d %d %d", \
		ANGLE_COORDS( _x, _y, _x, _y, _angle, _scale ),\
		ANGLE_COORDS( (_x) - ROTOR_WT2, (_y) - ROTOR_HT2, \
			_x, _y, _angle, _scale ), \
		ANGLE_COORDS( (_x), (_y) - ROTOR_HT, _x, _y, _angle, _scale ), \
		ANGLE_COORDS( (_x) + ROTOR_WT2, (_y) - ROTOR_HT2, \
			_x, _y, _angle, _scale ) )

#define ROTOR_ANGLE( _posn, _num ) \
( \
	(double) ( ( (_posn) + ( (_num) * ROTOR_CLICKS / 4 ) ) \
			% ROTOR_CLICKS ) \
		* 2.0 * PI / (double) ROTOR_CLICKS \
)

#define DRAW_ROTORS( _itp, _c, _id, _x, _y, _posn, _clr, _wt, _scale ) \
{ \
	ROTOR( TMP_COORD1, _x, _y, ROTOR_ANGLE( _posn, 1 ), _scale ); \
	ROTOR( TMP_COORD2, _x, _y, ROTOR_ANGLE( _posn, 2 ), _scale ); \
	ROTOR( TMP_COORD3, _x, _y, ROTOR_ANGLE( _posn, 3 ), _scale ); \
	ROTOR( TMP_COORD4, _x, _y, ROTOR_ANGLE( _posn, 4 ), _scale ); \
\
	sprintf( TMP_CMD, \
		"%s coords %d %s %s %s %s", \
		_c, _id, TMP_COORD1, TMP_COORD2, TMP_COORD3, TMP_COORD4 ); \
\
	Tcl_Eval( _itp, TMP_CMD ); \
\
	UPDATE_POLY( _itp, _c, _id, _clr, _clr, _wt ); \
}


/* Helicopter Body / Tail Drawing Macros */

#define DRAW_BODY( _itp, _c, _id, _x, _y, _angle, _clr, _wt, _scale ) \
{ \
	sprintf( TMP_CMD, \
		"%s coords %d %d %d %d %d %d %d %d %d", \
		_c, _id,  \
		ANGLE_COORDS( (_x), (_y) + BODY_HT2, _x, _y, _angle, _scale ),\
		ANGLE_COORDS( (_x) - BODY_WT2, (_y), _x, _y, _angle, _scale ), \
		ANGLE_COORDS( (_x), (_y) - BODY_HT2, _x, _y, _angle, _scale ), \
		ANGLE_COORDS( (_x) + BODY_WT2, (_y), _x, _y, _angle, _scale ) \
	); \
\
	Tcl_Eval( _itp, TMP_CMD ); \
\
	UPDATE_POLY( _itp, _c, _id, _clr, _clr, _wt ); \
}

#define DRAW_TAIL( _itp, _c, _id, _x, _y, _angle, _clr, _wt, _scale ) \
{ \
	sprintf( TMP_COORD1, "%d %d", \
		ANGLE_COORDS( (_x), (_y) + BODY_HT2, \
		_x, _y, _angle, _scale ) ); \
\
	sprintf( TMP_COORD2, "%d %d", \
		ANGLE_COORDS( (_x), (_y) + BODY_HT2 + TAIL_HT, \
		_x, _y, _angle, _scale ) ); \
\
	sprintf( TMP_COORD3, "%d %d", \
		ANGLE_COORDS( (_x) - TAIL_WT2, \
		(_y) + BODY_HT2 + TAIL_HT + TAIL_WT2, \
		_x, _y, _angle, _scale ) ); \
\
	sprintf( TMP_COORD4, "%d %d", \
		ANGLE_COORDS( (_x) + TAIL_WT2, \
		(_y) + BODY_HT2 + TAIL_HT + TAIL_WT2, \
		_x, _y, _angle, _scale ) ); \
\
	sprintf( TMP_CMD, \
		"%s coords %d %s %s %s %s %s %s", \
		_c, _id, TMP_COORD1, TMP_COORD2, TMP_COORD3, TMP_COORD2, \
		TMP_COORD4, TMP_COORD2 ); \
\
	Tcl_Eval( _itp, TMP_CMD ); \
\
	UPDATE_POLY( _itp, _c, _id, _clr, _clr, _wt ); \
}


/* Bullet Drawing Macro */

#define DRAW_BULLET( _itp, _c, _id, _x, _y, _angle, _scale ) \
{ \
	sprintf( TMP_CMD, \
		"%s coords %d %d %d %d %d %d %d %d %d %d %d %d %d", \
		_c, _id,  \
		ANGLE_COORDS( (_x), (_y) + BULLET_SIZE, \
			_x, _y, _angle, _scale ),\
		ANGLE_COORDS( (_x) - BULLET_SIZE, (_y), \
			_x, _y, _angle, _scale ), \
		ANGLE_COORDS( (_x), (_y) - BULLET_SIZE, \
			_x, _y, _angle, _scale ), \
		ANGLE_COORDS( (_x) + BULLET_SIZE, (_y), \
			_x, _y, _angle, _scale ), \
		ANGLE_COORDS( (_x), (_y) + BULLET_SIZE, \
			_x, _y, _angle, _scale ),\
		ANGLE_COORDS( (_x), (_y) + (3 * BULLET_SIZE), \
			_x, _y, _angle, _scale )\
	); \
\
	Tcl_Eval( _itp, TMP_CMD ); \
}


/* Explosion Drawing Macros */

#define DRAW_EXPLOSION( _itp, _c, _id, _x, _y, _angle, _rad, _scale ) \
{ \
	sprintf( TMP_CMD, \
	"%s coords %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d", \
		_c, _id,  \
		SPIKE( _x, _y, _angle, _angle, RAND_RAD( _rad ), _scale ), \
		SPIKE( _x, _y, (_angle)+(PI/4.0), _angle, \
			RAND_RAD( _rad ), _scale ), \
		SPIKE( _x, _y, (_angle)+(PI/2.0), _angle, \
			RAND_RAD( _rad ), _scale ), \
		SPIKE( _x, _y, (_angle)+(3.0*PI/4.0), _angle, \
			RAND_RAD( _rad ), _scale ), \
		SPIKE( _x, _y, (_angle)+(PI), _angle, \
			RAND_RAD( _rad ), _scale ), \
		SPIKE( _x, _y, (_angle)+(5.0*PI/4.0), _angle, \
			RAND_RAD( _rad ), _scale ), \
		SPIKE( _x, _y, (_angle)+(3.0*PI/2.0), _angle, \
			RAND_RAD( _rad ), _scale ), \
		SPIKE( _x, _y, (_angle)+(7.0*PI/4.0), _angle, \
			RAND_RAD( _rad ), _scale ) \
	); \
\
	Tcl_Eval( _itp, TMP_CMD ); \
}

#define RAND_RAD( _rad ) \
	( ( 3 * (_rad) / 4 ) \
		+ ( (int) random() \
			% ( ( (_rad) / 2 ) ? ( (_rad) / 2 ) : 1 ) ) )

#define RADX( _rad, _angle )    ( (_rad) * cos( _angle ) )

#define RADY( _rad, _angle )    ( (_rad) * sin( _angle ) )

#define SPIKE( _x, _y, _a, _angle, _rad, _scale ) \
	(_x) + X_OF_ANGLE( RADX( _rad, _a ), RADY( _rad, _a ), \
		_angle, _scale ), \
	(_y) + Y_OF_ANGLE( RADX( _rad, _a ), RADY( _rad, _a ), \
		_angle, _scale )


/* Un-Drawing Macro */

#define DELETE_ITEM( _itp, _c, _id ) \
(	sprintf( TMP_CMD, "%s delete %d", _c, _id ), \
	Tcl_Eval( _itp, TMP_CMD ) \
)

