
/* $Id: flyme.c,v 3.9 1999/06/25 18:45:46 kohl Exp $ */


/*
 *                      CUMULVS version 1.0:
 *      Computational Steering and Interactive Visualization
 *                   in Distributed Applications
 *           Oak Ridge National Laboratory, Oak Ridge TN.
 *      Authors:  James Arthur Kohl and Philip M. Papadopoulos
 *                   (C) 1996 All Rights Reserved
 *
 *                              NOTICE
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted
 * provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear
 * in supporting documentation.
 *
 * Neither the Institution, Oak Ridge National Laboratory, nor the
 * Authors make any representations about the suitability of this
 * software for any purpose.  This software is provided ``as is''
 * without express or implied warranty.
 *
 * CUMULVS was funded by the U.S. Department of Energy.
 */


#include "flyme.h"
#include "flyme_glob.h"


/* TCL Initialization Routine */

int Tcl_AppInit();


/* TCL C Command Routines */

int get_flyme_default_dir_proc();
int register_menuvar_proc();
int load_bitmap_file_proc();
int set_zoom_factor_proc();
int set_radar_range_proc();
int canvas_handle_proc();
int set_play_mode_proc();
int fix_help_line_proc();
int double_click_proc();
int region_slide_proc();
int strip_label_proc();
int resize_stv_proc();
int title_info_proc();
int set_opts_proc();
int pad_proc();


void quit_proc();
void usage();


/* MAIN */

int
main( argc, argv )
int argc;
char **argv;
{
	char **mung_argv;
	int mung_argc;
	int i;

	/* Mung Argv to Circumvent TK Stupidity */

	mung_argc = argc + 1;

	mung_argv = (char **) malloc( (STV_SIZE_T) ( mung_argc + 1 )
		* sizeof( char * ) );
	memcheck( mung_argv, "Munged Argv List" );

	mung_argv[0] = argv[0];

	mung_argv[1] = copy_str( "-" );

	for ( i=2 ; i < mung_argc ; i++ )
		mung_argv[i] = argv[ i - 1 ];

	mung_argv[ mung_argc ] = (char *) NULL;

	/* Pass to TK */

	Tk_Main( mung_argc, mung_argv, Tcl_AppInit );

	return( 0 );
}


int Tcl_AppInit( itp )
Tcl_Interp *itp;
{
	/* Save Interpreter */

	interp = itp;

	/* Initialize TCL & TK */

	if ( tcl_init() == TCL_ERROR )
		return( TCL_ERROR );

	/* Get Main Window */

	Top = Tk_MainWindow( interp );

	if ( Top == NULL )
		return( TCL_ERROR );

	Disp = Tk_Display( Top );

	/* Initialize Program Constants */

	if ( read_args() == TCL_ERROR )
		return( TCL_ERROR );

	/* Initialize Program Constants & Structs */

	if ( program_init() == TCL_ERROR )
		return( TCL_ERROR );

	/* Initialize Modules */

	if ( do_stv_init() == TCL_ERROR )
		return( TCL_ERROR );

	if ( window_init() == TCL_ERROR )
		return( TCL_ERROR );

	/* Set Up User-Specific Startup File */

#if ( TCL_MAJOR_VERSION > 7 ) || ( TCL_MINOR_VERSION >= 5 )

	SET_TCL_GLOBAL( interp, "tcl_rcFileName", "~/.xpvmrc" );

#else

	tcl_RcFileName = "~/.flymerc";

#endif

	/* Return to tkMain to Start Tk_MainLoop() */

	return( TCL_OK );
}


read_args()
{
	char tmp[1024];

	char *argv_str;

	int i, j, k;
	int do_usage;
	int len;

	argv_str = GET_TCL_GLOBAL( interp, "argv" );

	if ( Tcl_SplitList( interp, argv_str, &Argc, &Argv ) == TCL_ERROR )
		return( TCL_ERROR );

	TIMER_INTERVAL = 50;

	GRID_SPACING = 10000;

	ROTOR_CLICKS = 12;

	FREQ = 1;

	HEADING_GOAL_INDEX = -1;
	VELOCITY_GOAL_INDEX = -1;
	RADAR_RANGE_INDEX = -1;
	FIRE_BULLET_INDEX = -1;
	SHIELDED_INDEX = -1;
	HELI_INDEX = -1;
	RELEASE_INDEX = -1;

	vflag = 0;
	qflag = 0;

	stv_setopt( stvOptNull, stvOptVerbose, vflag ? STV_DEBUG_ALL : 0 );

	do_usage = 0;

	for ( i=0 ; i < Argc ; i++ )
	{
		if ( Argv[i][0] == '-' )
		{
			k = i + 1;

			len = strlen( Argv[i] );

			for ( j=1 ; j < len ; j++ )
			{
				switch ( Argv[i][j] )
				{
					case 'F':
					{
						FREQ = atoi( Argv[k++] );

						break;
					}

					case 'I':
					{
						TIMER_INTERVAL = atoi( Argv[k++] );

						break;
					}

					case 'G':
					{
						GRID_SPACING = atoi( Argv[k++] );

						break;
					}

					case 'C':
					{
						ROTOR_CLICKS = atoi( Argv[k++] );

						break;
					}

					case 'v':
					{
						vflag++;

						stv_setopt( stvOptNull, stvOptVerbose,
							vflag ? STV_DEBUG_ALL : 0 );

						break;
					}

					case 'q':
					{
						qflag++;

						break;
					}

					default:
					{
						printf( "Unknown Option -%c\n", Argv[i][j] );

						do_usage++;

						break;
					}
				}
			}

			i = k - 1;
		}
	}

	if ( do_usage )
		usage();

	return( TCL_OK );
}


void
usage()
{
	printf( "\nusage:  flyme [-v] " );
	printf( "[-F <freq>] [-I <intrvl>] [-C <clicks>]\n\n" );

	printf( "where:\n" );
	printf( "------\n" );

	printf( "-v\t\t   = Verbose Operation\n" );

	printf( "-q\t\t   = Quiet Operation\n" );

	printf( "-F <freq>\t   = Set STV Frame Frequency\n" );

	printf( "-I <intrvl>\t   = Set Helicopter Handle Interval\n" );

	printf( "-G <spacing>\t   = Set Grid Spacing for Display\n" );

	printf( "-C <clicks>\t   = Set Number of Rotor Rotation Clicks\n" );

	printf( "\n" );

	exit( 0 );
}


int
program_init()
{
	struct passwd *pw;

	char *flyme_dir;
	char *home;

	char tmp[1024];

	int i;

	/* Seed Random Generator */

	srandom( (int) time( (time_t *) NULL ) );

	/* Get FlyMe Main Directory */

	flyme_dir = getenv( "STV_ROOT" );

	if ( flyme_dir == NULL )
	{
		sprintf( tmp, "%s", FLYME_DEFAULT_DIR );

		if ( tmp[0] != '/' )
		{
			home = getenv( "HOME" );

			if ( home == NULL )
			{
				if ( (pw = getpwuid( getuid() )) != NULL )
					home = pw->pw_dir;
			
				else
					home = "/";
			}

			sprintf( tmp, "%s/%s/%s", home,
				STV_DEFAULT_DIR, FLYME_DEFAULT_DIR );
		}
	}

	else
		sprintf( tmp, "%s/%s", flyme_dir, FLYME_DEFAULT_DIR );

	FLYME_DIR = stv_copy_str( tmp );

	/* Initialize STV Globals */

	CURRENT_VFG = (STV_VIEW_FIELD_GROUP) NULL;

	RADAR_VFG = (STV_VIEW_FIELD_GROUP) NULL;

	CURRENT_VF = (STV_VIEW_FIELD) NULL;

	CURRENT_VP = (STV_VIEW_PARAM) NULL;

	VIEWER = (STV_VIEWER) NULL;

	stv_init_region( MAX_REGION );

	stv_init_region( VISREGION );

	stv_init_cell( VISCELL );

	INITIALIZED = stvFalse;

	CONNECTED = stvFalse;

	AUTO_ZOOM_FACTOR = 1.0;

	ZOOM_FACTOR = 1.0;

	ZOOM_LIMIT = stvFalse;

	RADAR_CHANGED = stvFalse;

	RADAR_RANGE = -1.0;

	PLAY_MODE = MONITORING;

	STEERING = stvFalse;

	HELI_RADIUS = -1;

	RESTART = 0;

	/* Initialize Heli Globals */

	RADAR_WIN = (WIN) NULL;

	MAIN_WIN = (WIN) NULL;

	CURRENT_HELI = (HELI) NULL;

	REGIONS = (REGION_STACK) NULL;

	STEER_RELEASE = stvTrue;

	STEER_ACTIVE = stvFalse;

	AUTO_STEER = stvFalse;

	AUTO_ZOOM = stvFalse;

	PTR_VALID = stvFalse;

	PTR_X = -1;
	PTR_Y = -1;

	REGION_ID = -1;

	BASE_SCORE = 0;

	MY_SCORE = 0;

	return( TCL_OK );
}


do_stv_init()
{
	int cc;

	cc = stv_myid( &MYTID );

	if ( !cc )
	{
		stv_mp_err( "stv_myid" );

		return( TCL_ERROR );
	}

	return( TCL_OK );
}


void
do_stv_cleanup()
{
	if ( CURRENT_HELI != NULL )
		stv_viewer_steering_release( CURRENT_VP, VIEWER );

	stv_init_region( MAIN_WIN->region );

	stv_init_region( MAX_REGION );

	stv_init_region( VISREGION );

	stv_init_cell( VISCELL );

	CURRENT_VFG = (STV_VIEW_FIELD_GROUP) NULL;

	RADAR_VFG = (STV_VIEW_FIELD_GROUP) NULL;

	/* Unset Flags */

	CURRENT_HELI = (HELI) NULL;

	INITIALIZED = stvFalse;

	CONNECTED = stvFalse;

	STEERING = stvFalse;

	/* Reset to Monitor Mode */

	SET_TCL_GLOBAL( interp, "play_mode", "Monitor" );

	Tcl_Eval( interp, "layout_main_panel" );

	Tcl_Eval( interp, "set_play_mode Monitor" );

	/* Sleep to Fix Nasty Race Condition. */
	/* NEVER USE THIS */

	stvSleep( 1 );
}


tcl_init()
{
	/* Initialize TCL / TK */

	if ( Tcl_Init( interp ) == TCL_ERROR )
		return( TCL_ERROR );

	if ( Tk_Init( interp ) == TCL_ERROR )
		return( TCL_ERROR );

	/* Create TCL Commands for Action Routines */

	Tcl_CreateCommand( interp, "canvas_handle",
		canvas_handle_proc, (ClientData) NULL, (vfp) NULL );

	Tcl_CreateCommand( interp, "region_slide",
		region_slide_proc, (ClientData) NULL, (vfp) NULL );

	Tcl_CreateCommand( interp, "set_play_mode",
		set_play_mode_proc, (ClientData) NULL, (vfp) NULL );

	Tcl_CreateCommand( interp, "set_zoom_factor",
		set_zoom_factor_proc, (ClientData) NULL, (vfp) NULL );

	Tcl_CreateCommand( interp, "set_radar_range",
		set_radar_range_proc, (ClientData) NULL, (vfp) NULL );

	/* Create TCL Utility Commands */

	Tcl_CreateCommand( interp, "get_flyme_default_dir",
		get_flyme_default_dir_proc, (ClientData) NULL, (vfp) NULL );

	Tcl_CreateCommand( interp, "register_menuvar",
		register_menuvar_proc, (ClientData) NULL, (vfp) NULL );

	Tcl_CreateCommand( interp, "load_bitmap_file",
		load_bitmap_file_proc, (ClientData) NULL, (vfp) NULL );

	Tcl_CreateCommand( interp, "fix_help_line",
		fix_help_line_proc, (ClientData) NULL, (vfp) NULL );

	Tcl_CreateCommand( interp, "double_click",
		double_click_proc, (ClientData) NULL, (vfp) NULL );

	Tcl_CreateCommand( interp, "strip_label",
		strip_label_proc, (ClientData) NULL, (vfp) NULL );

	Tcl_CreateCommand( interp, "resize_stv",
		resize_stv_proc, (ClientData) NULL, (vfp) NULL );

	Tcl_CreateCommand( interp, "title_info",
		title_info_proc, (ClientData) NULL, (vfp) NULL );

	Tcl_CreateCommand( interp, "set_opts",
		set_opts_proc, (ClientData) NULL, (vfp) NULL );

	Tcl_CreateCommand( interp, "quit",
		(Tcl_CmdProc *) quit_proc, (ClientData) NULL, (vfp) NULL );

	Tcl_CreateCommand( interp, "pad",
		pad_proc, (ClientData) NULL, (vfp) NULL );

	return( TCL_OK );
}


window_init()
{
	static char fname[1024];

	FILE *fptest;

	char *tmp;

	/* Create Interface - Run TCL Script */

	fptest = fopen( "flyme.tcl", "r" );

	if ( fptest != NULL )
	{
		strcpy( fname, "flyme.tcl" );

		fclose( fptest );
	}

	else
		sprintf( fname, "%s/flyme.tcl", FLYME_DIR );

	if ( Tcl_EvalFile( interp, fname ) == TCL_ERROR )
		return( TCL_ERROR );

	/* Read in TCL Global Constants */

	FM = GET_TCL_GLOBAL( interp, "FM" );
	FM_C = GET_TCL_GLOBAL( interp, "FM_C" );
	FM_R = GET_TCL_GLOBAL( interp, "FM_R" );
	FM_I = GET_TCL_GLOBAL( interp, "FM_I" );
	FM_IF = GET_TCL_GLOBAL( interp, "FM_IF" );
	FM_IS = GET_TCL_GLOBAL( interp, "FM_IS" );
	FM_IH = GET_TCL_GLOBAL( interp, "FM_IH" );

	/* Get TCL Spacing Constants */

	tmp = GET_TCL_GLOBAL( interp, "border_space" );
	BORDER_SPACE = atoi( tmp );

	tmp = GET_TCL_GLOBAL( interp, "frame_border" );
	FRAME_BORDER = atoi( tmp );

	tmp = GET_TCL_GLOBAL( interp, "row_height" );
	ROW_HEIGHT = atoi( tmp );

	tmp = GET_TCL_GLOBAL( interp, "col_width" );
	COL_WIDTH = atoi( tmp );

	/* Initialize Window Display Objects */

	RADAR_WIN = create_win();

	RADAR_WIN->canvas = FM_R;

	RADAR_WIN->htvar = "flyme_size";
	RADAR_WIN->wtvar = "flyme_size";

	RADAR_WIN->xoff = RADAR_WIN->yoff = 0;

	MAIN_WIN = create_win();

	MAIN_WIN->canvas = FM_C;

	MAIN_WIN->htvar = "flyme_height";
	MAIN_WIN->wtvar = "flyme_width";

	MAIN_WIN->xoff = MAIN_WIN->yoff = 0;

	/* Set Up Helicopter Handler */

	Tk_DoWhenIdle( heli_handle, (ClientData) NULL );

	return( TCL_OK );
}


/* ARGSUSED */
int
get_flyme_default_dir_proc( clientData, itp, argc, argv )
ClientData clientData;
Tcl_Interp *itp;
int argc;
char **argv;
{
	char tmp[1024];

	sprintf( tmp, "%s/%s", STV_DEFAULT_DIR, FLYME_DEFAULT_DIR );

	Tcl_SetResult( itp, stv_copy_str( tmp ), TCL_DYNAMIC );

	return( TCL_OK );
}

