
/* $Id: flyme.h,v 3.8 1999/11/03 01:33:09 kohl Exp $ */


/*
 *                      CUMULVS version 1.0:
 *      Computational Steering and Interactive Visualization
 *                   in Distributed Applications
 *           Oak Ridge National Laboratory, Oak Ridge TN.
 *      Authors:  James Arthur Kohl and Philip M. Papadopoulos
 *                   (C) 1996 All Rights Reserved
 *
 *                              NOTICE
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted
 * provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear
 * in supporting documentation.
 *
 * Neither the Institution, Oak Ridge National Laboratory, nor the
 * Authors make any representations about the suitability of this
 * software for any purpose.  This software is provided ``as is''
 * without express or implied warranty.
 *
 * CUMULVS was funded by the U.S. Department of Energy.
 */


/* System Header Files */

#include <stdio.h>
#include <sys/types.h>

#include <malloc.h>
#include <pwd.h>
#include <math.h>

#ifndef PI
#define PI (M_PI)
#endif


/* STV Header Files */

#include "stv.h"


/* TCL / TK Header Files */

#include <tk.h>

#include "const.h"


/* Data Structures */

struct heli_struct
{
	/* WIN */ struct win_struct *win;
	int hid;
	int steered;
	int shield_str;
	int shielded;
	int last_shield_str;
	int fire_count;
	int score;
	int flash;
	int updated;
	int screen_x, screen_y;
	int region_x, region_y;
	double heading;
	double vel_angle;
	double velocity;
	double radar_range;
	int rotor_posn;
	int rotor_id;
	int body_id;
	int tail_id;
	int radar_id;
	int shield_id;
	/* HELI */ struct heli_struct *next;
};

typedef struct heli_struct *HELI;


struct bullet_struct
{
	/* WIN */ struct win_struct *win;
	int bid;
	int updated;
	int screen_x, screen_y;
	int region_x, region_y;
	double vel_angle;
	int bullet_id;
	/* BULLET */ struct bullet_struct *next;
};

typedef struct bullet_struct *BULLET;


struct explosion_struct
{
	/* WIN */ struct win_struct *win;
	int eid;
	int updated;
	int screen_x, screen_y;
	int region_x, region_y;
	double angle;
	int strength;
	int radius;
	int explosion_id;
	/* EXPLOSION */ struct explosion_struct *next;
};

typedef struct explosion_struct *EXPLOSION;


struct grid_struct
{
	int id;
	int coord;
	/* GRID */ struct grid_struct *next;
};

typedef struct grid_struct *GRID;


struct win_struct
{
	char *canvas;
	char *htvar, *wtvar;
	double scale;
	int csx, csy;
	STV_REGION region;
	int xoff, yoff;
	HELI helis;
	BULLET bullets;
	EXPLOSION explosions;
	int border_id;
	GRID xgrid, ygrid;
};

typedef struct win_struct *WIN;


struct region_stack_struct
{
	STV_REGION region;
	/* REGION_STACK */ struct region_stack_struct *next;
};

typedef struct region_stack_struct *REGION_STACK;


typedef void (*vfp)();

typedef int (*ifp)();


/* Routines */

HELI	create_heli(), get_heli_hid(), add_heli(), closest_heli();

BULLET	create_bullet(), get_bullet_bid(), add_bullet();

EXPLOSION	create_explosion(), get_explosion_eid(), add_explosion(),
			closest_explosion();

WIN		create_win();

GRID	create_grid();

REGION_STACK	create_region_stack();

FILE	*fopen();

double	distance();
double	angle_of();
double	atof();

char	*upper_str();
char	*trunc_str();
char	*copy_str();
char	*pad_num();

void	activate_long_range_radar();
void	adjust_border();
void	attach_to_region();
void	center_display_to_heli();
void	center_region_to_heli();
void	check_auto_zoom();
void	draw_bullet();
void	draw_explosion();
void	draw_heli();
void	draw_radar();
void	draw_rotors();
void	do_stv_cleanup();
void	free_bullet();
void	free_explosion();
void	free_grid();
void	free_grid_list();
void	free_heli();
void	free_region_stack();
void	free_win();
void	handle_frame();
void	heli_handle();
void	memcheck();
void	push_region();
void	release_heli();
void	remove_missing_bullets();
void	remove_missing_explosions();
void	remove_missing_helis();
void	request_heli_steer();
void	reset_bullet_updated();
void	reset_explosion_updated();
void	reset_heli_updated();
void	reset_item_screen_coords();
void	set_display_scale();
void	set_display_to_heli();
void	set_visregion();
void	undraw_bullet();
void	undraw_explosion();
void	undraw_heli();
void	undraw_radar();
void	undraw_rotors();
void	update_info_displays();
void	update_win_stuff();

