#
# $Id: flyme.tcl,v 3.1 1998/01/27 16:03:19 kohl Exp $
#

#
#                       CUMULVS version 1.0:
#       Computational Steering and Interactive Visualization
#                    in Distributed Applications
#            Oak Ridge National Laboratory, Oak Ridge TN.
#       Authors:  James Arthur Kohl and Philip M. Papadopoulos
#                    (C) 1996 All Rights Reserved
#
#                               NOTICE
#
#  Permission to use, copy, modify, and distribute this software and
#  its documentation for any purpose and without fee is hereby granted
#  provided that the above copyright notice appear in all copies and
#  that both the copyright notice and this permission notice appear
#  in supporting documentation.
#
#  Neither the Institution, Oak Ridge National Laboratory, nor the
#  Authors make any representations about the suitability of this
#  software for any purpose.  This software is provided ``as is''
#  without express or implied warranty.
#
#  CUMULVS was funded by the U.S. Department of Energy.
#

#
# Main TCL Source File
#

#
# Hide Main Window
#

wm withdraw .

#
# Setup Main TCL Directories
#

set ckenv [ info exists env(STV_ROOT) ]

if { $ckenv == 1 } \
	{ set flyme_directory "$env(STV_ROOT)/examples/flyme" } \
\
else \
{
	set home    [ glob ~ ]
	set def     [ get_flyme_default_dir ]

	set flyme_directory "$home/$def"
}

set help_dir $flyme_directory/help

#
# Source Main TCL Procs
#

set ckdir_procs [ file exists procs.tcl ]

if { $ckdir_procs == 1 } \
	{ source procs.tcl } \
\
else \
	{ source $flyme_directory/procs.tcl }

#
# Source TCL Utility Procs
#

set ckdir_util [ file exists util.tcl ]

if { $ckdir_util == 1 } \
	{ source util.tcl } \
\
else \
	{ source $flyme_directory/util.tcl }

#
# Startup Message
#

puts ""

puts -nonewline "Initializing FlyMe..."
flush stdout

#
# Proc Debug Flag
#

set proc_debug FALSE
#set proc_debug TRUE

#
# Create Toplevel Window
#

set main_height 800
set main_width 600

toplevel .flyme

wm withdraw .flyme

set_geometry -1 -1 10 10

wm minsize .flyme 100 100

set depth [ winfo depth .flyme ]

puts -nonewline "."
flush stdout

#
# Set up Colors
#

if { $depth == 1 } \
{
	set active_fg_color white

	set selector_color black

	set fg_color black

	set bg_color white
} \
\
else \
{
	set active_fg_color purple

	set selector_color purple

	set fg_color blue

	set bg_color #d9d9d9
}

#
# Reset Main Window Background
#

.flyme configure -background $bg_color

#
# Set Default Data Font (Fixed Width)
#

set data_font "-*-courier-bold-r-normal--14-*-*-*-*-*-*-*"

puts -nonewline "."
flush stdout

#
# Load Bitmaps
#

set xbm_dir "$flyme_directory/xbm"

load_bitmap_file "$xbm_dir/up.xbm"
load_bitmap_file "$xbm_dir/down.xbm"
load_bitmap_file "$xbm_dir/left.xbm"
load_bitmap_file "$xbm_dir/right.xbm"

load_bitmap_file "$xbm_dir/ne.xbm"
load_bitmap_file "$xbm_dir/nw.xbm"
load_bitmap_file "$xbm_dir/sw.xbm"
load_bitmap_file "$xbm_dir/se.xbm"

puts -nonewline "."
flush stdout

#
# Create Dummy Frame for Text Sizing (see get_real_text_size{})
#

frame .flyme.dummy

get_base_text_size

#
# Set Up Other Spacing Constants
#

set scroll_width 20

set border_space 10

set frame_border 2

#
# Help Window Constants
#

set min_help_height [ expr (3 * $scroll_width) + (2 * $border_space) ]

set min_help_width 250

set help_stagger_index 0

puts -nonewline "."
flush stdout

#
# Create Title & Version Number
#

set title_str "F l y   M e   [ title_info ]"

label .flyme.title -text $title_str -foreground $fg_color -background $bg_color

place .flyme.title -relx 0.5 -y $border_space -anchor n

update

puts -nonewline "."
flush stdout

#
# Get Main Font for Entry Widgets
#

set main_font [ .flyme.title cget -font ]

#
# Create Main Panel
#

set FM .flyme

set FM_C $FM.canvas

set FM_R $FM.radar

set FM_I $FM.info

set FM_IF $FM_I.fire_canvas

set FM_IS $FM_I.shield_canvas

set FM_IH $FM_I.heading_canvas

label $FM.message -text {Status:   Welcome to Fly Me} \
	-foreground $fg_color -background $bg_color \
	-font $main_font -relief sunken -anchor nw

set msg_list ""

button $FM.full_unzoom -text "Full Un-Zoom" -padx 1 -pady 1 \
	-font $main_font \
	-foreground $fg_color -background $bg_color \
	-activeforeground $active_fg_color \
	-command "canvas_handle POP -1 -1"

button $FM.unzoom -text "Un-Zoom Region" -padx 1 -pady 1 \
	-font $main_font \
	-foreground $fg_color -background $bg_color \
	-activeforeground $active_fg_color \
	-command "canvas_handle pop -1 -1"

button $FM.quit -text "I'm Outta Here..." -padx 1 -pady 1 \
	-font $main_font \
	-foreground $fg_color -background $bg_color \
	-activeforeground $active_fg_color \
	-command "quit"

button $FM.up -bitmap "@$xbm_dir/up.xbm" \
	-foreground $fg_color -background $bg_color \
	-activeforeground $active_fg_color \
	-command "region_slide up"

button $FM.down -bitmap "@$xbm_dir/down.xbm" \
	-foreground $fg_color -background $bg_color \
	-activeforeground $active_fg_color \
	-command "region_slide down"

button $FM.left -bitmap "@$xbm_dir/left.xbm" \
	-foreground $fg_color -background $bg_color \
	-activeforeground $active_fg_color \
	-command "region_slide left"

button $FM.right -bitmap "@$xbm_dir/right.xbm" \
	-foreground $fg_color -background $bg_color \
	-activeforeground $active_fg_color \
	-command "region_slide right"

button $FM.ne -bitmap "@$xbm_dir/ne.xbm" \
	-foreground $fg_color -background $bg_color \
	-activeforeground $active_fg_color \
	-command "region_slide ne"

button $FM.nw -bitmap "@$xbm_dir/nw.xbm" \
	-foreground $fg_color -background $bg_color \
	-activeforeground $active_fg_color \
	-command "region_slide nw"

button $FM.sw -bitmap "@$xbm_dir/sw.xbm" \
	-foreground $fg_color -background $bg_color \
	-activeforeground $active_fg_color \
	-command "region_slide sw"

button $FM.se -bitmap "@$xbm_dir/se.xbm" \
	-foreground $fg_color -background $bg_color \
	-activeforeground $active_fg_color \
	-command "region_slide se"

radiobutton $FM.monitor -text { Monitor } \
	-padx 1 -pady 1 \
	-command "handleMode" \
	-bd $frame_border -relief sunken -value "Monitor" \
	-variable play_mode \
	-selectcolor $selector_color -foreground $fg_color -background $bg_color \
	-activeforeground $active_fg_color

radiobutton $FM.play -text { Play } \
	-padx 1 -pady 1 \
	-command "handleMode" \
	-bd $frame_border -relief sunken -value "Play" \
	-variable play_mode \
	-selectcolor $selector_color -foreground $fg_color -background $bg_color \
	-activeforeground $active_fg_color

set play_mode "Monitor"

checkbutton $FM.auto_zoom -text { Auto Zoom } \
	-padx 1 -pady 1 \
	-onvalue ON -offvalue OFF -variable auto_zoom \
	-command handleOpts \
	-bd $frame_border -relief sunken \
	-selectcolor $selector_color -foreground $fg_color -background $bg_color \
	-activeforeground $active_fg_color

set auto_zoom "OFF"

checkbutton $FM.auto_steer -text { Auto Steer } \
	-padx 1 -pady 1 \
	-onvalue ON -offvalue OFF -variable auto_steer \
	-command handleOpts \
	-bd $frame_border -relief sunken \
	-selectcolor $selector_color -foreground $fg_color -background $bg_color \
	-activeforeground $active_fg_color

set auto_steer "OFF"

puts -nonewline "."
flush stdout

label $FM.zoom_factor_label -text "Zoom Factor:" \
	-foreground $fg_color -background $bg_color

entry $FM.zoom_factor -font $main_font \
	-bd $frame_border -relief sunken \
	-foreground $fg_color -background $bg_color

entry_setup $FM.zoom_factor "zoom_factor_handle"

$FM.zoom_factor insert 0 1

label $FM.radar_range_label -text "Radar Range:" \
	-foreground $fg_color -background $bg_color

entry $FM.radar_range -font $main_font \
	-bd $frame_border -relief sunken \
	-foreground $fg_color -background $bg_color

entry_setup $FM.radar_range "radar_range_handle"

$FM.radar_range insert 0 50

label $FM.score_label -text "Score:          " \
	-foreground $fg_color -background $bg_color

canvas $FM_C -bd $frame_border -relief sunken -confine 1 -background $bg_color

canvas $FM_R -bd $frame_border -relief sunken -confine 1 -background $bg_color

frame $FM_I -bd $frame_border -relief sunken -background $bg_color

label $FM_I.fire_label -text "Fire Ready:" \
	-foreground $fg_color -background $bg_color

canvas $FM_IF -bd $frame_border -relief sunken -confine 1 -background $bg_color

label $FM_I.shield_label -text "Shield:" \
	-foreground $fg_color -background $bg_color

canvas $FM_IS -bd $frame_border -relief sunken -confine 1 -background $bg_color

canvas $FM_IH -bd $frame_border -relief sunken -confine 1 -background $bg_color

set FM_IH_N [ $FM_IH create text -100 -100 -text "N" -anchor n \
	-fill $fg_color ]

set FM_IH_S [ $FM_IH create text -100 -100 -text "S" -anchor s \
	-fill $fg_color ]

set FM_IH_E [ $FM_IH create text -100 -100 -text "E" -anchor e \
	-fill $fg_color ]

set FM_IH_W [ $FM_IH create text -100 -100 -text "W" -anchor w \
	-fill $fg_color ]

label $FM_I.speed -text "Speed:" \
	-foreground $fg_color -background $bg_color

label $FM_I.score -text "Score:" \
	-foreground $fg_color -background $bg_color

puts -nonewline "."
flush stdout

layout_main_panel

puts -nonewline "."
flush stdout

#
# Main Event Bindings
#

bind $FM.zoom_factor <Enter>	"focus $FM.zoom_factor"
bind $FM.zoom_factor <Leave>	"zoom_factor_handle; focus $FM"

bind $FM.radar_range <Enter>	"focus $FM.radar_range"
bind $FM.radar_range <Leave>	"radar_range_handle; focus $FM"

bind $FM_C <ButtonPress-1>		"canvas_handle press %x %y"
bind $FM_C <Button1-Motion>		"canvas_handle slide %x %y"
bind $FM_C <ButtonRelease-1>	"canvas_handle release %x %y"

bind $FM_C <ButtonPress-2>		"canvas_handle fire %x %y"

bind $FM_C <ButtonPress-3>		"canvas_handle shield %x %y"

bind $FM_C <Enter>				"focus $FM_C; canvas_handle enter %x %y"
bind $FM_C <Motion>				"canvas_handle move %x %y"
bind $FM_C <Leave>				"canvas_handle leave %x %y; focus $FM"

bind $FM_C <KeyPress>			"canvas_handle %A %x %y"

bind $FM <Configure>			"resize_main_panel"

bind $FM <Destroy>				"quit"

puts -nonewline "."
flush stdout

#
# Button Help Bindings
#


#
# Actually Bring Up Main XPVM Window
#

wm deiconify .flyme

puts "done."
flush stdout

#
# Reset Welcome Message
#

setMsg "Welcome to Fly Me"

update

