
/* $Id: heli.c,v 3.5 2000/06/12 18:01:00 kohl Exp $ */


/*
 *                      CUMULVS version 1.0:
 *      Computational Steering and Interactive Visualization
 *                   in Distributed Applications
 *           Oak Ridge National Laboratory, Oak Ridge TN.
 *      Authors:  James Arthur Kohl and Philip M. Papadopoulos
 *                   (C) 1996 All Rights Reserved
 *
 *                              NOTICE
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted
 * provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear
 * in supporting documentation.
 *
 * Neither the Institution, Oak Ridge National Laboratory, nor the
 * Authors make any representations about the suitability of this
 * software for any purpose.  This software is provided ``as is''
 * without express or implied warranty.
 *
 * CUMULVS was funded by the U.S. Department of Energy.
 */


#include "flyme.h"
#include "flyme_glob.h"


void
attach_to_region()
{
	STV_PARTICLEPTR PT;

	STV_VIEW_PARAM VP;

	STV_PARAMPTR P;

	int i;

	if ( CONNECTED )
	{
		printf( "\nSending FieldHaltAll\n\n" );
		fflush( stdout );

		stv_viewer_send_FieldHaltAll( CURRENT_VFG, VIEWER );

		CONNECTED = stvFalse;
	}

	if ( !INITIALIZED )
	{
		INITIALIZED =
			( stv_viewer_init( &VIEWER, MYTID, "Region", stvFalse )
				== stvStatusOk );
	}

	if ( INITIALIZED )
	{
		/* Clear View Field Selections */

		stv_clear_view_field_select( VIEWER );

		/* Select Desired Helicopter Fields */

		if ( !select_field( "hid", stvInt ) ) return;

		if ( !select_field( "steered", stvInt ) ) return;

		if ( !select_field( "shield_str", stvInt ) ) return;

		if ( !select_field( "shielded", stvInt ) ) return;

		if ( !select_field( "fire_count", stvInt ) ) return;

		if ( !select_field( "score", stvInt ) ) return;

		if ( !select_field( "heading", stvDouble ) ) return;

		if ( !select_field( "hvel_angle", stvDouble ) ) return;

		if ( !select_field( "hvel_magn", stvDouble ) ) return;

		if ( !select_field( "radar_range", stvDouble ) ) return;

		/* Select Desired Bullet Fields */

		if ( !select_field( "bid", stvInt ) ) return;

		if ( !select_field( "bvel_angle", stvDouble ) ) return;

		/* Select Desired Explosion Fields */

		if ( !select_field( "eid", stvInt ) ) return;

		if ( !select_field( "explosion_angle", stvDouble ) ) return;

		if ( !select_field( "explosion_strength", stvInt ) ) return;

		if ( !select_field( "explosion_radius", stvInt ) ) return;

		/* Set Up Vis Region */

		PT = stv_get_particle_name( "helicopter",
			VIEWER->particles, VIEWER->nparticles );

		if ( PT == NULL )
		{
			printf( "\nError: Particle \"helicopter\" Not Found\n\n" );
			fflush( stdout );

			return;
		}

		while ( pop_region( MAIN_WIN->region ) );

		STV_REGION_MIN( MAIN_WIN->region, 0 ) =
			STV_REGION_MIN( MAX_REGION, 0 ) = PT->glb[0];

		STV_REGION_MAX( MAIN_WIN->region, 0 ) =
			STV_REGION_MAX( MAX_REGION, 0 ) = PT->gub[0];

		STV_REGION_MIN( MAIN_WIN->region, 1 ) =
			STV_REGION_MIN( MAX_REGION, 1 ) = PT->glb[1];

		STV_REGION_MAX( MAIN_WIN->region, 1 ) =
			STV_REGION_MAX( MAX_REGION, 1 ) = PT->gub[1];

		STV_CELL_PARTICLE( VISCELL ) = 1;

		/* Request View Field Group */

		CURRENT_VF = stv_get_view_field_name( "hid",
			VIEWER->vfields, VIEWER->nvfields );

		set_visregion( stvFalse );

		CURRENT_VFG = stv_viewer_request_field( VIEWER,
			VISREGION, VISCELL, FREQ, stvFalse, &RESTART );

		if ( CURRENT_VFG == NULL )
			return;

		CONNECTED = stvTrue;

		MAIN_WIN->xoff = 2 * FRAME_BORDER;
		MAIN_WIN->yoff = 2 * FRAME_BORDER;

		set_display_scale( MAIN_WIN );

		adjust_border( MAIN_WIN );

		/* Get Steering Indices */

		VP = stv_get_view_param_name( "helicopter",
			VIEWER->vparams, VIEWER->nvparams );

		if ( VP != NULL )
		{
			P = VP->param;

			for ( i=0 ; i < P->num ; i++ )
			{
				if ( !strcmp( P->pnames[i], "heading goal" ) )
					HEADING_GOAL_INDEX = i;

				else if ( !strcmp( P->pnames[i], "velocity goal" ) )
					VELOCITY_GOAL_INDEX = i;

				else if ( !strcmp( P->pnames[i], "radar range" ) )
					RADAR_RANGE_INDEX = i;

				else if ( !strcmp( P->pnames[i], "fire bullet" ) )
					FIRE_BULLET_INDEX = i;

				else if ( !strcmp( P->pnames[i], "shielded" ) )
					SHIELDED_INDEX = i;

				else if ( !strcmp( P->pnames[i], "index" ) )
					HELI_INDEX = i;

				else if ( !strcmp( P->pnames[i], "release control" ) )
					RELEASE_INDEX = i;
			}
		}

		/* Get Heli Radius */

		VP = stv_get_view_param_name( "heli radius",
			VIEWER->vparams, VIEWER->nvparams );

		if ( VP != NULL )
			HELI_RADIUS = STV_VALUE_OF( VP->param->vars[0], int );
	}
}


void
activate_long_range_radar()
{
	STV_PARTICLEPTR PT;

	STV_VIEW_PARAM VP;

	STV_PARAMPTR P;

	int i;

	/* Get View Field Group (if not already) */

	if ( RADAR_VFG == NULL )
	{
		/* Clear View Field Selections */

		stv_clear_view_field_select( VIEWER );

		/* Select Desired Helicopter Fields */

		if ( !select_field( "hid", stvInt ) ) return;

		if ( !select_field( "steered", stvInt ) ) return;

		if ( !select_field( "shield_str", stvInt ) ) return;

		if ( !select_field( "shielded", stvInt ) ) return;

		if ( !select_field( "score", stvInt ) ) return;

		if ( !select_field( "heading", stvDouble ) ) return;

		if ( !select_field( "radar_range", stvDouble ) ) return;

		/* Select Desired Explosion Fields */

		if ( !select_field( "eid", stvInt ) ) return;

		if ( !select_field( "explosion_angle", stvDouble ) ) return;

		if ( !select_field( "explosion_radius", stvInt ) ) return;

		/* Set Whole Region */

		stv_copy_region( RADAR_WIN->region, MAX_REGION );

		/* Request View Field Group */

		RADAR_VFG = stv_viewer_request_field( VIEWER,
			RADAR_WIN->region, VISCELL, 10 * FREQ, stvFalse, &RESTART );

		if ( RADAR_VFG == NULL )
			return;

		RADAR_WIN->xoff = 2 * FRAME_BORDER;
		RADAR_WIN->yoff = 2 * FRAME_BORDER;
	}

	set_display_scale( RADAR_WIN );

	adjust_border( RADAR_WIN );
}


select_field( name, type )
char *name;
int type;
{
	STV_VIEW_FIELD VF;

	VF = stv_get_view_field_name( name,
		VIEWER->vfields, VIEWER->nvfields );

	if ( VF == NULL )
	{
		printf( "\nError: View Field \"%s\" Not Found\n\n", name );
		fflush( stdout );

		return( stvFalse );
	}
	
	VF->selected = stvTrue;

	VF->view_type = type;

	return( stvTrue );
}


void
set_visregion( setfreqflag )
int setfreqflag;
{
	int newfreq;
	int wt, ht;
	int maxsz;
	int sz;

	ht = STV_REGION_MAX( MAIN_WIN->region, 1 )
		- STV_REGION_MIN( MAIN_WIN->region, 1 ) + 1;

	ht = ht < 2000 ? 2000 : ht;

	wt = STV_REGION_MAX( MAIN_WIN->region, 0 )
		- STV_REGION_MIN( MAIN_WIN->region, 0 ) + 1;

	wt = wt < 2000 ? 2000 : wt;

	STV_REGION_MIN( VISREGION, 0 )
		= STV_REGION_MIN( MAIN_WIN->region, 0 ) - ( wt / 2 );

	STV_REGION_MAX( VISREGION, 0 )
		= STV_REGION_MAX( MAIN_WIN->region, 0 ) + ( wt / 2 );

	STV_REGION_MIN( VISREGION, 1 )
		= STV_REGION_MIN( MAIN_WIN->region, 1 ) - ( ht / 2 );

	STV_REGION_MAX( VISREGION, 1 )
		= STV_REGION_MAX( MAIN_WIN->region, 1 ) + ( ht / 2 );

	stv_check_region( CURRENT_VF->field, VISREGION, VISCELL );

	/* Set Frame Freq Based on Vis Region Ratio (When Monitoring) */

	if ( PLAY_MODE == MONITORING )
	{
		if ( wt < ht )
		{
			maxsz = STV_REGION_MAX( MAX_REGION, 0 )
				- STV_REGION_MIN( MAX_REGION, 0 ) + 1;

			sz = wt;
		}

		else
		{
			maxsz = STV_REGION_MAX( MAX_REGION, 1 )
				- STV_REGION_MIN( MAX_REGION, 1 ) + 1;

			sz = ht;
		}

		newfreq = ( sz / maxsz ) * 10;

		newfreq = newfreq < 1 ? 1 : newfreq;
	}

	else
		newfreq = 1;

	if ( newfreq != FREQ )
	{
		FREQ = newfreq;

		if ( setfreqflag )
			stv_viewer_set_VisFrequency( CURRENT_VFG, VIEWER, FREQ );
	}
}


void
set_display_scale( W )
WIN W;
{
	char *tmp;

	double hscale, wscale;

	int height, width;
	int fwt, fht;

	/* Get Canvas Dimensions */

	tmp = GET_TCL_GLOBAL( interp, W->htvar );
	height = atoi( tmp );

	tmp = GET_TCL_GLOBAL( interp, W->wtvar );
	width = atoi( tmp );

	height -= (4 * FRAME_BORDER);
	width -= (4 * FRAME_BORDER);

	W->csy = height / 2;
	W->csx = width / 2;

	/* Check for Valid (Updated) Canvas Size */

	if ( height > 0 && width > 0 )
	{
		fht = STV_REGION_MAX( W->region, 1 )
			- STV_REGION_MIN( W->region, 1 ) + 1;

		fwt = STV_REGION_MAX( W->region, 0 )
			- STV_REGION_MIN( W->region, 0 ) + 1;

		hscale = (double) height / (double) fht;
		wscale = (double) width / (double) fwt;

		W->scale = hscale < wscale ? hscale : wscale;
	}
}


/* Adjust Field Border */

void
adjust_border( W )
WIN W;
{
	GRID Glast;
	GRID G;

	double zoom;

	int maxx, maxy;
	int minx, miny;
	int rcoord;
	int rstart;
	int x1, y1;
	int x2, y2;
	int wt, ht;
	int coord;
	int r;

	if ( CURRENT_HELI != NULL && CURRENT_HELI->win == W )
	{
		zoom = AUTO_ZOOM ? AUTO_ZOOM_FACTOR : ZOOM_FACTOR;

		if ( !EQUIV( zoom, 0.0 ) )
			r = (int) ( CURRENT_HELI->radar_range / zoom );
	
		else
			r = (int) ( CURRENT_HELI->radar_range );

		minx = CURRENT_HELI->region_x - r;
		maxx = CURRENT_HELI->region_x + r;

		miny = CURRENT_HELI->region_y - r;
		maxy = CURRENT_HELI->region_y + r;

		minx = minx < STV_REGION_MIN( MAX_REGION, 0 ) ?
			STV_REGION_MIN( MAX_REGION, 0 ) : minx;

		miny = miny < STV_REGION_MIN( MAX_REGION, 1 ) ?
			STV_REGION_MIN( MAX_REGION, 1 ) : miny;

		maxx = maxx > STV_REGION_MAX( MAX_REGION, 0 ) ?
			STV_REGION_MAX( MAX_REGION, 0 ) : maxx;

		maxy = maxy > STV_REGION_MAX( MAX_REGION, 1 ) ?
			STV_REGION_MAX( MAX_REGION, 1 ) : maxy;
	}

	else
	{
		minx = STV_REGION_MIN( W->region, 0 );
		maxx = STV_REGION_MAX( W->region, 0 );

		miny = STV_REGION_MIN( W->region, 1 );
		maxy = STV_REGION_MAX( W->region, 1 );
	}

	ht = maxy - miny + 1;
	wt = maxx - minx + 1;

	x1 = y1 = 2 * FRAME_BORDER;

	x2 = REGION_TO_SCREEN( wt, 0, x1, W->scale );
	y2 = REGION_TO_SCREEN( ht, 0, y1, W->scale );

	if ( W->border_id == -1 )
	{
		W->border_id = DRAW_RECT( interp, W->canvas,
			x1, y1, x2, y2, "", "blue", 2 );
	}

	else
		RECT_COORDS( interp, W->canvas, W->border_id, x1, y1, x2, y2 );

	/* Adjust X Grid Lines */

	rstart = minx + ( GRID_SPACING - ( minx % GRID_SPACING ) );

	G = W->xgrid;

	Glast = (GRID) NULL;

	for ( rcoord=rstart ; rcoord < maxx ; rcoord += GRID_SPACING )
	{
		if ( G == NULL )
		{
			if ( Glast != NULL )
				G = Glast->next = create_grid();
			
			else
				G = W->xgrid = create_grid();
		}

		coord = REGION_TO_SCREEN( rcoord,
			STV_REGION_MIN( W->region, 0 ), W->xoff, W->scale );

		G->coord = coord;

		if ( G->id == -1 )
		{
			G->id = DRAW_LINE( interp, W->canvas,
				coord, y1, coord, y2 );
		}

		else
		{
			RECT_COORDS( interp, W->canvas, G->id,
				coord, y1, coord, y2 );
		}

		Glast = G;

		G = G->next;
	}

	/* Free Up Leftover Grid Lines */

	if ( G != NULL )
	{
		if ( Glast != NULL )
			free_grid_list( W, &(Glast->next) );
		
		else
			free_grid_list( W, &(W->xgrid) );
	}

	/* Adjust Y Grid Lines */

	rstart = miny + ( GRID_SPACING - ( miny % GRID_SPACING ) );

	G = W->ygrid;

	Glast = (GRID) NULL;

	for ( rcoord=rstart ; rcoord < maxy ; rcoord += GRID_SPACING )
	{
		if ( G == NULL )
		{
			if ( Glast != NULL )
				G = Glast->next = create_grid();
			
			else
				G = W->ygrid = create_grid();
		}

		coord = REGION_TO_SCREEN( rcoord,
			STV_REGION_MIN( W->region, 1 ), W->yoff, W->scale );

		G->coord = coord;

		if ( G->id == -1 )
		{
			G->id = DRAW_LINE( interp, W->canvas,
				x1, coord, x2, coord );
		}

		else
		{
			RECT_COORDS( interp, W->canvas, G->id,
				x1, coord, x2, coord );
		}

		Glast = G;

		G = G->next;
	}

	/* Free Up Leftover Grid Lines */

	if ( G != NULL )
	{
		if ( Glast != NULL )
			free_grid_list( W, &(Glast->next) );
		
		else
			free_grid_list( W, &(W->ygrid) );
	}
}


/* Check for Migration of Heli from Center of Region */

check_heli_migration( H )
HELI H;
{
	int diffx, diffy;
	int sizex, sizey;
	int cx, cy;
	int hx, hy;
	int ret;

	sizex = STV_REGION_MAX( VISREGION, 0 )
			- STV_REGION_MIN( VISREGION, 0 ) + 1;

	sizey = STV_REGION_MAX( VISREGION, 1 )
			- STV_REGION_MIN( VISREGION, 1 ) + 1;

	cx = STV_REGION_MIN( VISREGION, 0 ) + ( sizex / 2 );

	cy = STV_REGION_MIN( VISREGION, 1 ) + ( sizey / 2 );

	hx = H->region_x;

	hy = H->region_y;

	diffx = ( hx > cx ) ? hx - cx : cx - hx;

	diffy = ( hy > cy ) ? hy - cy : cy - hy;

	ret = stvFalse;

	if ( diffx > sizex / 4 )
	{
		/* Check for Boundaries */

		if ( hx > cx && STV_REGION_MAX( VISREGION, 0 )
				== STV_REGION_MAX( MAX_REGION, 0 ) )
		{
			ret = stvFalse;
		}

		else if ( hx < cx && STV_REGION_MIN( VISREGION, 0 )
				== STV_REGION_MIN( MAX_REGION, 0 ) )
		{
			ret = stvFalse;
		}

		else
			ret = stvTrue;
	}

	if ( !ret && diffy > sizey / 4 )
	{
		/* Check for Boundaries */

		if ( hy > cy && STV_REGION_MAX( VISREGION, 1 )
				== STV_REGION_MAX( MAX_REGION, 1 ) )
		{
			ret = stvFalse;
		}

		else if ( hy < cy && STV_REGION_MIN( VISREGION, 1 )
				== STV_REGION_MIN( MAX_REGION, 1 ) )
		{
			ret = stvFalse;
		}

		else
			ret = stvTrue;
	}
	
	return( ret );
}


/* Check for Auto Zooming */

void
check_auto_zoom()
{
	EXPLOSION E;

	HELI H;

	char *tmp;

	double close_dist;
	double close_heli;
	double close_exp;
	double thresh;
	double rdist0;
	double rdist;
	double ramp;

	int rx, ry;
	int width;
	int flag;
	int rwt;

	H = closest_heli( MAIN_WIN,
		CURRENT_HELI->screen_x, CURRENT_HELI->screen_y,
		CURRENT_HELI, &close_heli );

	E = closest_explosion( MAIN_WIN,
		CURRENT_HELI->screen_x, CURRENT_HELI->screen_y, &close_exp );

	flag = 0;

	if ( H != NULL || E != NULL )
	{
		/* Determine Closest Item */

		if ( ( H != NULL && close_heli < close_exp ) || E == NULL )
		{
			close_dist = close_heli;

			rx = H->region_x;
			ry = H->region_y;
		}

		else
		{
			close_dist = close_exp;

			rx = E->region_x;
			ry = E->region_y;
		}

		/* Get Canvas Width */

		tmp = GET_TCL_GLOBAL( interp, MAIN_WIN->wtvar );
		width = atoi( tmp );

		width -= (4 * FRAME_BORDER);

		thresh = 3.0 * (double) width / 8.0;

		if ( close_dist < thresh || AUTO_ZOOM_FACTOR > 1.0 )
		{
			rdist = distance(
				(double) CURRENT_HELI->region_x,
				(double) CURRENT_HELI->region_y,
				(double) rx, (double) ry );

			rdist0 = ( ( 2.0 * CURRENT_HELI->radar_range ) - 1.0 )
				* thresh / (double) width;

			if ( rdist < rdist0 )
			{
				ramp = ( rdist0 - rdist ) / rdist0;

				ramp *= ramp;

				rdist = rdist + ( ramp * 500.0 );
			}

			AUTO_ZOOM_FACTOR =
				( 2.0 * CURRENT_HELI->radar_range )
					/ ( ( (double) width * rdist / thresh ) + 1.0 );

			if ( AUTO_ZOOM_FACTOR < 1.0 )
				AUTO_ZOOM_FACTOR = 1.0;

			flag++;

			if ( vflag )
				printf( "Auto Zoom = %lf\n", AUTO_ZOOM_FACTOR );
		}
	}

	if ( !flag && AUTO_ZOOM_FACTOR != 1.0 )
	{
		AUTO_ZOOM_FACTOR = 1.0;
	
		if ( vflag )
			printf( "Auto Zoom = %lf\n", AUTO_ZOOM_FACTOR );

		flag++;
	}

	if ( flag )
		set_display_to_heli( CURRENT_HELI, stvFalse );
}


/* Set Vis Region to Heli's Radar */

void
set_display_to_heli( H, setvisflag )
HELI H;
int setvisflag;
{
	/* Reset Region Size Around Heli */

	center_region_to_heli( H );

	/* Set Display Scale For Full Region Size */

	set_display_scale( MAIN_WIN );

	/* Re-Center Region Around Heli with Modified Scale */

	center_region_to_heli( H );

	/* Adjust Border & Items */

	adjust_border( MAIN_WIN );

	reset_item_screen_coords( MAIN_WIN );

	/* Send New Vis Region */

	if ( setvisflag )
	{
		set_visregion( stvTrue );

		/* if is_static flag ever set, need to check return code here */
		stv_viewer_set_VisRegion( CURRENT_VFG, VIEWER,
			VISREGION, VISCELL );
	}
}


/* Center Display Region Over Heli */

void
center_display_to_heli( H )
HELI H;
{
	/* Center Region Around Heli */

	center_region_to_heli( H );

	/* Adjust Border & Update Items */

	adjust_border( MAIN_WIN );

	reset_item_screen_coords( MAIN_WIN );
}


void
center_region_to_heli( H )
HELI H;
{
	double zoom;

	int hold;
	int r;

	zoom = AUTO_ZOOM ? AUTO_ZOOM_FACTOR : ZOOM_FACTOR;

	if ( !EQUIV( zoom, 0.0 ) )
		r = (int) ( H->radar_range / zoom );
	
	else
		r = (int) ( H->radar_range );

	STV_REGION_MIN( MAIN_WIN->region, 0 ) = H->region_x - r;
	STV_REGION_MAX( MAIN_WIN->region, 0 ) = H->region_x + r;

	STV_REGION_MIN( MAIN_WIN->region, 1 ) = H->region_y - r;
	STV_REGION_MAX( MAIN_WIN->region, 1 ) = H->region_y + r;

	/* Handle X Coord */

	hold = 0;

	if ( STV_REGION_MIN( MAIN_WIN->region, 0 )
		< STV_REGION_MIN( MAX_REGION, 0 ) )
	{
		STV_REGION_MAX( MAIN_WIN->region, 0 ) +=
			STV_REGION_MIN( MAX_REGION, 0 )
				- STV_REGION_MIN( MAIN_WIN->region, 0 );
		
		STV_REGION_MIN( MAIN_WIN->region, 0 ) =
			STV_REGION_MIN( MAX_REGION, 0 );

		hold++;
	}

	else if ( STV_REGION_MAX( MAIN_WIN->region, 0 )
		> STV_REGION_MAX( MAX_REGION, 0 ) )
	{
		STV_REGION_MIN( MAIN_WIN->region, 0 ) -=
			STV_REGION_MAX( MAIN_WIN->region, 0 )
				- STV_REGION_MAX( MAX_REGION, 0 );

		STV_REGION_MAX( MAIN_WIN->region, 0 ) =
			STV_REGION_MAX( MAX_REGION, 0 );
	}

	if ( !hold )
	{
		MAIN_WIN->xoff = MAIN_WIN->csx
			- ( ( H->region_x - STV_REGION_MIN( MAIN_WIN->region, 0 ) )
				* MAIN_WIN->scale )
			+ ( 2 * FRAME_BORDER );
	}

	/* Handle Y Coord */

	hold = 0;

	if ( STV_REGION_MIN( MAIN_WIN->region, 1 )
		< STV_REGION_MIN( MAX_REGION, 1 ) )
	{
		STV_REGION_MAX( MAIN_WIN->region, 1 ) +=
			STV_REGION_MIN( MAX_REGION, 1 )
				- STV_REGION_MIN( MAIN_WIN->region, 1 );
		
		STV_REGION_MIN( MAIN_WIN->region, 1 ) =
			STV_REGION_MIN( MAX_REGION, 1 );

		hold++;
	}

	else if ( STV_REGION_MAX( MAIN_WIN->region, 1 )
		> STV_REGION_MAX( MAX_REGION, 1 ) )
	{
		STV_REGION_MIN( MAIN_WIN->region, 1 ) -=
			STV_REGION_MAX( MAIN_WIN->region, 1 )
				- STV_REGION_MAX( MAX_REGION, 1 );

		STV_REGION_MAX( MAIN_WIN->region, 1 ) =
			STV_REGION_MAX( MAX_REGION, 1 );
	}

	if ( !hold )
	{
		MAIN_WIN->yoff = MAIN_WIN->csy
			- ( ( H->region_y - STV_REGION_MIN( MAIN_WIN->region, 1 ) )
				* MAIN_WIN->scale )
			+ ( 2 * FRAME_BORDER );
	}
}


void
reset_item_screen_coords( W )
WIN W;
{
	EXPLOSION E;
	BULLET B;
	HELI H;

	/* Update Each Helicopter */

	H = W->helis;

	while ( H != NULL )
	{
		H->screen_x = REGION_TO_SCREEN( H->region_x,
			STV_REGION_MIN( W->region, 0 ), W->xoff, W->scale );

		H->screen_y = REGION_TO_SCREEN( H->region_y,
			STV_REGION_MIN( W->region, 1 ), W->yoff, W->scale );

		H->updated = UPDATED;

		H = H->next;
	}

	/* Update Each Bullet */

	B = W->bullets;

	while ( B != NULL )
	{
		B->screen_x = REGION_TO_SCREEN( B->region_x,
			STV_REGION_MIN( W->region, 0 ), W->xoff, W->scale );

		B->screen_y = REGION_TO_SCREEN( B->region_y,
			STV_REGION_MIN( W->region, 1 ), W->yoff, W->scale );

		B->updated = UPDATED;

		B = B->next;
	}

	/* Update Explosions */

	E = W->explosions;

	while ( E != NULL )
	{
		E->screen_x = REGION_TO_SCREEN( E->region_x,
			STV_REGION_MIN( W->region, 0 ), W->xoff, W->scale );

		E->screen_y = REGION_TO_SCREEN( E->region_y,
			STV_REGION_MIN( W->region, 1 ), W->yoff, W->scale );

		E->updated = UPDATED;

		E = E->next;
	}
}


item_in_display( W, x, y, size )
WIN W;
int x, y;
int size;
{
	int hx, hy;

	hx = SCREEN_TO_REGION( x, STV_REGION_MIN( W->region, 0 ),
		W->xoff, W->scale );

	hy = SCREEN_TO_REGION( y, STV_REGION_MIN( W->region, 1 ),
		W->yoff, W->scale );

	if ( hx >= STV_REGION_MIN( W->region, 0 ) - size
		&& hx <= STV_REGION_MAX( W->region, 0 ) + size
		&& hy >= STV_REGION_MIN( W->region, 1 ) - size
		&& hy <= STV_REGION_MAX( W->region, 1 ) + size )
	{
		return( stvTrue );
	}

	else
		return( stvFalse );
}


void
reset_heli_updated( W )
WIN W;
{
	HELI H;

	H = W->helis;

	while ( H != NULL )
	{
		H->updated = NOT_UPDATED;

		H = H->next;
	}
}


void
release_heli()
{
	int tmp;

	if ( CURRENT_HELI != NULL )
	{
		/* Release Helicopter */

		tmp = stvTrue;

		if ( stv_viewer_steer_vparameter( CURRENT_VP,
			(STV_VALUE) &tmp, stvInt, RELEASE_INDEX ) != stvStatusOk )
		{
			printf( "Error releasing helicopter steering...\n" );

			return;
		}

		stv_viewer_send_NewParams( VIEWER );

		/* Release Steering Token */

		stv_viewer_steering_release( CURRENT_VP, VIEWER );

		/* Reset Current Heli */

		CURRENT_HELI = (HELI) NULL;

		MAIN_WIN->xoff = MAIN_WIN->yoff = 2 * FRAME_BORDER;

		AUTO_ZOOM_FACTOR = 1.0;
	}
}


HELI add_heli( W, hid, steered, shield_str, shielded, fire_count, score,
	rx, ry, sx, sy, heading, vel_angle, velocity, radar_range )
WIN W;
int hid;
int steered;
int shield_str;
int shielded;
int fire_count;
int score;
int rx, ry;
int sx, sy;
double heading;
double vel_angle;
double velocity;
double radar_range;
{
	HELI H;

	if ( vflag )
	{
		printf(
		"adding heli hid=%d at (%d,%d)/(%d,%d) heading=%lf rr=%lf\n",
			hid, rx, ry, sx, sy, heading, radar_range );
		fflush( stdout );
	}

	if ( W->helis == NULL )
		H = W->helis = create_heli();
	
	else
	{
		H = W->helis;

		while ( H->next != NULL )
			H = H->next;

		H = H->next = create_heli();
	}

	H->updated = UPDATED;

	H->win = W;

	H->hid = hid;

	H->steered = steered;

	H->shield_str = shield_str;

	H->shielded = shielded;

	H->last_shield_str = shield_str;

	H->fire_count = fire_count;

	H->score = score;

	H->region_x = rx;
	H->region_y = ry;

	H->screen_x = sx;
	H->screen_y = sy;

	H->heading = heading;

	H->vel_angle = vel_angle;

	H->velocity = velocity;

	H->radar_range = radar_range;

	H->rotor_posn = (int) random() % ROTOR_CLICKS;

	draw_heli( H );

	draw_radar( H );

	return( H );
}


void
remove_missing_helis( W )
WIN W;
{
	HELI Hnext;
	HELI Hlast;
	HELI H;

	int whacked;

	H = W->helis;

	Hlast = (HELI) NULL;

	whacked = 0;

	while ( H != NULL )
	{
		/* Remove Lost Helis */

		if ( H->updated == NOT_UPDATED )
		{
			if ( vflag )
			{
				printf( "Removing Heli hid=%d\n", H->hid );
				fflush( stdout );
			}

			if ( IS_CURRENT( H ) )
				whacked++;

			undraw_heli( H );

			undraw_radar( H );

			Hnext = H->next;

			free_heli( &H );

			H = Hnext;

			if ( Hlast != NULL )
				Hlast->next = H;
			
			else
				W->helis = H;
		}

		else
		{
			Hlast = H;

			H = H->next;
		}
	}

	if ( whacked )
	{
		printf( "You Died, Baby!!\n" );

		SET_TCL_GLOBAL( interp, "play_mode", "Monitor" );

		Tcl_Eval( interp, "layout_main_panel" );

		Tcl_Eval( interp, "set_play_mode Monitor" );

		AUTO_ZOOM_FACTOR = 1.0;
	}
}


void
draw_heli( H )
HELI H;
{
	char *color;

	/* Verify Heli Visible */

	if ( item_in_display( H->win, H->screen_x, H->screen_y, 64 ) )
	{
		if ( IS_CURRENT( H ) && MY_SCORE > WINNING_SCORE )
			color = H->flash == 0 ? "green" : "purple";
		
		else
			color = "purple";

		/* Draw Rotors */

		if ( H->rotor_id == -1 )
		{
			H->rotor_id = CREATE_POLY( interp, H->win->canvas,
				H->screen_x, H->screen_y,
				"True", color, color, 2 );
		}

		DRAW_ROTORS( interp, H->win->canvas, H->rotor_id,
			H->screen_x, H->screen_y, H->rotor_posn, color, 2,
			H->win->scale );

		/* Draw Helicopter Body */

		if ( H->body_id == -1 )
		{
			H->body_id = CREATE_POLY( interp, H->win->canvas,
				H->screen_x, H->screen_y,
				"False", "", color, 2 );
		}

		DRAW_BODY( interp, H->win->canvas, H->body_id,
			H->screen_x, H->screen_y, H->heading, color, 2,
			H->win->scale );

		/* Draw Helicopter Tail */

		if ( H->tail_id == -1 )
		{
			H->tail_id = CREATE_POLY( interp, H->win->canvas,
				H->screen_x, H->screen_y,
				"False", "", color, 2 );
		}

		DRAW_TAIL( interp, H->win->canvas, H->tail_id,
			H->screen_x, H->screen_y, H->heading, color, 2,
			H->win->scale );

		/* Check for Hit, Beep Me */

		if ( !qflag
			&& H->win == MAIN_WIN
			&& H->last_shield_str != H->shield_str )
		{
			H->last_shield_str = H->shield_str;

			Tcl_Eval( interp, "bell" );
		}

		/* Draw Shield (if on) */

		if ( H->shielded && H->shield_str > 0 )
		{
			if ( H->shield_str > 6 )
				color = "green";
			
			else if ( H->shield_str > 3 )
				color = "orange";
			
			else
				color = "red";

			if ( H->shield_id == -1 )
			{
				H->shield_id = DRAW_CIRCLE( interp, H->win->canvas,
					H->screen_x, H->screen_y,
					HELI_RADIUS, "", color, 2, H->win->scale );
			}

			else
			{
				CIRCLE_COORDS( interp, H->win->canvas, H->shield_id,
					H->screen_x, H->screen_y,
					HELI_RADIUS, "", color, 2, H->win->scale );
			}
		}

		else
		{
			DELETE_ITEM( interp, H->win->canvas, H->shield_id );

			H->shield_id = -1;
		}
	}

	else
		undraw_heli( H );
}


void
undraw_heli( H )
HELI H;
{
	/* Undraw Rotors */

	if ( H->rotor_id != -1 )
	{
		DELETE_ITEM( interp, H->win->canvas, H->rotor_id );

		H->rotor_id = -1;
	}

	/* Undraw Helicopter Body */

	if ( H->body_id != -1 )
	{
		DELETE_ITEM( interp, H->win->canvas, H->body_id );

		H->body_id = -1;
	}

	/* Undraw Helicopter Tail */

	if ( H->tail_id != -1 )
	{
		DELETE_ITEM( interp, H->win->canvas, H->tail_id );

		H->tail_id = -1;
	}

	/* Undraw Helicopter Shield */

	if ( H->shield_id != -1 )
	{
		DELETE_ITEM( interp, H->win->canvas, H->shield_id );

		H->shield_id = -1;
	}
}


void
draw_rotors( H )
HELI H;
{
	char *color;

	/* Verify Heli Visible */

	if ( item_in_display( H->win, H->screen_x, H->screen_y, 64 ) )
	{
		if ( IS_CURRENT( H ) && MY_SCORE > WINNING_SCORE )
			color = H->flash == 0 ? "green" : "purple";
		
		else
			color = "purple";

		if ( H->rotor_id == -1 )
		{
			H->rotor_id = CREATE_POLY( interp, H->win->canvas,
				H->screen_x, H->screen_y,
				"True", color, color, 2 );
		}

		DRAW_ROTORS( interp, H->win->canvas, H->rotor_id,
			H->screen_x, H->screen_y, H->rotor_posn, color, 2,
			H->win->scale );
	}

	else
		undraw_heli( H );
}


void
undraw_rotors( H )
HELI H;
{
	if ( H->rotor_id != -1 )
	{
		DELETE_ITEM( interp, H->win->canvas, H->rotor_id );

		H->rotor_id = -1;
	}
}


void
draw_radar( H )
HELI H;
{
	char *color;

	if ( IS_CURRENT( H ) && MY_SCORE > WINNING_SCORE )
	{
		color = H->flash == 0 ? "green" : "yellow";

		H->flash = 1 - H->flash;
	}

	else
	{
		if ( IS_CURRENT( H ) )
			color = "purple";

		else if ( H->steered )
			color = "blue";

		else
			color = "yellow";
	}
	
	/* Draw Radar Range Circle */

	if ( H->radar_id == -1 )
	{
		H->radar_id = DRAW_CIRCLE( interp, H->win->canvas,
			H->screen_x, H->screen_y,
			H->radar_range, "", color, 2, H->win->scale );
	}

	else
	{
		CIRCLE_COORDS( interp, H->win->canvas, H->radar_id,
			H->screen_x, H->screen_y,
			H->radar_range, "", color, 2, H->win->scale );
	}
}


void
undraw_radar( H )
HELI H;
{
	/* Undraw Helicopter Radar Range Indicator */

	if ( H->radar_id != -1 )
	{
		DELETE_ITEM( interp, H->win->canvas, H->radar_id );

		H->radar_id = -1;
	}
}


void
reset_bullet_updated( W )
WIN W;
{
	BULLET B;

	B = W->bullets;

	while ( B != NULL )
	{
		B->updated = NOT_UPDATED;

		B = B->next;
	}
}


BULLET add_bullet( W, bid, rx, ry, sx, sy, vel_angle )
WIN W;
int bid;
int rx, ry;
int sx, sy;
double vel_angle;
{
	BULLET B;

	if ( vflag )
	{
		printf( "adding bullet bid=%d at (%d,%d)/(%d,%d)\n",
			bid, rx, ry, sx, sy );
		fflush( stdout );
	}

	if ( W->bullets == NULL )
		B = W->bullets = create_bullet();
	
	else
	{
		B = W->bullets;

		while ( B->next != NULL )
			B = B->next;

		B = B->next = create_bullet();
	}

	B->win = W;

	B->bid = bid;

	B->updated = UPDATED;

	B->region_x = rx;
	B->region_y = ry;

	B->screen_x = sx;
	B->screen_y = sy;

	B->vel_angle = vel_angle;

	draw_bullet( B );

	return( B );
}


void
remove_missing_bullets( W )
WIN W;
{
	BULLET Bnext;
	BULLET Blast;
	BULLET B;

	B = W->bullets;

	Blast = (BULLET) NULL;

	while ( B != NULL )
	{
		/* Remove Lost Bullets */

		if ( B->updated == NOT_UPDATED )
		{
			if ( vflag )
			{
				printf( "Removing Bullet bid=%d\n", B->bid );
				fflush( stdout );
			}

			undraw_bullet( B );

			Bnext = B->next;

			free_bullet( &B );

			B = Bnext;

			if ( Blast != NULL )
				Blast->next = B;
			
			else
				W->bullets = B;
		}

		else
		{
			Blast = B;

			B = B->next;
		}
	}
}


void
draw_bullet( B )
BULLET B;
{
	if ( item_in_display( B->win, B->screen_x, B->screen_y, 4 ) )
	{
		if ( B->bullet_id == -1 )
		{
			B->bullet_id = CREATE_POLY( interp, B->win->canvas,
				B->screen_x, B->screen_y, "True", "red", "red", 2 );
		}

		DRAW_BULLET( interp, B->win->canvas, B->bullet_id,
			B->screen_x, B->screen_y, B->vel_angle, B->win->scale );
	}

	else
		undraw_bullet( B );
}


void
undraw_bullet( B )
BULLET B;
{
	if ( B->bullet_id != -1 )
	{
		DELETE_ITEM( interp, B->win->canvas, B->bullet_id );

		B->bullet_id = -1;
	}
}


void
reset_explosion_updated( W )
WIN W;
{
	EXPLOSION E;

	E = W->explosions;

	while ( E != NULL )
	{
		E->updated = NOT_UPDATED;

		E = E->next;
	}
}


EXPLOSION add_explosion( W, eid, rx, ry, sx, sy,
	angle, strength, radius )
WIN W;
int eid;
int rx, ry;
int sx, sy;
double angle;
int strength;
int radius;
{
	EXPLOSION E;

	if ( vflag )
	{
		printf(
		"Explosion!!! => %d @ (%d,%d)/(%d,%d) -> angle=%lf radius=%d\n",
			strength, rx, ry, sx, sy, angle, radius );
		fflush( stdout );
	}

	E = create_explosion();

	if ( W->explosions == NULL )
		W->explosions = E;
	
	else
	{
		E->next = W->explosions;

		W->explosions = E;
	}

	E->win = W;

	E->eid = eid;

	E->region_x = rx;
	E->region_y = ry;

	E->screen_x = sx;
	E->screen_y = sy;

	E->angle = angle;

	E->strength = strength;

	E->radius = radius;

	draw_explosion( E );

	return( E );
}


void
remove_missing_explosions( W )
WIN W;
{
	EXPLOSION Enext;
	EXPLOSION Elast;
	EXPLOSION E;

	E = W->explosions;

	Elast = (EXPLOSION) NULL;

	while ( E != NULL )
	{
		/* Remove Lost Explosions */

		if ( E->updated == NOT_UPDATED )
		{
			if ( vflag )
			{
				printf( "Removing Explosion eid=%d\n", E->eid );
				fflush( stdout );
			}

			undraw_explosion( E );

			Enext = E->next;

			free_explosion( &E );

			E = Enext;

			if ( Elast != NULL )
				Elast->next = E;
			
			else
				W->explosions = E;
		}

		else
		{
			Elast = E;

			E = E->next;
		}
	}
}


void
draw_explosion( E )
EXPLOSION E;
{
	int numbell;
	int i;

	if ( item_in_display( E->win, E->screen_x, E->screen_y,
		E->radius ) )
	{
		if ( E->explosion_id == -1 )
		{
			if ( !qflag && E->win == MAIN_WIN )
			{
				numbell = E->strength / 10;

				numbell = numbell ? numbell : 1;

				for ( i=0 ; i < numbell ; i++ )
					Tcl_Eval( interp, "bell" );
			}

			E->explosion_id = CREATE_POLY( interp, E->win->canvas,
				E->screen_x, E->screen_y, "False", "red", "red", 2 );
		}

		DRAW_EXPLOSION( interp, E->win->canvas, E->explosion_id,
			E->screen_x, E->screen_y, E->angle, E->radius,
			E->win->scale );
	}

	else
		undraw_explosion( E );
}


void
undraw_explosion( E )
EXPLOSION E;
{
	if ( E->explosion_id != -1 )
	{
		DELETE_ITEM( interp, E->win->canvas, E->explosion_id );

		E->explosion_id = -1;
	}
}


void heli_handle( clientData )
ClientData clientData;
{
	static double last_head_goal = 0.0;

	static int cnt = 0;

	STV_VIEW_FIELD_GROUP VFG;

	char cmd[1024];

	double head_goal;

	double hdiff;
	double hmag;

	int get_frame;
	int dx, dy;
	int flag;
	int tmp;
	int cc;

	/* Attach to Region if Not Connected */

	if ( !CONNECTED )
		attach_to_region();

	/* Handle Primary Viewer Functionality (i.e. do stuff) */

	if ( CONNECTED )
	{
		/* Get Next Region Frame (Every FREQth Rotor Iter) */

		get_frame = FREQ / 2;

		get_frame = get_frame < 1 ? 1 : get_frame;

		if ( !(++cnt % get_frame) )
		{
			cc = stv_viewer_receive_frame( &VFG, VIEWER, &RESTART,
				stvMsgBlock );

			if ( cc == stvStatusOk || cc == stvStatusBadFrame )
			{
				flag = 0;

				/* Handle Long Range Radar */

				if ( VFG == RADAR_VFG )
				{
					stv_viewer_send_XON( VFG, VIEWER );

					if ( cc == stvStatusOk )
						handle_frame( VFG );

					stv_reset_Field_Particles( VFG );

					/* Grab Regular Frame */

					cc = stv_viewer_receive_frame( &VFG, VIEWER,
						&RESTART, stvMsgBlock );

					if ( cc != stvStatusOk && cc != stvStatusBadFrame )
					{
						stv_reset_Field_Particles( VFG );

						do_stv_cleanup();

						flag++;
					}
				}

				if ( !flag )
				{
					ZOOM_LIMIT = stvFalse;

					stv_viewer_send_XON( VFG, VIEWER );

					if ( cc == stvStatusOk )
						handle_frame( VFG );

					stv_reset_Field_Particles( VFG );
				}
			}

			else
			{
				stv_reset_Field_Particles( VFG );

				do_stv_cleanup();
			}
		}

		/* Handle Steered Heli */

		if ( CONNECTED && CURRENT_HELI != NULL )
		{
			/* Check for Auto Zoom */

			if ( AUTO_ZOOM )
				check_auto_zoom();

			/* Check for Vis Region Adjust */

			if ( CURRENT_HELI->updated == UPDATED )
			{
				if ( RADAR_CHANGED
					|| check_heli_migration( CURRENT_HELI ) )
				{
					set_display_to_heli( CURRENT_HELI, stvTrue );

					RADAR_CHANGED = stvFalse;
				}
				
				else
					center_display_to_heli( CURRENT_HELI );

				/* Update User Score Interface */

				MY_SCORE += CURRENT_HELI->score - BASE_SCORE;

				BASE_SCORE = CURRENT_HELI->score;

				sprintf( cmd,
					"%s.score_label configure -text \"Score:%10d\"",
					FM, MY_SCORE );

				Tcl_Eval( interp, cmd );

				/* Update Heli Score Interface */

				sprintf( cmd,
					"%s.score configure -text \"Score:%10d\"",
					FM_I, CURRENT_HELI->score );

				Tcl_Eval( interp, cmd );

				/* Update Heli Info Displays */

				update_info_displays( CURRENT_HELI );
			}

			/* Check Mouse Pointer for Heading Goal Adjustment */

			if ( STEER_ACTIVE || ( AUTO_STEER && PTR_VALID ) )
			{
				head_goal = angle_of(
					(double) CURRENT_HELI->screen_x,
					(double) CURRENT_HELI->screen_y,
					(double) PTR_X, (double) PTR_Y );

				if ( !EQUIV( head_goal, last_head_goal ) )
				{
					if ( stv_viewer_steer_vparameter( CURRENT_VP,
						(STV_VALUE) &head_goal, stvDouble,
						HEADING_GOAL_INDEX ) != stvStatusOk )
					{
						printf( "Error steering heli heading...\n" );
					}

					else
						last_head_goal = head_goal;
				}

				if ( STEER_RELEASE )
					STEER_ACTIVE = stvFalse;
			}

			/* Send New Steering Param Values */

			if ( CURRENT_VP->changed )
			{
				stv_viewer_send_NewParams( VIEWER );

				/* Reset Bullet Firing */

				tmp = stvFalse;

				if ( stv_viewer_steer_vparameter( CURRENT_VP,
					(STV_VALUE) &tmp, stvInt, FIRE_BULLET_INDEX )
					!= stvStatusOk )
				{
					printf( "Error resetting bullet firing...\n" );
				}

				else
					CURRENT_VP->changed = stvFalse;
			}

			/* Update Long Range Radar Window */

			update_win_stuff( RADAR_WIN );
		}

		/* Update Main Window */

		update_win_stuff( MAIN_WIN );

		/* Update Display */

		Tcl_Eval( interp, "update" );
	}

	/* Set Up Next Handler */

	Tk_DoWhenIdle( heli_handle, (ClientData) NULL );
}


void
update_win_stuff( W )
WIN W;
{
	EXPLOSION E;

	BULLET B;

	HELI H;

	/* Update Each Helicopter */

	H = W->helis;

	while ( H != NULL )
	{
		/* Spin Rotor */

		(H->rotor_posn)++;

		H->rotor_posn %= ROTOR_CLICKS;

		/* Draw It */

		if ( H->updated == UPDATED )
		{
			draw_heli( H );

			draw_radar( H );

			H->updated = NO_CHANGE;
		}

		else
			draw_rotors( H );

		H = H->next;
	}

	/* Update Each Bullet */

	B = W->bullets;

	while ( B != NULL )
	{
		if ( B->updated == UPDATED )
		{
			draw_bullet( B );

			B->updated = NO_CHANGE;
		}

		B = B->next;
	}

	/* Update Explosions */

	E = W->explosions;

	while ( E != NULL )
	{
		if ( E->updated == UPDATED )
		{
			draw_explosion( E );

			E->updated = NO_CHANGE;
		}

		E = E->next;
	}
}


/* Update Heli Info Displays */

void
update_info_displays( H )
HELI H;
{
	static int shield_ids[10] =
		{ -1, -1, -1, -1, -1, -1, -1, -1, -1, -1 };

	static int head_arrow = -1;
	static int vel_arrow = -1;
	static int fire_ball = -1;

	static int fire_wait = stvFalse;

	double scale;

	char cmd[1024];

	char *color;
	char *tmp;

	int flyme_ssize;
	int flyme_hsize;
	int shield_str;
	int fireck;
	int x1, x2;
	int x, y;
	int swt;
	int rh2;
	int i;

	/* Get Canvas Sizes */

	tmp = GET_TCL_GLOBAL( interp, "flyme_ssize" );
	flyme_ssize = atoi( tmp );

	tmp = GET_TCL_GLOBAL( interp, "flyme_hsize" );
	flyme_hsize = atoi( tmp );

	/* Update Fire Ready Display */

	if ( !fire_wait )
	{
		fireck = STV_VALUE_OF(
			CURRENT_VP->param->vars[ FIRE_BULLET_INDEX ],
			int );

		if ( fireck )
			fire_wait = stvTrue;
	}

	if ( H->fire_count > 2 || fire_wait )
		color = "red";

	else
		color = "green";

	if ( H->fire_count > 0 )
		fire_wait = stvFalse;

	if ( fire_ball == -1 )
	{
		rh2 = ROW_HEIGHT / 2;

		fire_ball = DRAW_CIRCLE( interp, FM_IF, rh2, rh2, rh2 / 2,
			 color, color, 1, 1.0 );
	}

	else
	{
		UPDATE_POLY( interp, FM_IF, fire_ball, color, color, 1 );
	}

	/* Update Shield Display */

	swt = flyme_ssize / 10;

	if ( H->shielded )
		shield_str = H->shield_str;
	
	else
		shield_str = 0;

	for ( i=0 ; i < shield_str ; i++ )
	{
		if ( i + 1 > 6 )
			color = "green";
		
		else if ( i + 1 > 3 )
			color = "orange";
		
		else
			color = "red";

		x1 = i * swt;

		x2 = x1 + swt - 1;

		if ( shield_ids[i] == -1 )
		{
			shield_ids[i] = DRAW_RECT( interp, FM_IS,
				x1, 0, x2, ROW_HEIGHT, color, color, 1 );
		}
	}

	for ( i=shield_str ; i < 10 ; i++ )
	{
		if ( shield_ids[i] != -1 )
		{
			DELETE_ITEM( interp, FM_IS, shield_ids[i] );

			shield_ids[i] = -1;
		}
	}

	/* Update Heading Display */

	x = y = flyme_hsize / 2;

	scale = (double) ( flyme_hsize - ( 2 * ROW_HEIGHT ) ) / 100.0;

	if ( head_arrow == -1 )
	{
		head_arrow = CREATE_POLY( interp, FM_IH, x, y, "False",
			"", "purple", 2 );
	}

	DRAW_ARROW( interp, FM_IH, head_arrow, x, y, H->heading, scale );

	scale *= 0.666;

	if ( vel_arrow == -1 )
	{
		vel_arrow = CREATE_POLY( interp, FM_IH, x, y, "False",
			"", "orange", 1 );
	}

	DRAW_ARROW( interp, FM_IH, vel_arrow, x, y, H->vel_angle, scale );

	/* Update Speed Display */

	sprintf( cmd, "%s.speed configure -text \"Speed:%3d\"",
		FM_I, (int) H->velocity );

	Tcl_Eval( interp, cmd );
}


/* Handle New Helicopter Data */

void
handle_frame( VFG )
STV_VIEW_FIELD_GROUP VFG;
{
	STV_VIEW_PARTICLE_DATA VPD;

	STV_VIEW_PARTICLE VPT;

	EXPLOSION E;

	BULLET B;

	HELI H;

	int ex_strength_index;
	int ex_radius_index;
	int ex_angle_index;

	int radar_range_index;
	int vel_angle_index;
	int velocity_index;
	int heading_index;

	int fire_count_index;
	int shield_str_index;
	int shielded_index;
	int steered_index;
	int score_index;

	int hid_index;
	int bid_index;
	int eid_index;

	double radar_range;
	double vel_angle;
	double velocity;
	double heading;
	double angle;

	int fire_count;
	int shield_str;
	int shielded;
	int strength;
	int steered;
	int radius;
	int rx, ry;
	int sx, sy;
	int score;

	int hid;
	int bid;
	int eid;

	if ( VFG == CURRENT_VFG )
	{
		/*** Get Helicopter Particles ***/
	
		VPT = stv_get_view_particle_name( VFG->particle_list,
			"helicopter" );
	
		if ( VPT != NULL )
		{
			/* Get Indices */
	
			hid_index = get_particle_field_index( VPT, "hid" );
	
			steered_index = get_particle_field_index( VPT, "steered" );
	
			fire_count_index = get_particle_field_index( VPT,
				"fire_count" );
	
			shield_str_index = get_particle_field_index( VPT,
				"shield_str" );
	
			shielded_index = get_particle_field_index( VPT,
				"shielded" );
	
			score_index = get_particle_field_index( VPT, "score" );
	
			heading_index = get_particle_field_index( VPT, "heading" );
	
			vel_angle_index = get_particle_field_index( VPT,
				"hvel_angle" );
	
			velocity_index = get_particle_field_index( VPT,
				"hvel_magn" );
	
			radar_range_index = get_particle_field_index( VPT,
				"radar_range" );
	
			if ( hid_index != -1 && steered_index != -1
				&& shield_str_index != -1 && shielded_index != -1
				&& fire_count_index != -1 && score_index != -1
				&& heading_index != -1 && vel_angle_index != -1
				&& velocity_index != -1 && radar_range_index != -1 )
			{
				/* Reset Heli Updated Flags */
	
				reset_heli_updated( MAIN_WIN );
	
				/* Get Data */
	
				VPD = VPT->pdata;
	
				while ( VPD != NULL )
				{
					/* Get Heli Values */
	
					rx = VPD->coords[0];
					ry = VPD->coords[1];
	
					sx = REGION_TO_SCREEN( rx,
						STV_REGION_MIN( MAIN_WIN->region, 0 ),
						MAIN_WIN->xoff, MAIN_WIN->scale );
	
					sy = REGION_TO_SCREEN( ry,
						STV_REGION_MIN( MAIN_WIN->region, 1 ),
						MAIN_WIN->yoff, MAIN_WIN->scale );
	
					hid = STV_VALUE_OF( VPD->field_values[ hid_index ],
						int );
	
					steered = STV_VALUE_OF(
						VPD->field_values[ steered_index ], int );
	
					shield_str = STV_VALUE_OF(
						VPD->field_values[ shield_str_index ], int );
	
					shielded = STV_VALUE_OF(
						VPD->field_values[ shielded_index ], int );
	
					fire_count = STV_VALUE_OF(
						VPD->field_values[ fire_count_index ], int );
	
					score = STV_VALUE_OF(
						VPD->field_values[ score_index ], int );
	
					heading = STV_VALUE_OF(
						VPD->field_values[ heading_index ], double );
	
					vel_angle = STV_VALUE_OF(
						VPD->field_values[ vel_angle_index ], double );
	
					velocity = STV_VALUE_OF(
						VPD->field_values[ velocity_index ], double );
	
					radar_range = STV_VALUE_OF(
						VPD->field_values[ radar_range_index ],
						double );
	
					/* Look for Heli */
	
					H = get_heli_hid( MAIN_WIN, hid );
	
					/* Add New Heli */
	
					if ( H == NULL )
					{
						H = add_heli( MAIN_WIN, hid, steered,
							shield_str, shielded, fire_count, score,
							rx, ry, sx, sy, heading,
							vel_angle, velocity, radar_range );
					}
	
					/* Adjust Existing Heli */
	
					else
					{
						if ( IS_CURRENT( H )
							&& !EQUIV( H->radar_range, radar_range ) )
						{
							RADAR_CHANGED = stvTrue;
						}
	
						if ( H->region_x != rx || H->region_y != ry
							|| H->steered != steered
							|| H->shield_str != shield_str
							|| H->shielded != shielded
							|| H->fire_count != fire_count
							|| H->score != score
							|| !EQUIV( H->heading, heading )
							|| !EQUIV( H->vel_angle, vel_angle )
							|| !EQUIV( H->velocity, velocity )
							|| !EQUIV( H->radar_range, radar_range ) )
						{
							H->steered = steered;

							H->shield_str = shield_str;
							H->shielded = shielded;
	
							H->fire_count = fire_count;
	
							H->score = score;
	
							H->region_x = rx;
							H->region_y = ry;
	
							H->screen_x = sx;
							H->screen_y = sy;
	
							H->heading = heading;
	
							H->vel_angle = vel_angle;
	
							H->velocity = velocity;
	
							H->radar_range = radar_range;
	
							H->updated = UPDATED;
						}
	
						else
							H->updated = NO_CHANGE;
					}
	
					VPD = VPD->next;
				}
	
				/* Remove Missing Helis */
	
				remove_missing_helis( MAIN_WIN );
			}
		}
	
		/*** Get Bullet Particles ***/
	
		VPT = stv_get_view_particle_name( VFG->particle_list,
			"bullet" );
	
		if ( VPT != NULL )
		{
			/* Get Index */
	
			bid_index = get_particle_field_index( VPT, "bid" );
	
			vel_angle_index = get_particle_field_index( VPT,
				"bvel_angle" );
	
			if ( bid_index != -1 && vel_angle_index != -1 )
			{
				/* Reset Bullet Updated Flags */
	
				reset_bullet_updated( MAIN_WIN );
	
				/* Get Data */
	
				VPD = VPT->pdata;
	
				while ( VPD != NULL )
				{
					/* Get Bullet Values */
	
					rx = VPD->coords[0];
					ry = VPD->coords[1];
	
					sx = REGION_TO_SCREEN( rx,
						STV_REGION_MIN( MAIN_WIN->region, 0 ),
						MAIN_WIN->xoff, MAIN_WIN->scale );
	
					sy = REGION_TO_SCREEN( ry,
						STV_REGION_MIN( MAIN_WIN->region, 1 ),
						MAIN_WIN->yoff, MAIN_WIN->scale );
	
					bid = STV_VALUE_OF( VPD->field_values[ bid_index ],
						int );
	
					vel_angle = STV_VALUE_OF(
						VPD->field_values[ vel_angle_index ], double );
	
					/* Look for Bullet */
	
					B = get_bullet_bid( MAIN_WIN, bid );
	
					/* Add New Bullet */
	
					if ( B == NULL )
					{
						B = add_bullet( MAIN_WIN, bid, rx, ry, sx, sy,
							vel_angle );
					}
	
					/* Adjust Existing Bullet */
	
					else
					{
						if ( B->region_x != rx || B->region_y != ry
							|| !EQUIV( B->vel_angle, vel_angle ) )
						{
							B->region_x = rx;
							B->region_y = ry;
	
							B->screen_x = sx;
							B->screen_y = sy;
	
							B->vel_angle = vel_angle;
	
							B->updated = UPDATED;
						}
	
						else
							B->updated = NO_CHANGE;
					}
	
					VPD = VPD->next;
				}
	
				/* Remove Missing Bullets */
	
				remove_missing_bullets( MAIN_WIN );
			}
		}
	
		/*** Get Explosion Particles ***/
	
		VPT = stv_get_view_particle_name( VFG->particle_list,
			"explosion" );
	
		if ( VPT != NULL )
		{
			/* Get Index */
	
			eid_index = get_particle_field_index( VPT, "eid" );
	
			ex_angle_index = get_particle_field_index( VPT,
				"explosion_angle" );
	
			ex_strength_index = get_particle_field_index( VPT,
				"explosion_strength" );
	
			ex_radius_index = get_particle_field_index( VPT,
				"explosion_radius" );
	
			if ( eid_index != -1 && ex_angle_index != -1
				&& ex_strength_index != -1 && ex_radius_index != -1 )
			{
				/* Reset Explosion Updated Flags */
	
				reset_explosion_updated( MAIN_WIN );
	
				/* Get Data */
	
				VPD = VPT->pdata;
	
				while ( VPD != NULL )
				{
					/* Get Explosion Values */
	
					rx = VPD->coords[0];
					ry = VPD->coords[1];
	
					sx = REGION_TO_SCREEN( rx,
						STV_REGION_MIN( MAIN_WIN->region, 0 ),
						MAIN_WIN->xoff, MAIN_WIN->scale );
	
					sy = REGION_TO_SCREEN( ry,
						STV_REGION_MIN( MAIN_WIN->region, 1 ),
						MAIN_WIN->yoff, MAIN_WIN->scale );
	
					eid = STV_VALUE_OF( VPD->field_values[ eid_index ],
						int );
	
					angle = STV_VALUE_OF(
						VPD->field_values[ ex_angle_index ], double );
	
					strength = STV_VALUE_OF(
						VPD->field_values[ ex_strength_index ], int );
	
					radius = STV_VALUE_OF(
						VPD->field_values[ ex_radius_index ], int );
	
					/* Look for Explosion */
	
					E = get_explosion_eid( MAIN_WIN, eid );
	
					/* Add New Explosion */
	
					if ( E == NULL )
					{
						E = add_explosion( MAIN_WIN, eid,
							rx, ry, sx, sy,
							angle, strength, radius );
					}
	
					/* Adjust Existing Explosion */
	
					else
					{
						if ( E->region_x != rx || E->region_y != ry
							|| !EQUIV( E->angle, angle )
							|| E->strength != strength
							|| E->radius != radius )
						{
							E->region_x = rx;
							E->region_y = ry;
	
							E->screen_x = sx;
							E->screen_y = sy;
	
							E->angle = angle;
	
							E->strength = strength;
	
							E->radius = radius;
	
							E->updated = UPDATED;
						}
	
						else
							E->updated = NO_CHANGE;
					}
	
					VPD = VPD->next;
				}
	
				/* Remove Missing Explosions */
	
				remove_missing_explosions( MAIN_WIN );
			}
		}
	}

	else if ( VFG == RADAR_VFG )
	{
		/*** Get Helicopter Particles ***/
	
		VPT = stv_get_view_particle_name( VFG->particle_list,
			"helicopter" );
	
		if ( VPT != NULL )
		{
			/* Get Indices */
	
			hid_index = get_particle_field_index( VPT, "hid" );
	
			steered_index = get_particle_field_index( VPT, "steered" );
	
			shield_str_index = get_particle_field_index( VPT,
				"shield_str" );
	
			shielded_index = get_particle_field_index( VPT,
				"shielded" );
	
			score_index = get_particle_field_index( VPT, "score" );
	
			heading_index = get_particle_field_index( VPT, "heading" );
	
			radar_range_index = get_particle_field_index( VPT,
				"radar_range" );
	
			if ( hid_index != -1 && steered_index != -1
				&& shield_str_index != -1 && shielded_index != -1
				&& score_index != -1 && heading_index != -1
				&& radar_range_index != -1 )
			{
				/* Reset Heli Updated Flags */
	
				reset_heli_updated( RADAR_WIN );
	
				/* Get Data */
	
				VPD = VPT->pdata;
	
				while ( VPD != NULL )
				{
					/* Get Heli Values */
	
					rx = VPD->coords[0];
					ry = VPD->coords[1];
	
					sx = REGION_TO_SCREEN( rx,
						STV_REGION_MIN( RADAR_WIN->region, 0 ),
						RADAR_WIN->xoff, RADAR_WIN->scale );
	
					sy = REGION_TO_SCREEN( ry,
						STV_REGION_MIN( RADAR_WIN->region, 1 ),
						RADAR_WIN->yoff, RADAR_WIN->scale );
	
					hid = STV_VALUE_OF( VPD->field_values[ hid_index ],
						int );
	
					steered = STV_VALUE_OF(
						VPD->field_values[ steered_index ], int );
	
					shield_str = STV_VALUE_OF(
						VPD->field_values[ shield_str_index ], int );
	
					shielded = STV_VALUE_OF(
						VPD->field_values[ shielded_index ], int );
	
					score = STV_VALUE_OF(
						VPD->field_values[ score_index ], int );
	
					heading = STV_VALUE_OF(
						VPD->field_values[ heading_index ], double );
	
					radar_range = STV_VALUE_OF(
						VPD->field_values[ radar_range_index ],
						double );
	
					/* Look for Heli */
	
					H = get_heli_hid( RADAR_WIN, hid );
	
					/* Add New Heli */
	
					if ( H == NULL )
					{
						H = add_heli( RADAR_WIN, hid, steered,
							shield_str, shielded, 0, score,
							rx, ry, sx, sy, heading,
							-1.0, -1.0, radar_range );
					}
	
					/* Adjust Existing Heli */
	
					else
					{
						if ( H->region_x != rx || H->region_y != ry
							|| H->steered != steered
							|| H->shield_str != shield_str
							|| H->shielded != shielded
							|| H->score != score
							|| !EQUIV( H->heading, heading )
							|| !EQUIV( H->radar_range, radar_range ) )
						{
							H->steered = steered;

							H->shield_str = shield_str;
							H->shielded = shielded;
	
							H->score = score;
	
							H->region_x = rx;
							H->region_y = ry;
	
							H->screen_x = sx;
							H->screen_y = sy;
	
							H->heading = heading;
	
							H->radar_range = radar_range;
	
							H->updated = UPDATED;
						}
	
						else
							H->updated = NO_CHANGE;
					}
	
					VPD = VPD->next;
				}
	
				/* Remove Missing Helis */
	
				remove_missing_helis( RADAR_WIN );
			}
		}
	
		/*** Get Explosion Particles ***/
	
		VPT = stv_get_view_particle_name( VFG->particle_list,
			"explosion" );
	
		if ( VPT != NULL )
		{
			/* Get Index */
	
			eid_index = get_particle_field_index( VPT, "eid" );
	
			ex_angle_index = get_particle_field_index( VPT,
				"explosion_angle" );
	
			ex_radius_index = get_particle_field_index( VPT,
				"explosion_radius" );
	
			if ( eid_index != -1 && ex_angle_index != -1
				&& ex_radius_index != -1 )
			{
				/* Reset Explosion Updated Flags */
	
				reset_explosion_updated( RADAR_WIN );
	
				/* Get Data */
	
				VPD = VPT->pdata;
	
				while ( VPD != NULL )
				{
					/* Get Explosion Values */
	
					rx = VPD->coords[0];
					ry = VPD->coords[1];
	
					sx = REGION_TO_SCREEN( rx,
						STV_REGION_MIN( RADAR_WIN->region, 0 ),
						RADAR_WIN->xoff, RADAR_WIN->scale );
	
					sy = REGION_TO_SCREEN( ry,
						STV_REGION_MIN( RADAR_WIN->region, 1 ),
						RADAR_WIN->yoff, RADAR_WIN->scale );
	
					eid = STV_VALUE_OF( VPD->field_values[ eid_index ],
						int );
	
					angle = STV_VALUE_OF(
						VPD->field_values[ ex_angle_index ], double );
	
					radius = STV_VALUE_OF(
						VPD->field_values[ ex_radius_index ], int );
	
					/* Look for Explosion */
	
					E = get_explosion_eid( RADAR_WIN, eid );
	
					/* Add New Explosion */
	
					if ( E == NULL )
					{
						E = add_explosion( RADAR_WIN, eid,
							rx, ry, sx, sy,
							angle, -1, radius );
					}
	
					/* Adjust Existing Explosion */
	
					else
					{
						if ( E->region_x != rx || E->region_y != ry
							|| !EQUIV( E->angle, angle )
							|| E->radius != radius )
						{
							E->region_x = rx;
							E->region_y = ry;
	
							E->screen_x = sx;
							E->screen_y = sy;
	
							E->angle = angle;
	
							E->radius = radius;
	
							E->updated = UPDATED;
						}
	
						else
							E->updated = NO_CHANGE;
					}
	
					VPD = VPD->next;
				}
	
				/* Remove Missing Explosions */
	
				remove_missing_explosions( RADAR_WIN );
			}
		}
	}
}


get_particle_field_index( VPT, name )
STV_VIEW_PARTICLE VPT;
char *name;
{
	int i;

	for ( i=0 ; i < VPT->nvfields ; i++ )
	{
		if ( !strcmp( (VPT->view_field_list[i])->field->name, name ) )
			return( i );
	}

	return( -1 );
}

