
/* $Id: commands.c,v 1.60 2000/06/12 18:01:02 kohl Exp $ */


/*
 *                      CUMULVS version 1.0:
 *      Computational Steering and Interactive Visualization
 *                   in Distributed Applications
 *           Oak Ridge National Laboratory, Oak Ridge TN.
 *      Authors:  James Arthur Kohl and Philip M. Papadopoulos
 *                   (C) 1996 All Rights Reserved
 *
 *                              NOTICE
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted
 * provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear
 * in supporting documentation.
 *
 * Neither the Institution, Oak Ridge National Laboratory, nor the
 * Authors make any representations about the suitability of this
 * software for any purpose.  This software is provided ``as is''
 * without express or implied warranty.
 *
 * CUMULVS was funded by the U.S. Department of Energy.
 */


#include "slicer.h"
#include "slicer_glob.h"


/* ARGSUSED */
int set_app_name_proc( clientData, itp, argc, argv )
ClientData clientData;
Tcl_Interp *itp;
int argc;
char **argv;
{
	if ( argc != 2 )
	{
		Tcl_SetResult( itp, "usage: set_app_name <name>",
			TCL_STATIC );

		return( TCL_ERROR );
	}

	if ( SLICER_APP == NULL || strcmp( argv[1], SLICER_APP ) )
	{
		if ( SLICER_APP != NULL )
			FREE( SLICER_APP );

		if ( strcmp( argv[1], "" ) )
			SLICER_APP = stv_copy_str( argv[1] );

		else
			SLICER_APP = (char *) NULL;

		if ( SLICER_FIELD != NULL )
			FREE( SLICER_FIELD );
		
		SLICER_FIELD = (char *) NULL;

		if ( SLICER_RADIUS_FIELD != NULL )
			FREE( SLICER_RADIUS_FIELD );
		
		SLICER_RADIUS_FIELD = (char *) NULL;

		if ( SLICER_COLOR_FIELD != NULL )
			FREE( SLICER_COLOR_FIELD );
		
		SLICER_COLOR_FIELD = (char *) NULL;

		INITIALIZED = stvFalse;

		if ( RESTART )
			RESTART = stvFalse;

		REALLOC = stvTrue;

		attach_to_app();
	}

	return( TCL_OK );
}


/* ARGSUSED */
int get_app_name_proc( clientData, itp, argc, argv )
ClientData clientData;
Tcl_Interp *itp;
int argc;
char **argv;
{
	Tcl_SetResult( itp, stv_copy_str( SLICER_APP ? SLICER_APP : "" ),
		TCL_DYNAMIC );

	return( TCL_OK );
}


/* ARGSUSED */
int get_field_list_proc( clientData, itp, argc, argv )
ClientData clientData;
Tcl_Interp *itp;
int argc;
char **argv;
{
	STV_VIEW_FIELD VF;

	char *field_lists[2048];
	char *field_data[4];
	char *field_info;

	char tmp[255];

	int flag;
	int num;

	VF = VIEWER->vfields;

	num = 0;

	while ( VF != NULL )
	{
		flag = 0;

		if ( VF->field->decomp != NULL )
		{
			field_data[0] = stv_copy_str( VF->field->name );

			field_data[1] = stv_copy_str( "command" );

			sprintf( tmp, "Select \"%s\" as Data Field for Request",
				VF->field->name );

			field_data[2] = stv_copy_str( "none" );

			field_data[3] = stv_copy_str( tmp );

			flag++;
		}

		else if ( VF->field->particle != NULL )
		{
			field_data[0] = stv_copy_str( VF->field->name );

			field_data[1] = stv_copy_str( "particle" );

			sprintf( tmp,
				"Select \"%s\" Data Field for Particle Radius / Color",
				VF->field->name );

			field_data[2] = stv_copy_str( VF->field->name );

			field_data[3] = stv_copy_str( tmp );

			flag++;
		}

		if ( flag )
		{
			field_lists[num++] = Tcl_Merge( 4, field_data );

			FREE( field_data[0] );
			FREE( field_data[1] );
			FREE( field_data[2] );
			FREE( field_data[3] );
		}

		VF = VF->next;
	}

	field_info = Tcl_Merge( num, field_lists );

	Tcl_SetResult( itp, field_info, TCL_DYNAMIC );

	return( TCL_OK );
}


/* ARGSUSED */
int set_field_name_proc( clientData, itp, argc, argv )
ClientData clientData;
Tcl_Interp *itp;
int argc;
char **argv;
{
	char tmp[1024];

	char *name[2];
	int type[2];

	int i;

	/* Get Args */

	if ( argc < 2 || argc > 5 )
	{
		Tcl_SetResult( itp,
		"usage: set_field_name <name> [ radius ] [ <name2> [ color ] ]",
			TCL_STATIC );

		return( TCL_ERROR );
	}

	for ( i=0 ; i < 2 ; i++ )
	{
		name[i] = (char *) NULL;
		type[i] = 0;
	}

	for ( i=1 ; i < argc ; i += 2 )
	{
		name[ i / 2 ] = argv[i];

		if ( i + 1 < argc )
		{
			if ( !strcmp( argv[ i + 1 ], "radius" ) )
				type[ i / 2 ] = FOR_RADIUS;

			else if ( !strcmp( argv[ i + 1 ], "color" ) )
				type[ i / 2 ] = FOR_COLOR;
		}
	}

	/* Sanity Check & Merge for Particles */

	if ( name[1] != NULL )
	{
		if ( !strcmp( name[0], name[1] ) )
		{
			type[0] |= type[1];

			name[1] = (char *) NULL;
		}

		else
		{
			/* No Particle Usage Type, Second Field - Ignore */

			if ( !(type[1]) )
				name[1] = (char *) NULL;

			/* No Particle Usage Type, First Field - Use Second */
		
			else if ( !(type[0]) )
			{
				name[0] = name[1];
				type[0] = type[1];

				name[1] = (char *) NULL;
			}
		}
	}

	/* Set Slicer Fields */

	if ( strcmp( name[0], "" ) )
	{
		/* Particle Fields */

		if ( type[0] )
		{
			for ( i=0 ; i < 2 && name[i] ; i++ )
			{
				if ( ( type[i] & FOR_RADIUS )
					&& ( SLICER_RADIUS_FIELD == NULL
						|| strcmp( name[i], SLICER_RADIUS_FIELD ) ) )
				{
					if ( SLICER_RADIUS_FIELD != NULL )
						FREE( SLICER_RADIUS_FIELD );

					SLICER_RADIUS_FIELD = stv_copy_str( name[i] );
				}

				if ( ( type[i] & FOR_COLOR )
					&& ( SLICER_COLOR_FIELD == NULL
						|| strcmp( name[i], SLICER_COLOR_FIELD ) ) )
				{
					if ( SLICER_COLOR_FIELD != NULL )
						FREE( SLICER_COLOR_FIELD );

					SLICER_COLOR_FIELD = stv_copy_str( name[i] );
				}
			}

			if ( SLICER_RADIUS_FIELD != NULL
				&& SLICER_COLOR_FIELD != NULL )
			{
				if ( strcmp( SLICER_RADIUS_FIELD, SLICER_COLOR_FIELD ) )
				{
					sprintf( tmp,
						"setMsg {%s Radius=\"%s\" Color=\"%s\".}",
						"Particle Fields Selected",
						SLICER_RADIUS_FIELD, SLICER_COLOR_FIELD );
				}

				else
				{
					sprintf( tmp,
						"setMsg {%s Radius/Color=\"%s\".}",
						"Particle Field Selected",
						SLICER_RADIUS_FIELD );
				}
			}

			else
			{
				sprintf( tmp,
					"setMsg {Particle Field \"%s\" Selected (%s).}",
					( SLICER_RADIUS_FIELD != NULL )
						? SLICER_RADIUS_FIELD : SLICER_COLOR_FIELD,
					( SLICER_RADIUS_FIELD != NULL )
						? "Radius" : "Color" );
			}

			safe_eval( itp, tmp );

			attach_to_app();
		}

		/* Contiguous Fields */

		else if ( SLICER_FIELD == NULL
			|| strcmp( name[0], SLICER_FIELD ) )
		{
			if ( SLICER_FIELD != NULL )
				FREE( SLICER_FIELD );

			SLICER_FIELD = stv_copy_str( name[0] );

			sprintf( tmp,
				"setMsg {Field \"%s\" Selected.}", SLICER_FIELD );

			safe_eval( itp, tmp );

			attach_to_app();
		}
	}

	return( TCL_OK );
}


/* ARGSUSED */
int set_plane_choice_proc( clientData, itp, argc, argv )
ClientData clientData;
Tcl_Interp *itp;
int argc;
char **argv;
{
	if ( argc != 2 )
	{
		Tcl_SetResult( itp, "usage: set_plane_choice < XY | YZ | XZ >",
			TCL_STATIC );

		return( TCL_ERROR );
	}

	if ( !strcmp( argv[1], "XY" ) )
		PLANE_CHOICE = PLANE_XY;

	else if ( !strcmp( argv[1], "YZ" ) )
		PLANE_CHOICE = PLANE_YZ;

	else if ( !strcmp( argv[1], "XZ" ) )
		PLANE_CHOICE = PLANE_XZ;

	else
	{
		Tcl_SetResult( itp, "usage: set_plane_choice < XY | YZ | XZ >",
			TCL_STATIC );

		return( TCL_ERROR );
	}

	return( TCL_OK );
}


/* ARGSUSED */
int set_color_range_proc( clientData, itp, argc, argv )
ClientData clientData;
Tcl_Interp *itp;
int argc;
char **argv;
{
	CIRCLE C;

	RECT R;

	char color[16];

	char *minstr;
	char *maxstr;

	float min, max;

	int i;

	if ( argc != 3 )
	{
		Tcl_SetResult( itp, "usage: set_color_range <min> <max>",
			TCL_STATIC );

		return( TCL_ERROR );
	}

	minstr = argv[1];
	maxstr = argv[2];

	min = atof( minstr );
	max = atof( maxstr );

	if ( min != COLOR_MIN_VALUE || max != COLOR_MAX_VALUE )
	{
		safe_eval( itp, "setMsg {Redrawing for New Color Range...}" );

		COLOR_MIN_VALUE = min;
		COLOR_MAX_VALUE = max;

		if ( COLOR_MAX_VALUE < COLOR_MIN_VALUE )
			COLOR_MAX_VALUE = COLOR_MIN_VALUE;

		R = &(RECTIDS[0]);

		for ( i=0 ; i < NRECTIDS ; i++ )
		{
			if ( R->id != -1 )
			{
				color_value( color, R->value );

				if ( strcmp( R->color, color ) )
				{
					strcpy( R->color, color );

					RECT_COLOR( itp, SL_C, R->id, color, color );
				}
			}

			R++;
		}

		C = &(CIRCLEIDS[0]);

		for ( i=0 ; i < NCIRCLEIDS ; i++ )
		{
			if ( C->id != -1 )
			{
				color_value( color, C->value );

				if ( strcmp( C->color, color ) )
				{
					strcpy( C->color, color );

					CIRCLE_COLOR( itp, SL_C, C->id, color, color );
				}
			}

			C++;
		}

		update_color_range();

		safe_eval( itp, "setMsg {New Color Range Redraw Complete.}" );

		Tcl_SetResult( itp, "1", TCL_STATIC );
	}

	else
		Tcl_SetResult( itp, "0", TCL_STATIC );

	return( TCL_OK );
}


void
update_color_range()
{
	UPDATE_FENTRY( "color_min", COLOR_MIN_VALUE );
	UPDATE_FENTRY( "color_max", COLOR_MAX_VALUE );
}


/* ARGSUSED */
int draw_color_map_proc( clientData, itp, argc, argv )
ClientData clientData;
Tcl_Interp *itp;
int argc;
char **argv;
{
	static int equiv_save = -1;
	static int csize_save = -1;

	char color[16];

	char *color_size;
	char *sl_m;

	float cstart;
	float cstep;
	float c;

	int *newids;

	int old_ncolorids;
	int csize;
	int equiv;
	int i;

	/* Manually Get Color Map Canvas, In Case Startup */

	sl_m = GET_TCL_GLOBAL( itp, "SL_M" );

	/* Get New Map Size */

	color_size = GET_TCL_GLOBAL( itp, "color_size" );

	csize = atoi( color_size );

	if ( !csize )
	{
		Tcl_SetResult( itp, "Error:  Color Map Width of Zero",
			TCL_STATIC );

		return( TCL_ERROR );
	}

	if ( csize > NCOLORIDS || COLORIDS == NULL )
	{
		newids = (int *) MALLOC( (STV_SIZE_T) csize * sizeof(int) );
		memcheck( newids, "Color Map ID Array" );

		if ( COLORIDS != NULL )
		{
			for ( i=0 ; i < NCOLORIDS ; i++ )
			{
				newids[i] = COLORIDS[i];

				COLORIDS[i] = -1;
			}

			for ( i=NCOLORIDS ; i < csize ; i++ )
				newids[i] = -1;

			FREE( COLORIDS );
		}

		else
		{
			for ( i=0 ; i < csize ; i++ )
				newids[i] = -1;
		}

		COLORIDS = newids;

		NCOLORIDS = csize;
	}

	/* Verify Map Needs Redraw */

	if ( EQUIV( COLOR_MIN_VALUE, COLOR_MAX_VALUE ) )
	{
		cstart = COLOR_MIN_VALUE - ( (float) csize / 2.0 );
		cstep = 1.0;

		equiv = 1;
	}

	else
	{
		cstart = COLOR_MIN_VALUE;
		cstep = ( COLOR_MAX_VALUE - COLOR_MIN_VALUE )
			 / (float) csize;

		equiv = 0;
	}

	if ( csize == csize_save && equiv == equiv_save )
		return( TCL_OK );

	else
	{
		csize_save = csize;
		equiv_save = equiv;
	}

	/* Draw New Map */

	c = cstart;

	for ( i=0 ; i < csize ; i++ )
	{
		color_value( color, c );

		if ( COLORIDS[i] != -1 )
		{
			RECT_COORDS( itp, sl_m, COLORIDS[i], i, 0, i + 1, 100 );

			RECT_COLOR( itp, sl_m, COLORIDS[i], color, color );
		}

		else
		{
			COLORIDS[i] = DRAW_RECT( itp, sl_m, i, 0, i + 1, 100,
				color, color, 1 );
		}

		c += cstep;
	}

	/* Free Any Leftover Color Rects */

	for ( i=csize ; i < NCOLORIDS ; i++ )
	{
		if ( COLORIDS[i] != -1 )
		{
			DELETE_ITEM( itp, sl_m, COLORIDS[i] );

			COLORIDS[i] = -1;
		}
	}

	return( TCL_OK );
}


/* ARGSUSED */
int set_vis_region_proc( clientData, itp, argc, argv )
ClientData clientData;
Tcl_Interp *itp;
int argc;
char **argv;
{
	int xmin, xmax, xcell;
	int ymin, ymax, ycell;
	int xi, yi, zi;
	int zval;

	if ( argc != 8 )
	{
		Tcl_SetResult( itp,
		"usage: set_vis_region xmin xmax xcell ymin ymax ycell zval",
			TCL_STATIC );

		return( TCL_ERROR );
	}

	xmin =	atoi( argv[1] );
	xmax =	atoi( argv[2] );
	xcell =	atoi( argv[3] );

	ymin =	atoi( argv[4] );
	ymax =	atoi( argv[5] );
	ycell =	atoi( argv[6] );

	zval =	atoi( argv[7] );

	if ( !get_plane_indices( &xi, &yi, &zi ) )
	{
		Tcl_SetResult( itp,
			"Error:  Invalid Plane Selection in set_vis_region{}",
			TCL_STATIC );

		return( TCL_ERROR );
	}

	if ( STV_REGION_MIN( VISREGION, xi ) != xmin
		|| STV_REGION_MAX( VISREGION, xi ) != xmax
		|| STV_CELL_OF_DIM( VISCELL, xi ) != xcell
		|| STV_REGION_MIN( VISREGION, yi ) != ymin
		|| STV_REGION_MAX( VISREGION, yi ) != ymax
		|| STV_CELL_OF_DIM( VISCELL, yi ) != ycell
		|| STV_REGION_MIN( VISREGION, zi ) != zval
		|| STV_REGION_MAX( VISREGION, zi ) != zval
		|| STV_CELL_OF_DIM( VISCELL, zi ) != 1 )
	{
		STV_REGION_MIN( VISREGION, xi ) = xmin;
		STV_REGION_MAX( VISREGION, xi ) = xmax;

		STV_CELL_OF_DIM( VISCELL, xi ) = xcell;

		STV_REGION_MIN( VISREGION, yi ) = ymin;
		STV_REGION_MAX( VISREGION, yi ) = ymax;

		STV_CELL_OF_DIM( VISCELL, yi ) = ycell;

		STV_REGION_MIN( VISREGION, zi ) =
			STV_REGION_MAX( VISREGION, zi ) = zval;

		STV_CELL_OF_DIM( VISCELL, zi ) = 1;

		if ( CURRENT_VF != NULL
			&& stv_check_region( CURRENT_VF->field,
				VISREGION, VISCELL ) == stvTrue )
		{
			update_vis_region();
		}

		if ( CONNECTED )
		{
			/* if is_static ever set, check return code here */
			stv_viewer_set_VisRegion( CURRENT_VFG, VIEWER,
				VISREGION, VISCELL );

			allocate_data_frame();
		}
	}

	return( TCL_OK );
}


/* ARGSUSED */
int set_vis_freq_proc( clientData, itp, argc, argv )
ClientData clientData;
Tcl_Interp *itp;
int argc;
char **argv;
{
	int freq;

	if ( argc != 2 )
	{
		Tcl_SetResult( itp, "usage: set_vis_freq freq", TCL_STATIC );

		return( TCL_ERROR );
	}

	freq = atoi( argv[1] );

	if ( CONNECTED && freq != CURRENT_VFG->freq )
		stv_viewer_set_VisFrequency( CURRENT_VFG, VIEWER, freq );

	return( TCL_OK );
}


/* ARGSUSED */
int set_vis_mode_proc( clientData, itp, argc, argv )
ClientData clientData;
Tcl_Interp *itp;
int argc;
char **argv;
{
	char *mode;

	int tmp;

	if ( argc != 2 )
	{
		Tcl_SetResult( itp,
			"usage: set_vis_mode [ normal | diff | diffsave ]",
			TCL_STATIC );

		return( TCL_ERROR );
	}

	mode = argv[1];

	if ( !strcmp( mode, "normal" ) )
		tmp = VIS_MODE_NORMAL;

	else if ( !strcmp( mode, "diff" ) )
		tmp = VIS_MODE_DIFF;

	else if ( !strcmp( mode, "diffsave" ) )
	{
		if ( SAVE_FIELD != NULL )
			tmp = VIS_MODE_DIFFSAVE;
		
		else
		{
			printf( "set_vis_mode{}:  No Saved Field - " );
			printf( "Reverting to Normal Mode\n" );

			tmp = VIS_MODE_NORMAL;

			SET_TCL_GLOBAL( itp, "vis_mode", "normal" );
		}
	}

	else
	{
		printf( "set_vis_mode{}:  Unknown Vis Mode \"%s\".\n", mode );

		Tcl_SetResult( itp,
			"usage: set_vis_mode [ normal | diff | diffsave ]",
			TCL_STATIC );

		return( TCL_ERROR );
	}

	if ( tmp != VIS_MODE )
	{
		VIS_MODE = tmp;

		if ( CONNECTED )
			allocate_data_frame();
	}

	return( TCL_OK );
}


/* ARGSUSED */
int set_opt_proc( clientData, itp, argc, argv )
ClientData clientData;
Tcl_Interp *itp;
int argc;
char **argv;
{
	char usage[255];

	char *onoff;
	char *opt;

	int tmp;

	if ( argc != 3 )
		goto err;

	opt = argv[1];
	onoff = argv[2];

	if ( !strcmp( onoff, "ON" ) )
		tmp = stvTrue;

	else if ( !strcmp( onoff, "OFF" ) )
		tmp = stvFalse;

	else if ( strcmp( onoff, "none" )
			|| strcmp( opt, "reset_cumulative" ) )
		goto err;

	if ( !strcmp( opt, "auto_color" ) )
		AUTO_COLOR_MAP = tmp;

	else if ( !strcmp( opt, "min_max" ) )
		SHOW_MIN_MAX = tmp;

	else if ( !strcmp( opt, "cumulative" ) )
		SHOW_CUMULATIVE = tmp;

	else if ( !strcmp( opt, "reset_cumulative" ) )
		CUMULATIVE_SET = 0;

	else
		goto err;

	return( TCL_OK );

err:

	sprintf( usage,
		"usage: set_opt [%s|%s|%s|%s] [ON|OFF]",
		"auto_color", "min_max", "cumulative", "reset_cumulative" );

	Tcl_SetResult( itp, stv_copy_str( usage ), TCL_DYNAMIC );

	return( TCL_ERROR );
}


void
update_vis_region()
{
	int xi, yi, zi;

	if ( !get_plane_indices( &xi, &yi, &zi ) )
	{
		printf( "\nError:  Invalid Plane Selection in %s\n\n",
			"update_vis_reg()" );

		return;
	}

	UPDATE_IENTRY( "xmin_entry", STV_REGION_MIN( VISREGION, xi ) );
	UPDATE_IENTRY( "xmax_entry", STV_REGION_MAX( VISREGION, xi ) );
	UPDATE_IENTRY( "xcell_entry", STV_CELL_OF_DIM( VISCELL, xi ) );

	UPDATE_IENTRY( "ymin_entry", STV_REGION_MIN( VISREGION, yi ) );
	UPDATE_IENTRY( "ymax_entry", STV_REGION_MAX( VISREGION, yi ) );
	UPDATE_IENTRY( "ycell_entry", STV_CELL_OF_DIM( VISCELL, yi ) );

	UPDATE_IENTRY( "zval_entry", STV_REGION_MIN( VISREGION, zi ) );
}


int
get_plane_indices( xi, yi, zi )
int *xi;
int *yi;
int *zi;
{
	switch ( PLANE_CHOICE )
	{
		case PLANE_XY:
			*xi = 0;
			*yi = 1;
			*zi = 2;
			break;

		case PLANE_YZ:
			*xi = 1;
			*yi = 2;
			*zi = 0;
			break;

		case PLANE_XZ:
			*xi = 0;
			*yi = 2;
			*zi = 1;
			break;

		default:
		{
			printf( "\nError in %s:  Invalid Plane Selection (%d)\n\n",
				"get_plane_indices()", PLANE_CHOICE );

			return( stvFalse );
		}
	}

	return( stvTrue );
}


/* ARGSUSED */
int resize_stv_proc( clientData, itp, argc, argv )
ClientData clientData;
Tcl_Interp *itp;
int argc;
char **argv;
{
	if ( CONNECTED && SLEEP )
		redraw_view();

	return( TCL_OK );
}


/* ARGSUSED */
int save_field_proc( clientData, itp, argc, argv )
ClientData clientData;
Tcl_Interp *itp;
int argc;
char **argv;
{
	char *cmd;

	int do_usage = 0;

	if ( argc < 2 || argc > 3 )
		do_usage++;

	else
	{
		cmd = argv[1];

		if ( !strcmp( cmd, "set" ) )
			set_current_field();

		else if ( !strcmp( cmd, "free" ) )
			free_current_field();

		else if ( !strcmp( cmd, "save" ) )
		{
			if ( argc != 3 || !strcmp( argv[2], "" ) )
				do_usage++;
			
			else
				save_current_field( argv[2] );
		}

		else if ( !strcmp( cmd, "load" ) )
		{
			if ( argc != 3 || !strcmp( argv[2], "" ) )
				do_usage++;
			
			else
				load_current_field( argv[2] );
		}

		else
			do_usage++;
	}

	if ( do_usage )
	{
		Tcl_SetResult( itp,
		"usage: save_field [set] [free] [save <file>] [load <file>]",
			TCL_STATIC );

		return( TCL_ERROR );
	}

	return( TCL_OK );
}


/* ARGSUSED */
int canvas_handle_proc( clientData, itp, argc, argv )
ClientData clientData;
Tcl_Interp *itp;
int argc;
char **argv;
{
	static CIRCLE SAVEC = (CIRCLE) NULL;

	static RECT SAVER = (RECT) NULL;

	static int markx = -1;
	static int marky = -1;

	STV_VIEW_PARAM VP;

	CIRCLE C;

	RECT R;

	char rstr[1024];
	char cstr[1024];
	char cmd[1024];

	char *op;

	int xi, yi, zi;
	int coords[3];

	int xmin, xmax;
	int ymin, ymax;

	int x, y;
	int flag;
	int tmp;

	if ( MAIN_WIN == NULL )
		return( TCL_OK );

	if ( argc != 4 )
	{
		Tcl_SetResult( itp, "usage: canvas_handle <op> <x> <y>",
			TCL_STATIC );

		return( TCL_ERROR );
	}

	op = argv[1];

	x = atoi( argv[2] );
	y = atoi( argv[3] );

	if ( vflag )
	{
		printf( "(%d, %d) -> (%d, %d)\n", x, y,
			SCREEN_TO_REGION( x, STV_REGION_MIN( VISREGION, 0 ),
				MAIN_WIN->xoff, CURRENT_SCALE ),
			SCREEN_TO_REGION( y, STV_REGION_MIN( VISREGION, 1 ),
				MAIN_WIN->yoff, CURRENT_SCALE ) );
	}

	if ( !strcmp( op, "query_press" ) )
	{
		PTR_X = x;
		PTR_Y = y;

		R = find_rect( x, y );

		if ( R != NULL )
		{
			if ( get_plane_indices( &xi, &yi, &zi ) )
			{
				SAVER = R;

				sprintf( cmd, "%s itemconfigure %d -outline black",
					SL_C, R->id );

				safe_eval( itp, cmd );

				coords[xi] = R->x;
				coords[yi] = R->y;
				coords[zi] = R->z;

				sprintf( cmd,
					"query_handle %d %d \"%s(%d,%d,%d) = %f\"",
					x, y,
					SLICER_FIELD_NAME ? SLICER_FIELD_NAME : "",
					coords[0], coords[1], coords[2], R->value );

				safe_eval( itp, cmd );
			}
		}

		else
		{
			C = find_circle( x, y );

			if ( C != NULL )
			{
				if ( get_plane_indices( &xi, &yi, &zi ) )
				{
					SAVEC = C;

					sprintf( cmd, "%s itemconfigure %d -outline black",
						SL_C, C->id );

					safe_eval( itp, cmd );

					coords[xi] = C->x;
					coords[yi] = C->y;
					coords[zi] = C->z;

					if ( SLICER_RADIUS_FIELD != NULL )
					{
						sprintf( rstr, " <%s = %f>",
							SLICER_RADIUS_FIELD, C->radius );
					}

					else
						strcpy( rstr, "" );

					if ( SLICER_COLOR_FIELD != NULL )
					{
						sprintf( cstr, " <%s = %f>",
							SLICER_COLOR_FIELD, C->value );
					}

					else
						strcpy( cstr, "" );

					sprintf( cmd,
						"query_handle %d %d \"%s(%d,%d,%d):%s%s\"",
						x, y,
						SLICER_FIELD_NAME ? SLICER_FIELD_NAME : "",
						coords[0], coords[1], coords[2], rstr, cstr );

					safe_eval( itp, cmd );
				}
			}
		}
	}

	else if ( !strcmp( op, "query_slide" ) )
	{
		PTR_X = x;
		PTR_Y = y;

		if ( SAVER != NULL )
		{
			sprintf( cmd, "%s itemconfigure %d -outline %s",
				SL_C, SAVER->id, SAVER->color );

			safe_eval( itp, cmd );

			SAVER = (RECT) NULL;
		}

		else if ( SAVEC != NULL )
		{
			sprintf( cmd, "%s itemconfigure %d -outline %s",
				SL_C, SAVEC->id, SAVEC->color );

			safe_eval( itp, cmd );

			SAVEC = (CIRCLE) NULL;
		}

		R = find_rect( x, y );

		if ( R != NULL )
		{
			if ( get_plane_indices( &xi, &yi, &zi ) )
			{
				SAVER = R;

				sprintf( cmd, "%s itemconfigure %d -outline black",
					SL_C, R->id );

				safe_eval( itp, cmd );

				coords[xi] = R->x;
				coords[yi] = R->y;
				coords[zi] = R->z;

				sprintf( cmd,
					"query_handle %d %d \"%s(%d,%d,%d) = %f\"",
					x, y,
					SLICER_FIELD_NAME ? SLICER_FIELD_NAME : "",
					coords[0], coords[1], coords[2], R->value );

				safe_eval( itp, cmd );
			}
		}

		else
		{
			C = find_circle( x, y );

			if ( C != NULL )
			{
				if ( get_plane_indices( &xi, &yi, &zi ) )
				{
					SAVEC = C;

					sprintf( cmd, "%s itemconfigure %d -outline black",
						SL_C, C->id );

					safe_eval( itp, cmd );

					coords[xi] = C->x;
					coords[yi] = C->y;
					coords[zi] = C->z;

					if ( SLICER_RADIUS_FIELD != NULL )
					{
						sprintf( rstr, " <%s = %f>",
							SLICER_RADIUS_FIELD, C->radius );
					}

					else
						strcpy( rstr, "" );

					if ( SLICER_COLOR_FIELD != NULL )
					{
						sprintf( cstr, " <%s = %f>",
							SLICER_COLOR_FIELD, C->value );
					}

					else
						strcpy( cstr, "" );

					sprintf( cmd,
						"query_handle %d %d \"%s(%d,%d,%d):%s%s\"",
						x, y,
						SLICER_FIELD_NAME ? SLICER_FIELD_NAME : "",
						coords[0], coords[1], coords[2], rstr, cstr );

					safe_eval( itp, cmd );
				}
			}

			else
			{
				if ( SAVER != NULL )
				{
					sprintf( cmd, "%s itemconfigure %d -outline %s",
						SL_C, SAVER->id, SAVER->color );

					safe_eval( itp, cmd );

					SAVER = (RECT) NULL;
				}

				else if ( SAVEC != NULL )
				{
					sprintf( cmd, "%s itemconfigure %d -outline %s",
						SL_C, SAVEC->id, SAVEC->color );

					safe_eval( itp, cmd );

					SAVEC = (CIRCLE) NULL;
				}

				safe_eval( itp, "query_handle -1 -1 \"\"" );
			}
		}
	}

	else if ( !strcmp( op, "query_release" ) )
	{
		PTR_X = x;
		PTR_Y = y;

		if ( SAVER != NULL )
		{
			sprintf( cmd, "%s itemconfigure %d -outline %s",
				SL_C, SAVER->id, SAVER->color );

			safe_eval( itp, cmd );

			SAVER = (RECT) NULL;
		}

		else if ( SAVEC != NULL )
		{
			sprintf( cmd, "%s itemconfigure %d -outline %s",
				SL_C, SAVEC->id, SAVEC->color );

			safe_eval( itp, cmd );

			SAVEC = (CIRCLE) NULL;
		}

		safe_eval( itp, "query_handle -1 -1 \"\"" );
	}

	else if ( !strcmp( op, "zoom_press" ) )
	{
		PTR_X = x;
		PTR_Y = y;

		/* Create Zoom Rect */

		if ( REGION_ID != -1 )
			DELETE_ITEM( itp, MAIN_WIN->canvas, REGION_ID );
	
		REGION_ID = DRAW_RECT( itp, MAIN_WIN->canvas, x, y, x, y,
			"", "black", 1 );

		markx = x;
		marky = y;
	}

	else if ( !strcmp( op, "zoom_slide" ) )
	{
		PTR_X = x;
		PTR_Y = y;

		/* Adjust Zoom Rect */

		if ( REGION_ID != -1 )
		{
			RECT_COORDS( itp, MAIN_WIN->canvas, REGION_ID,
				markx, marky, x, y );
		}
	}

	else if ( !strcmp( op, "zoom_release" ) )
	{
		PTR_X = x;
		PTR_Y = y;

		/* Set Desired Vis Region */

		DELETE_ITEM( itp, MAIN_WIN->canvas, REGION_ID );

		if ( CONNECTED && markx != x && marky != y )
		{
			if ( get_plane_indices( &xi, &yi, &zi ) )
			{
				/* Save Old Region */

				push_region( VISREGION );

				/* Determine Min/Max Coords */

				if ( markx < x )
				{
					xmin = markx;
					xmax = x;
				}

				else
				{
					xmin = x;
					xmax = markx;
				}

				if ( marky < y )
				{
					ymin = marky;
					ymax = y;
				}

				else
				{
					ymin = y;
					ymax = marky;
				}

				/* Set New Region Bounds */

				STV_REGION_MAX( VISREGION, xi ) =
					SCREEN_TO_REGION( xmax,
						STV_REGION_MIN( VISREGION, xi ),
						BORDER_SPACE, CURRENT_SCALE );

				STV_REGION_MAX( VISREGION, yi ) =
					SCREEN_TO_REGION( ymax,
						STV_REGION_MIN( VISREGION, yi ),
						BORDER_SPACE, CURRENT_SCALE );

				STV_REGION_MIN( VISREGION, xi ) =
					SCREEN_TO_REGION( xmin,
						STV_REGION_MIN( VISREGION, xi ),
						BORDER_SPACE, CURRENT_SCALE );

				STV_REGION_MIN( VISREGION, yi ) =
					SCREEN_TO_REGION( ymin,
						STV_REGION_MIN( VISREGION, yi ),
						BORDER_SPACE, CURRENT_SCALE );

				printf( "New Display Region [%d - %d][%d - %d]\n",
					STV_REGION_MIN( VISREGION, xi ),
					STV_REGION_MAX( VISREGION, xi ),
					STV_REGION_MIN( VISREGION, yi ),
					STV_REGION_MAX( VISREGION, yi ) );

				stv_check_region( CURRENT_VF->field,
					VISREGION, VISCELL );

				printf(
					"After stv_check_region():  [%d - %d][%d - %d]\n",
					STV_REGION_MIN( VISREGION, xi ),
					STV_REGION_MAX( VISREGION, xi ),
					STV_REGION_MIN( VISREGION, yi ),
					STV_REGION_MAX( VISREGION, yi ) );

				/* Send New Vis Region */

				/* if is_static ever set, check return code here */
				stv_viewer_set_VisRegion( CURRENT_VFG, VIEWER,
					VISREGION, VISCELL );

				update_vis_region();

				allocate_data_frame();
			}
		}
	}

	else if ( !strcmp( op, "pop" ) )
	{
		if ( CONNECTED && pop_region( VISREGION ) )
		{
			/* if is_static ever set, check return code here */
			stv_viewer_set_VisRegion( CURRENT_VFG, VIEWER,
				VISREGION, VISCELL );

			update_vis_region();

			allocate_data_frame();
		}
	}

	else if ( !strcmp( op, "POP" ) )
	{
		if ( CONNECTED )
		{
			flag = 0;

			while ( pop_region( VISREGION ) )
				flag++;

			if ( flag )
			{
				/* if is_static ever set, check return code here */
				stv_viewer_set_VisRegion( CURRENT_VFG, VIEWER,
					VISREGION, VISCELL );

				update_vis_region();

				allocate_data_frame();
			}
		}
	}

	else if ( !strcmp( op, "enter" ) )
	{
		PTR_VALID = stvTrue;

		PTR_X = x;
		PTR_Y = y;
	}

	else if ( !strcmp( op, "move" ) )
	{
		PTR_X = x;
		PTR_Y = y;
	}

	else if ( !strcmp( op, "leave" ) )
	{
		PTR_VALID = stvFalse;

		PTR_X = -1;
		PTR_Y = -1;
	}

	else if ( strlen( op ) == 1 )
	{
		switch ( op[0] )
		{
			case 'q':
			case 'Q':
			{
				Tcl_Eval( itp, "quit" );

				break;
			}

			default:
			{
				printf( "Error in canvas_handle{}: Unknown Key '%c'\n",
					op[0] );
				
				break;
			}
		}
	}

	else
		printf( "Error in canvas_handle{}: Unknown Op = \"%s\"\n", op );

	return( TCL_OK );
}


RECT
find_rect( x, y )
int x;
int y;
{
	RECT R;

	int i;

	R = &(RECTIDS[0]);

	for ( i=0 ; i < NRECTIDS ; i++ )
	{
		if ( x >= R->x1 && x <= R->x2
			&& y >= R->y1 && y <= R->y2 )
		{
			return( R );
		}

		R++;
	}

	return( (RECT) NULL );
}


CIRCLE
find_circle( x, y )
int x;
int y;
{
	CIRCLE Cclose;
	CIRCLE C;

	double close_dist = -1.0;
	double dist;

	int i;

	Cclose = (CIRCLE) NULL;

	C = &(CIRCLEIDS[0]);

	for ( i=0 ; i < NCIRCLEIDS ; i++ )
	{
		dist = distance( (double) x, (double) y,
			(double) C->xc, (double) C->yc );

		if ( dist < close_dist || close_dist == -1.0 )
		{
			close_dist = dist;

			Cclose = C;
		}

		C++;
	}

	return( Cclose );
}


/* ARGSUSED */
int register_menuvar_proc( clientData, itp, argc, argv )
ClientData clientData;
Tcl_Interp *itp;
int argc;
char **argv;
{
	char *menuvar;
	char *index;
	char *menu;

	if ( argc != 4 )
	{
		Tcl_SetResult( itp,
			"usage: register_menuvar <menu> <index> <menuvar>",
			TCL_STATIC );

		return( TCL_ERROR );
	}

	menu	= argv[1];
	index	= argv[2];
	menuvar	= argv[3];

	if ( !strcmp( menu, ".slicer.mode_menu" ) )
	{
		if ( !strcmp( index, "auto_color" ) )
		{
			if ( MENUVAR_AUTO_COLOR != NULL )
				FREE( MENUVAR_AUTO_COLOR );
			
			MENUVAR_AUTO_COLOR = stv_copy_str( menuvar );
		}

		else if ( !strcmp( index, "min_max" ) )
		{
			if ( MENUVAR_MIN_MAX != NULL )
				FREE( MENUVAR_MIN_MAX );
			
			MENUVAR_MIN_MAX = stv_copy_str( menuvar );
		}

		else if ( !strcmp( index, "cumulative" ) )
		{
			if ( MENUVAR_CUMULATIVE != NULL )
				FREE( MENUVAR_CUMULATIVE );
			
			MENUVAR_CUMULATIVE = stv_copy_str( menuvar );
		}
	}

	return( TCL_OK );
}


/* ARGSUSED */
int load_bitmap_file_proc( clientData, itp, argc, argv )
ClientData clientData;
Tcl_Interp *itp;
int argc;
char **argv;
{
	Pixmap bitmap;

	char tmp[1024];

	if ( argc != 2 )
	{
		Tcl_SetResult( itp, "usage: load_bitmap_file <file>",
			TCL_STATIC );

		return( TCL_ERROR );
	}

	sprintf( tmp, "@%s", argv[1] );

	bitmap = Tk_GetBitmap( itp, Top, Tk_GetUid( tmp ) );

	if ( bitmap <= 0 )
	{
		sprintf( tmp, "Error Opening Bitmap %s", argv[1] );

		Tcl_SetResult( itp, stv_copy_str( tmp ), TCL_DYNAMIC );

		return( TCL_ERROR );
	}

	return( TCL_OK );
}


/* ARGSUSED */
int fix_help_line_proc( clientData, itp, argc, argv )
ClientData clientData;
Tcl_Interp *itp;
int argc;
char **argv;
{
	char tmp[1024];

	char *result;
	char *ptr;
	char *str;

	int len;

	if ( argc != 2 )
	{
		Tcl_SetResult( itp, "usage: fix_help_line <line>", TCL_STATIC );

		return( TCL_ERROR );
	}

	str = argv[1];

	ptr = tmp;

	while ( *str != '\0' )
	{
		if ( *str == '' )
			*ptr++ = '{';

		else if ( *str == '' )
			*ptr++ = '}';

		else
			*ptr++ = *str;

		str++;
	}

	*ptr = '\0';

	result = stv_copy_str( tmp );

	Tcl_SetResult( itp, result, TCL_DYNAMIC );

	return( TCL_OK );
}


/* ARGSUSED */
int double_click_proc( clientData, itp, argc, argv )
ClientData clientData;
Tcl_Interp *itp;
int argc;
char **argv;
{
	static struct timeval last = { -1, -1 };

	struct timeval tmp;

	float diff;

	gettimeofday( &tmp, (struct timezone *) NULL );

	diff = ( (float) ( tmp.tv_sec - last.tv_sec ) * 1000000.0 )
		+ (float) ( tmp.tv_usec - last.tv_usec );

	if ( diff < 500000.0 )
		Tcl_SetResult( itp, "1", TCL_STATIC );

	else
		Tcl_SetResult( itp, "0", TCL_STATIC );

	last.tv_sec = tmp.tv_sec;
	last.tv_usec = tmp.tv_usec;

	return( TCL_OK );
}


/* ARGSUSED */
int strip_label_proc( clientData, itp, argc, argv )
ClientData clientData;
Tcl_Interp *itp;
int argc;
char **argv;
{
	char tmp[1024];

	char *result;
	char *ptr;
	char *str;

	int len;

	if ( argc != 2 )
	{
		Tcl_SetResult( itp, "usage: strip_label <str>", TCL_STATIC );

		return( TCL_ERROR );
	}

	str = argv[1];

	ptr = tmp;

	while ( *str != '\0' )
	{
		if ( *str == '-' || *str == ' ' || *str == '.' )
			*ptr++ = '_';

		else
			*ptr++ = *str;

		str++;
	}

	*ptr = '\0';

	result = stv_copy_str( tmp );

	Tcl_SetResult( itp, result, TCL_DYNAMIC );

	return( TCL_OK );
}


/* ARGSUSED */
int title_info_proc( clientData, itp, argc, argv )
ClientData clientData;
Tcl_Interp *itp;
int argc;
char **argv;
{
	char tmp[255];

	sprintf( tmp, "%s", SLICER_VERSION );

	Tcl_SetResult( itp, stv_copy_str( tmp ), TCL_DYNAMIC );

	return( TCL_OK );
}


/* ARGSUSED */
int pad_proc( clientData, itp, argc, argv )
ClientData clientData;
Tcl_Interp *itp;
int argc;
char **argv;
{
	char *result;
	char *ptr;
	char *str;

	int len;
	int nd;
	int nz;
	int i;

	if ( argc < 3 )
	{
		Tcl_SetResult( itp,
			"usage: pad <number_str> <desired_length>",
			TCL_STATIC );

		return( TCL_ERROR );
	}

	str = argv[1];

	len = atoi( argv[2] );

	nd = strlen( str );

	nz = len - nd;

	result = (char *) MALLOC( (STV_SIZE_T) (len + 1) * sizeof(char) );
	memcheck( result, "Numerical Pad String" );

	ptr = result;

	for ( i=0 ; i < nz ; i++ )
		*ptr++ = '0';

	sprintf( (char *) (result + nz), "%s", str );

	Tcl_SetResult( itp, result, TCL_DYNAMIC );

	return( TCL_OK );
}


/* ARGSUSED */
int sleep_set_proc( clientData, itp, argc, argv )
ClientData clientData;
Tcl_Interp *itp;
int argc;
char **argv;
{
	int xon_chk = 0;
	int flag = 0;
	int kick = 0;

	if ( argc > 2 )
	{
		Tcl_SetResult( itp, "usage: sleep_set [ ON | OFF | STEP ]",
			TCL_STATIC );

		return( TCL_ERROR );
	}

	else if ( argc == 2 )
	{
		if ( !strcmp( argv[1], "ON" ) )
		{
			SLEEP = stvTrue;

			flag++;
		}

		else if ( !strcmp( argv[1], "OFF" ) )
		{
			/* Don't Un-Sleep If Locked - Just Mucks Things Up */
			/* (But if trying to unlock, let Un-Sleep...) */
			if ( CONNECTED && CURRENT_VFG->locked == stvAppLocked
					&& APP_LOCKED != APP_LOCK_UNLOCKED )
			{
				NSTEPS++;
			}

			else
				SLEEP = stvFalse;

			xon_chk++;

			kick++;

			flag++;
		}

		else if ( !strcmp( argv[1], "STEP" ) )
		{
			NSTEPS++;

			xon_chk++;

			kick++;
		}

		else
		{
			Tcl_SetResult( itp, "usage: sleep_set [ ON | OFF ]",
				TCL_STATIC );

			return( TCL_ERROR );
		}
	}

	else
		Tcl_SetResult( itp, SLEEP ? "ON" : "OFF", TCL_STATIC );

	if ( flag )
		SET_TCL_GLOBAL( itp, "sleeping", SLEEP ? "ON" : "OFF" );

	/* Check for XON Kick */

	if ( xon_chk && XON_KICK )
	{
		if ( CONNECTED )
			stv_viewer_send_XON( CURRENT_VFG, VIEWER );

		if ( NSTEPS > 0 )
			NSTEPS--;

		XON_KICK = stvFalse;
	}

	/* Kick Viewer In The Butt If It Was Asleep */

	if ( kick && VIEWER_KICK )
	{
		Tk_DoWhenIdle( viewer_handle, (ClientData) NULL );

		VIEWER_KICK = stvFalse;
	}

	return( TCL_OK );
}


/* ARGSUSED */
int lock_set_proc( clientData, itp, argc, argv )
ClientData clientData;
Tcl_Interp *itp;
int argc;
char **argv;
{
	if ( argc != 2 )
	{
		Tcl_SetResult( itp,
			"usage: lock_set [ lock | unlock ]",
			TCL_STATIC );

		return( TCL_ERROR );
	}

	if ( CONNECTED )
	{
		/* lock */

		if ( !strcmp( argv[1], "lock" ) )
		{
			stv_viewer_send_AppLock( CURRENT_VFG, VIEWER, 0 );

			APP_LOCKED = APP_LOCK_LOCKED;
		}

		/* unlock */

		else if ( !strcmp( argv[1], "unlock" ) )
		{
			stv_viewer_send_AppUnlock( CURRENT_VFG, VIEWER );

			APP_LOCKED = APP_LOCK_UNLOCKED;

			if ( SLEEP )
				Tcl_Eval( itp, "sleep_set OFF" );
		}
	}

	return( TCL_OK );
}


/* ARGSUSED */
void
quit_proc( clientData, itp, argc, argv )
ClientData clientData;
Tcl_Interp *itp;
int argc;
char **argv;
{
	if ( CURRENT_VFG != NULL )
		stv_viewer_send_FieldHaltAll( CURRENT_VFG, VIEWER );

	exit( 0 );
}

