
/* $Id: cp_charge.c,v 3.17 2000/07/18 15:53:15 kohl Exp $ */


/*
 *                      CUMULVS version 1.0:
 *      Computational Steering and Interactive Visualization
 *                   in Distributed Applications
 *           Oak Ridge National Laboratory, Oak Ridge TN.
 *      Authors:  James Arthur Kohl and Philip M. Papadopoulos
 *                   (C) 1996 All Rights Reserved
 *
 *                              NOTICE
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted
 * provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear
 * in supporting documentation.
 *
 * Neither the Institution, Oak Ridge National Laboratory, nor the
 * Authors make any representations about the suitability of this
 * software for any purpose.  This software is provided ``as is''
 * without express or implied warranty.
 *
 * CUMULVS was funded by the U.S. Department of Energy.
 */


/* System Header Files */

#include <stdio.h>
#include <sys/types.h>

#include <malloc.h>
#include <math.h>

#ifndef PI
#define PI (M_PI)
#endif


/* STV Headers */

#include "stv.h"


/* Communication Constants */

#define APPNAME				"cp_charge"

#define GROUPNAME			"ChargeItBig"

#define STV_MSGTAG			10

#define TRANSFER_MSG		100

#define END_TRANSFER		-1
#define BLIP_TRANSFER		-2
#define CHARGE_TRANSFER		-3


/* Simulation Constants */

#define FIELD_HEIGHT		1000
#define FIELD_WIDTH			1000

#define CHARGE_BOUNCE		10000


/* Macros */

#define NUM_CHARGE_IDS		100000

#define CHARGE_ID( _region ) \
( NEXT_CHARGE_ID = NEXT_CHARGE_ID >= NUM_CHARGE_IDS - 1 ? 0 : \
	NEXT_CHARGE_ID + 1, \
	( ( (_region) + 1 ) * NUM_CHARGE_IDS ) + NEXT_CHARGE_ID )

#define DUMP_CHARGE( _str, _cid, _x, _y, _val, _r, _m, _va, _vm ) \
{ \
	if ( vflag ) \
	{ \
		printf( "%s charge %d (%.2lf,%.2lf) ", _str, _cid, _x, _y ); \
		printf( "V(%.2lf) R(%.2lf) M(%.2lf) ", _val, _r, _m ); \
		printf( "V(%.2lf,%.2lf)\n", _va, _vm ); \
		fflush( stdout ); \
	} \
}


#define EQUIV_DELTA			0.000001

#define EQUIV( _d1, _d2 ) \
( ( (_d1) != 0.0 ) ? \
		( ( fabs( (_d1) - (_d2) ) / fabs( _d1 ) ) < EQUIV_DELTA ) \
		: ( fabs( (_d1) - (_d2) ) < EQUIV_DELTA ) )


#define IN_RANGE( _x, _x0, _a ) \
( ( (_a) > 0 ) ? \
	( (_x0) <= (_x) && (_x) <= ( (_x0) + (_a) ) ) \
	: ( (_x0) >= (_x) && (_x) >= ( (_x0) + (_a) ) ) )


/* Data Structures */

struct region_struct
{
	int tid;
	int xmin, xmax;
	int ymin, ymax;
	int nneigh;
	/* REGION */ struct region_struct *neigh;
};

typedef struct region_struct *REGION;


struct charge_struct
{
	int cid;
	double loc_x, loc_y;
	int bounce;
	double value;
	double radius;
	double mass;
	double vel_angle;
	double vel_magn;
	/* CHARGE */ struct charge_struct *next;
};

typedef struct charge_struct *CHARGE;


struct blip_struct
{
	double loc_x, loc_y;
	double value;
	double radius;
	double mass;
	/* BLIP */ struct blip_struct *next;
};

typedef struct blip_struct *BLIP;


/* Routines */

void	get_charge __STV_Prototype__(( int, STV_REGION,
			STV_PARTICLE_ID *, int * ));

void	get_charge_pfield __STV_Prototype__(( STV_PARTICLE_ID,
			STV_VALUE, STV_VALUE * ));


void	make_charge __STV_Prototype__(( int *, STV_PARTICLE_ID * ));

void	set_charge_pfield __STV_Prototype__(( STV_PARTICLE_ID,
			STV_VALUE, STV_VALUE ));


#define PF_CID			0
#define PF_VALUE		1
#define PF_RADIUS		2
#define PF_MASS			3
#define PF_VEL_ANGLE	4
#define PF_VEL_MAGN		5
#define PF_X_COORD		6
#define PF_Y_COORD		7


REGION		create_region();

CHARGE		create_charge(), add_charge(), get_charge_by_cid();

BLIP		create_blip(), add_blip();

double		CHARGE_VALUE, CHARGE_RADIUS, CHARGE_MASS;
int			CHARGE_INDEX;

double		FIELD_ANGLE, FIELD_MAGN;

double		angle_of();
double		atof();
double		distance();
double		pwr3();
double		rand_angle();
double		rand_double();
double		sq();

void		add_local_blips();
void		add_rand_charge();
void		add_to_vector();
void		app_init();
void		charge_action();
void		charge_handle();
void		charge_init();
void		check_exits();
void		cleanup_region();
void		do_stv_stuff();
void		eff_vel();
void		free_blip();
void		free_blip_list();
void		free_charge();
void		free_region();
void		free_region_list();
void		handle_steering();
void		memcheck();
void		neigh_init();
void		normalize_angle();
void		program_init();
void		pvm_init();
void		rand_point();
void		read_args();
void		region_init();
void		send_transfers();
void		usage();

int			NEXT_CHARGE_ID;

int			CP_RESTART;


/* Externals */

extern		int 	errno;


/* Global Variables */

REGION		GLOBREGION;

REGION		MYREGION;

CHARGE		CHARGES;
int			NCHARGES;

BLIP		BLIPS;

double		APPEAR_PROB;
double		DISAPPEAR_PROB;

char		**Argv;
int			Argc;

int			MYINST;
int			NTASKS;
int			NPARTS;
int			MYTID;
int			WALLS;
int			SPMD;

int 		vflag;
int			qflag;
int 		dflag;
int 		cflag;


/* Main */

main( argc, argv )
int argc;
char **argv;
{
	int cc;

	Argc = argc;
	Argv = argv;

	read_args();

	program_init();

	pvm_init();

	do_stv_stuff();

	app_init();

	if ( !CP_RESTART )
		charge_init();

	cc = stvTrue;

	while ( stvTrue )
	{
		if ( !cc || CP_RESTART )
		{
			cleanup_region();

			if ( !cc )
				region_init();

			stv_loadFromCP();

			CP_RESTART = stvFalse;

			neigh_init();
		}

		cc = iterate();
	}
}


iterate()
{
	static int cnt = 0;

	CHARGE C;

	int new_params;
	int cc;

	/* Dump Out Iteration Tick */

	if ( !qflag && !(++cnt % 100) )
	{
		printf( "." );
		fflush( stdout );
	}

	/* Update Objects */

	charge_handle();

	/* Reset Radar Blips */

	free_blip_list( &BLIPS );

	add_local_blips();

	/* Communicate Transfers */

	send_transfers();

	check_exits();

	cc = recv_transfers();

	if ( !cc )
		return( stvFalse );

	/* Check for Actions */

	charge_action();

	/* Handle STV Requests */

	new_params = stv_sendToFE();

	if ( new_params )
		handle_steering();

	/* Do Checkpointing!!! */

	if ( !(cnt % 100) )
		stv_checkpoint();

	return( stvTrue );
}


void
charge_handle()
{
	CHARGE Clast;
	CHARGE Cnext;
	CHARGE C, C2;

	double cx1, cy1, cx2, cy2;
	double factor1, factor2;
	double vol1, vol2;
	double ba1, bm1;
	double ba2, bm2;
	double ra1, ra2;
	double vc1, vc2;
	double a1, a2;
	double dx, dy;
	double hx, hy;
	double t;

	int strength;
	int absorb;
	int radius;
	int hit;

	/* Update Each Charge */

	C = CHARGES;

	Clast = (CHARGE) NULL;

	while ( C != NULL )
	{
		/* Randomly Remove Some Charges... */
		/* (Call it Quantum Effects, Baby) */

		if ( rand_double() < DISAPPEAR_PROB )
		{
			if ( vflag )
			{
				printf(
					"\n*** Charge %d Spontaneously Annihilates ***\n",
					C->cid );
				fflush( stdout );
			}

			else if ( !qflag )
			{
				printf( "X" );
				fflush( stdout );
			}

			if ( Clast == NULL )
				CHARGES = C->next;

			else
				Clast->next = C->next;

			Cnext = C->next;

			free_charge( &C );

			C = Cnext;

			NCHARGES--;
		}

		else
		{
			/* Check for Other Charge Collision */

			C2 = C->next;

			absorb = stvTrue;

			hit = stvFalse;

			while ( C2 != NULL && !hit )
			{
				if ( charge_hits_charge( C, C2, &hx, &hy, &t ) )
				{
					if ( vflag )
					{
						printf( "\n*** Charge Collision %d/%d ***\n",
							C->cid, C2->cid );
						fflush( stdout );
					}

					else if ( !qflag )
					{
						printf( "*" );
						fflush( stdout );
					}

					/* Determine Impact Strength */

					cx1 = C->loc_x + ( t * C->vel_magn
						* cos( C->vel_angle + (PI / 2.0) ) );
					cy1 = C->loc_y - ( t * C->vel_magn
						* sin( C->vel_angle + (PI / 2.0) ) );

					cx2 = C2->loc_x + ( t * C2->vel_magn
						* cos( C2->vel_angle + (PI / 2.0) ) );
					cy2 = C2->loc_y - ( t * C2->vel_magn
						* sin( C2->vel_angle + (PI / 2.0) ) );

					a1 = angle_of( cx1, cy1, hx, hy );

					a2 = a1 - PI;

					ra1 = a1 - C->vel_angle;
					ra2 = a2 - C2->vel_angle;

					factor1 = ( 1.0 + cos( ra1 ) ) / 2.0;
					factor2 = ( 1.0 + cos( ra2 ) ) / 2.0;

					strength = (int)
						( ( C->mass * sq( C->vel_magn ) * factor1 )
						+ ( C2->mass * sq( C2->vel_magn ) * factor2 ) )
						/ 2;

					if ( !qflag )
					{
						printf( "%d", strength );
						fflush( stdout );
					}

					/* Charges Bounce Apart */

					if ( strength < CHARGE_BOUNCE )
					{
						ba1 = bm1 = 0.0;
						ba2 = bm2 = 0.0;

						vc1 = ( C->mass - C2->mass )
							/ ( C->mass + C2->mass );
						vc2 = ( 2.0 * C2->mass )
							/ ( C->mass + C2->mass );

						add_to_vector( &ba1, &bm1,
							C->vel_angle * vc1, C->vel_magn * vc1 );
						add_to_vector( &ba1, &bm1,
							C2->vel_angle * vc2, C2->vel_magn * vc2 );

						vc2 = -1.0 * vc1;
						vc1 = ( 2.0 * C->mass )
							/ ( C->mass + C2->mass );

						add_to_vector( &ba2, &bm2,
							C->vel_angle * vc1, C->vel_magn * vc1 );
						add_to_vector( &ba2, &bm2,
							C2->vel_angle * vc2, C2->vel_magn * vc2 );

						C->vel_angle = ba1;
						C->vel_magn = bm1;

						C2->vel_angle = ba2;
						C2->vel_magn = bm2;

						C->loc_x = cx1 + ( ( 1.0 - t ) * C->vel_magn
							* cos( C->vel_angle + (PI / 2.0) ) );
						C->loc_y = cy1 - ( ( 1.0 - t ) * C->vel_magn
							* sin( C->vel_angle + (PI / 2.0) ) );

						C2->loc_x = cx2 + ( ( 1.0 - t ) * C2->vel_magn
							* cos( C2->vel_angle + (PI / 2.0) ) );
						C2->loc_y = cy2 - ( ( 1.0 - t ) * C2->vel_magn
							* sin( C2->vel_angle + (PI / 2.0) ) );

						if ( strength > 0 )
						{
							C->bounce = stvTrue;
							C2->bounce = stvTrue;
						}

						absorb = stvFalse;
					}

					hit = stvTrue;
				}

				else
					C2 = C2->next;
			}

			if ( hit && absorb )
			{
				/* Merge Charges */

				if ( vflag )
				{
					printf( "\n*** Charges Merge %d/%d ***\n",
						C->cid, C2->cid );
					fflush( stdout );
				}

				else if ( !qflag )
				{
					printf( "M" );
					fflush( stdout );
				}

				ba1 = bm1 = 0.0;

				vc1 = C->mass / ( C->mass + C2->mass );
				vc2 = C2->mass / ( C->mass + C2->mass );

				add_to_vector( &ba1, &bm1,
					C->vel_angle * vc1, C->vel_magn * vc1 );
				add_to_vector( &ba1, &bm1,
					C2->vel_angle * vc2, C2->vel_magn * vc2 );

				C2->vel_angle = ba1;
				C2->vel_magn = bm1;

				C2->mass += C->mass;

				C2->loc_x = hx + ( ( 1.0 - t ) * C2->vel_magn
					* cos( C2->vel_angle + (PI / 2.0) ) );
				C2->loc_y = hy - ( ( 1.0 - t ) * C2->vel_magn
					* sin( C2->vel_angle + (PI / 2.0) ) );

				C2->value += C->value;

				vol1 = PI * pwr3( C->radius );
				vol2 = PI * pwr3( C2->radius );

				C2->radius = cbrt( ( vol1 + vol2 ) / PI );

				/* Remove Original Charge */

				if ( Clast == NULL )
					CHARGES = C->next;

				else
					Clast->next = C->next;

				Cnext = C->next;

				free_charge( &C );

				C = Cnext;

				NCHARGES--;
			}

			else
			{
				/* Move Charge */

				if ( !(C->bounce) )
				{
					dx = C->vel_magn * cos( C->vel_angle + (PI / 2.0) );
					dy = C->vel_magn * sin( C->vel_angle + (PI / 2.0) );

					C->loc_x += dx;
					C->loc_y -= dy;

					/* Keep frigging charge in walled region */

					if ( WALLS )
					{
						/* Bounce off walls */

						radius = (int) C->radius;

						if ( (int) C->loc_x - radius
								<= GLOBREGION->xmin
							|| (int) C->loc_x + radius
								>= GLOBREGION->xmax
							|| (int) C->loc_y - radius
								<= GLOBREGION->ymin
							|| (int) C->loc_y + radius
								>= GLOBREGION->ymax )
						{
							/* Reverse Angle */

							if ( (int) C->loc_x - radius
										< GLOBREGION->xmin
									|| (int) C->loc_x + radius
										> GLOBREGION->xmax )
								C->vel_angle *= -1.0;

							else if ( (int) C->loc_y - radius
										< GLOBREGION->ymin
									|| (int) C->loc_y + radius
										> GLOBREGION->ymax )
								C->vel_angle = PI - C->vel_angle;

							normalize_angle( &(C->vel_angle) );

							/* Keep in Region */

							if ( (int) C->loc_x - radius
									<= GLOBREGION->xmin )
								C->loc_x = (double) ( GLOBREGION->xmin
									+ radius );

							if ( (int) C->loc_x + radius
									>= GLOBREGION->xmax )
								C->loc_x = (double) ( GLOBREGION->xmax
									- radius );

							if ( (int) C->loc_y - radius
									<= GLOBREGION->ymin )
								C->loc_y = (double) ( GLOBREGION->ymin
									+ radius );

							if ( (int) C->loc_y + radius
									>= GLOBREGION->ymax )
								C->loc_y = (double) ( GLOBREGION->ymax
									- radius );
						}
					}
				}

				else
					C->bounce = stvFalse;

				/* Next Charge */

				Clast = C;

				C = C->next;
			}
		}
	}
}


charge_hits_charge( C, C2, gx, gy, t )
CHARGE C, C2;
double *gx, *gy;
double *t;
{
	double eff_hvel_angle1;
	double eff_hvel_angle2;

	double eff_hvel_magn1;
	double eff_hvel_magn2;

	int hit;

	/* Determine Effective Velocity Angles & Magns */

	eff_vel( C->vel_magn, C->vel_angle,
		&eff_hvel_magn1, &eff_hvel_angle1 );

	eff_vel( C2->vel_magn, C2->vel_angle,
		&eff_hvel_magn2, &eff_hvel_angle2 );

	/* Check for Possibility of Collision */

	if ( not_even_close(
		C->loc_x, C->loc_y, eff_hvel_magn1, C->radius,
		C2->loc_x, C2->loc_y, eff_hvel_magn2, C2->radius ) )
	{
		*gx = *gy = -1.0;

		return( stvFalse );
	}

	hit = collision_detect(
		C->loc_x, C->loc_y,
			eff_hvel_angle1, eff_hvel_magn1, C->radius,
		C2->loc_x, C2->loc_y,
			eff_hvel_angle2, eff_hvel_magn2, C2->radius,
		gx, gy, t );

	return( hit );
}


void
eff_vel( magn, angle, eff_magn, eff_angle )
double magn, angle;
double *eff_magn, *eff_angle;
{
	double x, y;

	x = magn * cos( angle + (PI / 2.0) );
	y = magn * sin( angle + (PI / 2.0) );

	*eff_magn = sqrt( ( x * x ) + ( y * y ) );

	*eff_angle = angle_of( (double) 0.0, (double) 0.0, x, -1.0 * y );
}


not_even_close( x1, y1, vel1, rad1, x2, y2, vel2, rad2 )
double x1, y1;
double vel1;
double rad1;
double x2, y2;
double vel2;
double rad2;
{
	double dist;

	dist = distance( x1, y1, x2, y2 );

	if ( dist > vel1 + rad1 + vel2 + rad2 + 1.0 )
		return( stvTrue );

	else
		return( stvFalse );
}


collision_detect(
	x10, y10, angle1, magn1, rad1,
	x20, y20, angle2, magn2, rad2,
	xh, yh, timpact )
double x10, y10;
double angle1;
double magn1;
double rad1;
double x20, y20;
double angle2;
double magn2;
double rad2;
double *xh, *yh;
double *timpact;
{
	double B1, B2, C1, G, H1, H2, I1, J1;
	double A, B, C;
	double DET;
	double R;

	double theta1, theta2;
	double a1, b1;
	double a2, b2;

	double a1sq;
	double a2sq;

	double pdiff, mdiff;
	double x1p, x1m;
	double x2p, x2m;
	double y1p, y1m;
	double x1, y1;
	double x2, y2;

	double dck1, dck2;

	y10 = ( (double) FIELD_HEIGHT ) - y10;
	y20 = ( (double) FIELD_HEIGHT ) - y20;

	theta1 = angle1 + ( PI / 2.0 );
	theta2 = angle2 + ( PI / 2.0 );

	a1 = magn1 * cos( theta1 );
	b1 = magn1 * sin( theta1 );

	a2 = magn2 * cos( theta2 );
	b2 = magn2 * sin( theta2 );

	if ( cflag )
	{
		printf( "\n[(%lf,%lf):%lf@%lf->%lf,%lf,%lf,%lf]\n",
			x10, y10, magn1, angle1, theta1, a1, b1, rad1 );

		printf( "[(%lf,%lf):%lf@%lf->%lf,%lf,%lf,%lf]\n",
			x20, y20, magn2, angle2, theta2, a2, b2, rad2 );
	}

	if ( EQUIV( a1, 0.0 ) && EQUIV( a2, 0.0 ) )
	{
		I1 = y20 - ( ( b2 * y10 ) / b1 );

		J1 = ( b2 / b1 ) - 1;

		R = rad1 + rad2;

		A = sq( J1 );

		B = 2.0 * I1 * J1;

		C = sq( I1 ) + sq( x10 - x20 ) - sq( R );

		DET = sq( B ) - ( 4.0 * A * C );

		/* Check For No Intersection - MISS */

		if ( DET < 0 )
		{
			if ( cflag )
				{ printf( "\tDETA\n" ); fflush( stdout ); }

			*xh = *yh = -1.0;

			return( stvFalse );
		}

		/* Determine Correct Solution Point */

		y1p = ( ( -1.0 * B ) + sqrt( DET ) ) / ( 2.0 * A );
		y1m = ( ( -1.0 * B ) - sqrt( DET ) ) / ( 2.0 * A );

		pdiff = fabs( y1p - y10 );

		mdiff = fabs( y1m - y10 );

		if ( pdiff < mdiff )
			y1 = y1p;

		else
			y1 = y1m;

		/* Get Rest of Object Coords */

		x1 = x10;

		x2 = x20;

		y2 = ( ( b2 * y1 ) / b1 ) + I1;
	}

	else if ( EQUIV( a1, 0.0 ) && !EQUIV( a2, 0.0 ) )
	{
		a2sq = sq( a2 );

		B2 = y20 - ( ( b2 * x20 ) / a2 );

		H2 = y10 - ( ( b1 * x20 ) / a2 );

		R = rad1 + rad2;

		A = ( sq( a2 ) + sq( b2 - b1 ) ) / a2sq;

		B = ( 2.0
			* ( ( ( b2 - b1 ) * ( B2 - H2 ) ) - ( a2 * x10 ) ) )
			/ a2;

		C = sq( B2 - H2 ) + sq( x10 ) - sq( R );

		DET = sq( B ) - ( 4.0 * A * C );

		/* Check For No Intersection - MISS */

		if ( DET < 0 )
		{
			if ( cflag )
				{ printf( "\tDETB\n" ); fflush( stdout ); }

			*xh = *yh = -1.0;

			return( stvFalse );
		}

		/* Determine Correct Solution Point */

		x2p = ( ( -1.0 * B ) + sqrt( DET ) ) / ( 2.0 * A );
		x2m = ( ( -1.0 * B ) - sqrt( DET ) ) / ( 2.0 * A );

		pdiff = fabs( x2p - x20 );

		mdiff = fabs( x2m - x20 );

		if ( pdiff < mdiff )
			x2 = x2p;

		else
			x2 = x2m;

		/* Get Rest of Object Coords */

		x1 = x10;

		y1 = ( ( b1 * x2 ) / a2 ) + H2;

		y2 = ( ( b2 * x2 ) / a2 ) + B2;
	}

	else if ( !EQUIV( a1, 0.0 ) && EQUIV( a2, 0.0 ) )
	{
		a1sq = sq( a1 );

		B1 = y10 - ( ( b1 * x10 ) / a1 );

		H1 = y20 - ( ( b2 * x10 ) / a1 );

		R = rad1 + rad2;

		A = ( sq( a1 ) + sq( b2 - b1 ) ) / a1sq;

		B = ( 2.0
			* ( ( ( b1 - b2 ) * ( B1 - H1 ) ) - ( a1 * x20 ) ) )
			/ a1;

		C = sq( B1 - H1 ) + sq( x20 ) - sq( R );

		DET = sq( B ) - ( 4.0 * A * C );

		/* Check For No Intersection - MISS */

		if ( DET < 0 )
		{
			if ( cflag )
				{ printf( "\tDETC\n" ); fflush( stdout ); }

			*xh = *yh = -1.0;

			return( stvFalse );
		}

		/* Determine Correct Solution Point */

		x1p = ( ( -1.0 * B ) + sqrt( DET ) ) / ( 2.0 * A );
		x1m = ( ( -1.0 * B ) - sqrt( DET ) ) / ( 2.0 * A );

		pdiff = fabs( x1p - x10 );

		mdiff = fabs( x1m - x10 );

		if ( pdiff < mdiff )
			x1 = x1p;

		else
			x1 = x1m;

		/* Get Rest of Object Coords */

		x2 = x20;

		y2 = ( ( b2 * x1 ) / a1 ) + H1;

		y1 = ( ( b1 * x1 ) / a1 ) + B1;
	}

	else
	{
		a1sq = sq( a1 );

		B1 = y10 - ( ( b1 * x10 ) / a1 );
		B2 = y20 - ( ( b2 * x20 ) / a2 );

		C1 = x20 - ( ( a2 * x10 ) / a1 );

		G = ( ( b2 * C1 ) / a2 ) + B2;

		R = rad1 + rad2;

		A = ( sq( a2 - a1 ) + sq( b2 - b1 ) ) / a1sq;

		B = ( 2.0
			* ( ( C1 * ( a2 - a1 ) ) + ( ( G - B1 ) * ( b2 - b1 ) ) ) )
			/ a1;

		C = sq( B1 - G ) + sq( C1 ) - sq( R );

		DET = sq( B ) - ( 4.0 * A * C );

		/* Check For No Intersection - MISS */

		if ( DET < 0 )
		{
			if ( cflag )
				{ printf( "\tDETD\n" ); fflush( stdout ); }

			*xh = *yh = -1.0;

			return( stvFalse );
		}

		/* Determine Correct Solution Point */

		x1p = ( ( -1.0 * B ) + sqrt( DET ) ) / ( 2.0 * A );
		x1m = ( ( -1.0 * B ) - sqrt( DET ) ) / ( 2.0 * A );

		pdiff = fabs( x1p - x10 );

		mdiff = fabs( x1m - x10 );

		if ( pdiff < mdiff )
			x1 = x1p;

		else
			x1 = x1m;

		/* Get Rest of Object Coords */

		y1 =  (( b1 * x1 ) / a1 ) + B1;

		x2 = ( ( a2 * x1 ) / a1 ) + C1;

		y2 = ( ( b2 * x2 ) / a2 ) + B2;
	}

	/* Verify Coords Reached in Time */

	if ( !IN_RANGE( x1, x10, a1 )
		|| !IN_RANGE( y1, y10, b1 )
		|| !IN_RANGE( x2, x20, a2 )
		|| !IN_RANGE( y2, y20, b2 ) )
	{
		if ( cflag )
		{
			printf( "\tNIR\n" );

			printf(
	"\t\t%lf,%lf,%lf\n\t%lf,%lf,%lf\n\t%lf,%lf,%lf\n\t%lf,%lf,%lf\n",
				x1, x10, a1, y1, y10, b1, x2, x20, a2, y2, y20, b2 );
		}

		*xh = *yh = -1.0;

		return( stvFalse );
	}

	/* Calculate Glance Point */

	if ( cflag )
	{
		printf( "\t***\n" );

		printf(
		"\t\t(%lf,%lf,%lf/%lf,%lf,%lf/%lf,%lf,%lf/%lf,%lf,%lf)\n",
			x1, x10, a1, y1, y10, b1, x2, x20, a2, y2, y20, b2 );
	}

	y1 = ( (double) FIELD_HEIGHT ) - y1;
	y2 = ( (double) FIELD_HEIGHT ) - y2;

	if ( !EQUIV( R, 0.0 ) )
	{
		*xh = ( ( rad1 * x2 ) + ( rad2 * x1 ) ) / R;

		*yh = ( ( rad1 * y2 ) + ( rad2 * y1 ) ) / R;
	}

	else
	{
		*xh = x1;
		*yh = y1;
	}

	if ( cflag )
		printf( "\t\tH( %lf, %lf )\n", *xh, *yh );

	/* Calculate Time of Impact (Relative to Starting Point */

	if ( !EQUIV( a1, 0.0 ) )
		*timpact = ( x1 - x10 ) / a1;
	else if ( !EQUIV( b1, 0.0 ) )
		*timpact = ( y1 - y10 ) / b1;
	else if ( !EQUIV( a2, 0.0 ) )
		*timpact = ( x2 - x20 ) / a2;
	else if ( !EQUIV( b2, 0.0 ) )
		*timpact = ( y2 - y20 ) / b2;

	/* Return Hit */

	return( stvTrue );
}


void
add_local_blips()
{
	CHARGE C;

	/* Add Charge Blips */

	C = CHARGES;

	while ( C != NULL )
	{
		add_blip( C->loc_x, C->loc_y, C->value, C->radius, C->mass );

		C = C->next;
	}
}


void
send_transfers()
{
	REGION R;

	CHARGE Clast;
	CHARGE Cnext;
	CHARGE C;

	int tmp;

	/* Check for Transfers to Each Neighbor */

	R = MYREGION->neigh;

	while ( R != NULL )
	{
		/* Initialize Message */

		stv_userInitSend( PvmDataDefault, TRANSFER_MSG );

		/* Check for Charges */

		C = CHARGES;

		Clast = (CHARGE) NULL;

		while ( C != NULL )
		{
			if ( (int) C->loc_x >= R->xmin
				&& (int) C->loc_x <= R->xmax
				&& (int) C->loc_y >= R->ymin
				&& (int) C->loc_y <= R->ymax )
			{
				/* Transfer Charge */

				DUMP_CHARGE( "\nsending",
					C->cid, C->loc_x, C->loc_y,
					C->value, C->radius, C->mass,
					C->vel_angle, C->vel_magn );

				tmp = CHARGE_TRANSFER;
				pvm_pkint( &tmp, 1, 1 );

				pvm_pkint( &(C->cid), 1, 1 );
				pvm_pkdouble( &(C->loc_x), 1, 1 );
				pvm_pkdouble( &(C->loc_y), 1, 1 );
				pvm_pkdouble( &(C->value), 1, 1 );
				pvm_pkdouble( &(C->radius), 1, 1 );
				pvm_pkdouble( &(C->mass), 1, 1 );
				pvm_pkdouble( &(C->vel_angle), 1, 1 );
				pvm_pkdouble( &(C->vel_magn), 1, 1 );

				/* Remove From List */

				if ( Clast == NULL )
					CHARGES = C->next;

				else
					Clast->next = C->next;

				Cnext = C->next;

				free_charge( &C );

				C = Cnext;

				NCHARGES--;
			}

			else
			{
				/* Send Radar Blip */

				tmp = BLIP_TRANSFER;
				pvm_pkint( &tmp, 1, 1 );

				pvm_pkdouble( &(C->loc_x), 1, 1 );
				pvm_pkdouble( &(C->loc_y), 1, 1 );

				pvm_pkdouble( &(C->value), 1, 1 );
				pvm_pkdouble( &(C->radius), 1, 1 );
				pvm_pkdouble( &(C->mass), 1, 1 );

				Clast = C;

				C = C->next;
			}
		}

		tmp = END_TRANSFER;
		pvm_pkint( &tmp, 1, 1 );

		pvm_send( R->tid, TRANSFER_MSG );

		R = R->neigh;
	}
}


void
check_exits()
{
	CHARGE Clast;
	CHARGE Cnext;
	CHARGE C;

	int radius;

	if ( WALLS )
		return;

	/* Check for Charge Exits */

	C = CHARGES;

	Clast = (CHARGE) NULL;

	while ( C != NULL )
	{
		if ( (int) C->loc_x < GLOBREGION->xmin
			|| (int) C->loc_x > GLOBREGION->xmax
			|| (int) C->loc_y < GLOBREGION->ymin
			|| (int) C->loc_y > GLOBREGION->ymax )
		{
			/* Remove From List */

			DUMP_CHARGE( "\nexit",
				C->cid, C->loc_x, C->loc_y,
				C->value, C->radius, C->mass,
				C->vel_angle, C->vel_magn );

			if ( !qflag )
			{
				printf( "E" );
				fflush( stdout );
			}

			if ( Clast == NULL )
				CHARGES = C->next;

			else
				Clast->next = C->next;

			Cnext = C->next;

			free_charge( &C );

			C = Cnext;

			NCHARGES--;
		}

		else
		{
			Clast = C;

			C = C->next;
		}
	}
}


recv_transfers()
{
	double value;
	double radius;
	double mass;
	double vel_angle;
	double vel_magn;
	double x, y;

	int msgtag;
	int parent;
	int buf;
	int bid;
	int age;
	int cnt;
	int cid;
	int tmp;

	/* Receive Transfers Back from Each Neighbor */

	if ( MYREGION->nneigh == 0 )
		return( stvTrue );

	cnt = 0;

	do
	{
		buf = stv_userRecv( -1, TRANSFER_MSG, TRANSFER_MSG );

		if ( buf < 0 )
		{
			printf( "\nError Receiving Transfer Message.\n\n" );

			return( stvFalse );
		}

		/* Transfer Message Received */

		if ( buf > 0 )
		{
			cnt++;

			/* Get Transfer Op */

			pvm_upkint( &tmp, 1, 1 );

			while ( tmp != END_TRANSFER )
			{
				switch ( tmp )
				{
					case BLIP_TRANSFER:
					{
						pvm_upkdouble( &x, 1, 1 );
						pvm_upkdouble( &y, 1, 1 );
						pvm_upkdouble( &value, 1, 1 );
						pvm_upkdouble( &radius, 1, 1 );
						pvm_upkdouble( &mass, 1, 1 );

						add_blip( x, y, value, radius, mass );

						break;
					}

					case CHARGE_TRANSFER:
					{
						pvm_upkint( &cid, 1, 1 );
						pvm_upkdouble( &x, 1, 1 );
						pvm_upkdouble( &y, 1, 1 );
						pvm_upkdouble( &value, 1, 1 );
						pvm_upkdouble( &radius, 1, 1 );
						pvm_upkdouble( &mass, 1, 1 );
						pvm_upkdouble( &vel_angle, 1, 1 );
						pvm_upkdouble( &vel_magn, 1, 1 );

						if ( vflag )
							printf( "\ncharge recvd:\n" );

						add_charge( cid, x, y, value, radius, mass,
							vel_angle, vel_magn );

						break;
					}
				}

				/* Get Next Transfer Op */

				pvm_upkint( &tmp, 1, 1 );
			}
		}
	}
	while ( cnt < MYREGION->nneigh );

	return( stvTrue );
}


void
charge_action()
{
	BLIP B;

	CHARGE C;

	double net_angle, net_magn;
	double angle, magn;

	double factor;
	double dist;
	double sign;

	int flag;

	/* Check Each Charge */

	C = CHARGES;

	flag = 0;

	while ( C != NULL )
	{
		/* Start With Field Forces */

		factor = ( C->value < 0.0 ) ? -1.0 : 1.0;

		net_angle = FIELD_ANGLE;
		net_magn = factor * FIELD_MAGN;

		/* Accumulate Net Force On Charge */

		B = BLIPS;

		while ( B != NULL )
		{
			dist = distance( C->loc_x, C->loc_y, B->loc_x, B->loc_y );

			/* Not Me */

			if ( dist > 0.0 )
			{
				/* Get Angle & Magn of Force */

				angle = angle_of( B->loc_x, B->loc_y,
					C->loc_x, C->loc_y );

				sign = C->value / B->value;

				magn = ( ( sign > 0.0 ) ? 1.0 : -1.0 )
					* 100.0 * fabs( B->value ) / dist;

				/* Add to Net Angle & Magn */

				add_to_vector( &net_angle, &net_magn, angle, magn );

				normalize_angle( &net_angle );

				/* No Inactivity */

				flag++;
			}

			B = B->next;
		}

		/* Adjust Charge Velocity */

		if ( flag )
		{
			/* Inertia */

			net_magn /= C->mass;

			/* Apply */

			add_to_vector( &(C->vel_angle), &(C->vel_magn),
				net_angle, net_magn );

			normalize_angle( &(C->vel_angle) );
		}

		C = C->next;
	}

	/* Check for Charge Additions */

	if ( CHARGES == NULL )
		add_rand_charge();

	else if ( CHARGES->next == NULL )
	{
		if ( rand_double() < 0.5 )
			add_rand_charge();
	}

	else
	{
		if ( rand_double() < APPEAR_PROB )
			add_rand_charge();
	}
}


void
handle_steering()
{
	CHARGE C;

	do
	{
		/* Handle Charge Steering */

		C = get_charge_by_cid( CHARGE_INDEX );

		if ( C != NULL )
		{
			C->value = CHARGE_VALUE;
			C->radius = CHARGE_RADIUS;
			C->mass = CHARGE_MASS;
		}

		/* Reset Parameters */

		CHARGE_VALUE = CHARGE_RADIUS = CHARGE_MASS = -1.0;

		CHARGE_INDEX = -1;
	}
	while ( stv_get_remaining_params() > 0 );
}


void
read_args()
{
	int i, j, k;
	int do_usage;
	int len;

	APPEAR_PROB = 0.005;

	DISAPPEAR_PROB = 0.00005;

	FIELD_ANGLE = 0.0;

	FIELD_MAGN = 0.0;

	NPARTS = 1;

	WALLS = stvTrue;

	SPMD = stvFalse;

	NTASKS = 1;

	vflag = 0;
	qflag = 0;
	dflag = 0;
	cflag = 0;

	stv_setopt( stvOptNull, stvOptVerbose, vflag ? STV_DEBUG_ALL : 0 );

	do_usage = 0;

	for ( i=0 ; i < Argc ; i++ )
	{
		k = i + 1;

		if ( !strcmp( Argv[i], "-SPMD" ) )
			SPMD = stvTrue;

		else if ( !strcmp( Argv[i], "-FA" ) )
			FIELD_ANGLE = atof( Argv[k++] );

		else if ( !strcmp( Argv[i], "-FM" ) )
			FIELD_MAGN = atof( Argv[k++] );

		else if ( Argv[i][0] == '-' )
		{
			len = strlen( Argv[i] );

			for ( j=1 ; j < len ; j++ )
			{
				switch ( Argv[i][j] )
				{
					case 'N':
					{
						NTASKS = atoi( Argv[k++] );

						break;
					}

					case 'P':
					{
						NPARTS = atoi( Argv[k++] );

						break;
					}

					case 'A':
					{
						APPEAR_PROB = atof( Argv[k++] );

						break;
					}

					case 'D':
					{
						DISAPPEAR_PROB = atof( Argv[k++] );

						break;
					}

					case 'W':	WALLS = stvTrue;	break;

					case 'O':	WALLS = stvFalse;	break;

					case 'v':
					{
						vflag++;

						stv_setopt( stvOptNull, stvOptVerbose,
							vflag ? STV_DEBUG_ALL : 0 );

						break;
					}

					case 'q': qflag++; break;

					case 'd': dflag++; break;

					case 'c': cflag++; break;

					case 'h': do_usage++; break;

					default:
					{
						printf( "Unknown Option -%c\n", Argv[i][j] );

						do_usage++;

						break;
					}
				}
			}
		}

		i = k - 1;
	}

	if ( do_usage )
		usage();
}


void
usage()
{
	printf( "\nusage:  cp_charge [-vqdc] [-N <ntasks>] [-SPMD]\n" );
	printf( "\t[-P <parts>] [-(A|D) <prob>] [-WO]\n" );
	printf( "\t[-FA <field_angle>] [-FM <field_magn>]\n" );

	printf( "\nwhere:\n" );
	printf( "------\n" );

	printf( "\t-v = Verbose Operation (%s)\n", vflag ? "On" : "Off" );

	printf( "\t-q = Quiet Operation (%s)\n", qflag ? "On" : "Off" );

	printf( "\t-d = Spawn in Debug Mode (%s)\n", dflag ? "On" : "Off" );

	printf( "\t-c = Verbose for Collision Detect (%s)\n",
		cflag ? "On" : "Off" );

	printf( "\t-N <ntasks> = # of Initial Processors (%d)\n", NTASKS );

	printf( "\t-SPMD = do not use master to spawn rest of tasks (%s)\n",
		SPMD ? "On" : "Off" );

	printf( "\t-P <parts> = Set Initial Number of Particles (%d)\n",
		NPARTS );

	printf( "\t-A <prob> = Set Charge Appearance Probability (%lf)\n",
		APPEAR_PROB );

	printf(
		"\t-D <prob> = Set Charge Disappearance Probability (%lf)\n",
		DISAPPEAR_PROB );

	printf( "\t-W = Set Closed (Turn On Walls) (%d)\n", WALLS );

	printf( "\t-O = Set Open (Turn Off Walls) (%d)\n", !WALLS );

	printf( "\t-FA <field_angle> = Set Field Angle (%lf)\n",
		FIELD_ANGLE );

	printf( "\t-FM <field_magn> = Set Field Magnitude (%lf)\n",
		FIELD_MAGN );

	printf( "\n" );

	exit( 0 );
}


void
program_init()
{
	/* Seed Random Generator */

	srandom( getpid() * (int) time( (time_t *) NULL ) );

	/* Initialize Globals */

	GLOBREGION = (REGION) NULL;

	MYREGION = (REGION) NULL;

	CHARGES = (CHARGE) NULL;
	NCHARGES = 0;

	BLIPS = (BLIP) NULL;

	NEXT_CHARGE_ID = 0;

	/* Steering Parameters */

	CHARGE_VALUE = CHARGE_RADIUS = CHARGE_MASS = -1.0;

	CHARGE_INDEX = -1;
}


void
pvm_init()
{
	/* Join PVM */

	MYTID = pvm_mytid();

	printf( "My TID = [ 0x%x / %d ]\n", MYTID, MYTID );
	fflush( stdout );

	/* Join Application Group */

	MYINST = pvm_joingroup( GROUPNAME );

	if ( MYINST < 0 )
	{
		pvm_perror( "Joining Group" );

		exit( 0 );
	}
}


void
app_init()
{
	int flags;
	int left;
	int cnt;
	int num;

	/* Determine if Master - Spawn Tasks if Necessary */

	if ( !SPMD && !CP_RESTART )
	{
		if ( !MYINST && NTASKS > 1 )
		{
			printf( "Spawning %d Other Tasks... ", NTASKS - 1 );
			fflush( stdout );

			if ( dflag )
				flags = PvmTaskDefault | PvmTaskDebug;

			else
				flags = PvmTaskDefault;

			left = NTASKS - 1;

			num = 0;

			cnt = 0;

			do
			{
				left -= num;

				num = pvm_spawn( "cp_charge", Argv + 1, flags,
					(char *) NULL, left, (int *) NULL );

				if ( num < 0 )
				{
					printf( "\n\nPVM System Error - Exiting.\n\n" );
					fflush( stdout );

					exit( 0 );
				}

				else if ( num != left )
				{
					printf( "\n\tOnly Spawned %d Tasks, Retrying... ",
						num );
					fflush( stdout );
				}

				cnt++;
			}
			while ( num != left && cnt < 3 );

			if ( num != left )
			{
				printf( "\n\nGiving Up on Spawn.  Exiting.\n\n" );
				printf( "Please Reset the Virtual Machine.\n\n" );
				fflush( stdout );

				exit( 0 );
			}

			else
			{
				printf( "done.\n" );
				fflush( stdout );
			}
		}
	}

	/* Set Up Neighbors */

	neigh_init();
}


void
region_init()
{
	/* Free Up Existing Regions */

	if ( GLOBREGION != NULL )
		free_region( &GLOBREGION );

	if ( MYREGION != NULL )
		free_region_list( &MYREGION );

	/* Define Global Region */

	GLOBREGION = create_region();

	GLOBREGION->xmin = 0;
	GLOBREGION->xmax = FIELD_WIDTH - 1;

	GLOBREGION->ymin = 0;
	GLOBREGION->ymax = FIELD_HEIGHT - 1;
}


void
neigh_init()
{
	REGION R;

	int neigh;
	int slice;

	/* Slice Size */

	slice = FIELD_WIDTH / NTASKS;

	/* My Slice */

	MYREGION = create_region();

	MYREGION->tid = MYTID;

	MYREGION->xmin = MYINST * slice;
	MYREGION->xmax = MYREGION->xmin + slice - 1;

	MYREGION->ymin = GLOBREGION->ymin;
	MYREGION->ymax = GLOBREGION->ymax;

	MYREGION->nneigh = 0;

	R = MYREGION;

	/* Left Neighbor */

	neigh = MYINST - 1;

	if ( neigh >= 0 )
	{
		R = R->neigh = create_region();

		R->xmin = neigh * slice;
		R->xmax = R->xmin + slice - 1;

		R->ymin = GLOBREGION->ymin;
		R->ymax = GLOBREGION->ymax;

		while ( (R->tid = pvm_gettid( GROUPNAME, neigh )) < 0 );

		(MYREGION->nneigh)++;
	}

	/* Right Neighbor */

	neigh = MYINST + 1;

	if ( neigh < NTASKS )
	{
		R = R->neigh = create_region();

		R->xmin = neigh * slice;
		R->xmax = R->xmin + slice - 1;

		R->ymin = GLOBREGION->ymin;
		R->ymax = GLOBREGION->ymax;

		while ( (R->tid = pvm_gettid( GROUPNAME, neigh )) < 0 );

		(MYREGION->nneigh)++;
	}
}


void
charge_init()
{
	int ext;
	int num;
	int i;

	/* Distribute Charges Evenly Over Tasks */

	num = NPARTS / NTASKS;

	ext = NPARTS - ( num * NTASKS );

	if ( MYINST < ext )
		num++;

	printf( "Creating %d Initial Points.\n", num );

	for ( i=0 ; i < num ; i++ )
		add_rand_charge();
}


void
do_stv_stuff()
{
	STV_VALUE charge_params[7];
	STV_VALUE field_params[2];

	STV_VALUE vptr;

	char *charge_names[3];
	char *field_names[2];

	char *sptr;

	int charge_types[3];
	int field_types[2];

	int glb[2];
	int gub[2];

	int charge_id;

	int type;

	stv_setopt( stvOptNull, stvOptVerbose, vflag ? STV_DEBUG_ALL : 0 );

	stv_init( APPNAME, STV_MSGTAG, NTASKS, MYINST );

	stv_setopt( stvOptNull, stvOptCpRecovery, stvOptCpRollback );

	CP_RESTART = stv_cpInit( "cp_charge", TRANSFER_MSG, &NTASKS );

	region_init();

	glb[0] = GLOBREGION->xmin;
	glb[1] = GLOBREGION->ymin;

	gub[0] = GLOBREGION->xmax;
	gub[1] = GLOBREGION->ymax;

	/* Charge Particle / Fields */

	charge_id = stv_particleDefine( "charge", 2, glb, gub, NTASKS,
		get_charge, make_charge );

	stv_pfieldDefine( "cid", charge_id, stvInt, 1,
		get_charge_pfield, (STV_VALUE) PF_CID,
		set_charge_pfield, (STV_VALUE) PF_CID,
		stvVisOnly | stvCpOnly );

	stv_pfieldDefine( "value", charge_id, stvDouble, 1,
		get_charge_pfield, (STV_VALUE) PF_VALUE,
		set_charge_pfield, (STV_VALUE) PF_VALUE,
		stvVisOnly | stvCpOnly );

	stv_pfieldDefine( "radius", charge_id, stvDouble, 1,
		get_charge_pfield, (STV_VALUE) PF_RADIUS,
		set_charge_pfield, (STV_VALUE) PF_RADIUS,
		stvVisOnly | stvCpOnly );

	stv_pfieldDefine( "mass", charge_id, stvDouble, 1,
		get_charge_pfield, (STV_VALUE) PF_MASS,
		set_charge_pfield, (STV_VALUE) PF_MASS,
		stvVisOnly | stvCpOnly );

	stv_pfieldDefine( "vel_angle", charge_id, stvDouble, 1,
		get_charge_pfield, (STV_VALUE) PF_VEL_ANGLE,
		set_charge_pfield, (STV_VALUE) PF_VEL_ANGLE,
		stvVisOnly | stvCpOnly );

	stv_pfieldDefine( "vel_magn", charge_id, stvDouble, 1,
		get_charge_pfield, (STV_VALUE) PF_VEL_MAGN,
		set_charge_pfield, (STV_VALUE) PF_VEL_MAGN,
		stvVisOnly | stvCpOnly );

	stv_pfieldDefine( "x_coord", charge_id, stvDouble, 1,
		get_charge_pfield, (STV_VALUE) PF_X_COORD,
		set_charge_pfield, (STV_VALUE) PF_X_COORD,
		stvCpOnly );

	stv_pfieldDefine( "y_coord", charge_id, stvDouble, 1,
		get_charge_pfield, (STV_VALUE) PF_Y_COORD,
		set_charge_pfield, (STV_VALUE) PF_Y_COORD,
		stvCpOnly );

	/* Set Field Options */

	stv_setopt( stvOptAll, stvOptFieldTimeScale,
			stv_time_of( 10, stvTimeMilliSec ) );

	stv_setopt( stvOptAll, stvOptSetFieldTime,
			stv_time_of( 60, stvTimeSec ) );

	/* Steering Parameters */

	/* Elec Field */

	field_params[0] = (STV_VALUE) &FIELD_ANGLE;
	field_params[1] = (STV_VALUE) &FIELD_MAGN;

	field_names[0] = "angle";
	field_names[1] = "magnitude";

	field_types[0] = stvDouble;
	field_types[1] = stvDouble;

	stv_vparamDefine( "field",
		field_params, field_names, field_types, 2,
		stvVisOnly | stvCpOnly );

	/* Walls */

	stv_paramDefine( "walls", (STV_VALUE) &WALLS, stvInt,
		stvVisOnly | stvCpOnly );

	/* Charges */

	charge_params[0] = (STV_VALUE) &CHARGE_VALUE;
	charge_params[1] = (STV_VALUE) &CHARGE_RADIUS;
	charge_params[2] = (STV_VALUE) &CHARGE_MASS;

	charge_names[0] = "value";
	charge_names[1] = "radius";
	charge_names[2] = "mass";

	charge_types[0] = stvDouble;
	charge_types[1] = stvDouble;
	charge_types[2] = stvDouble;

	stv_vparamDefine( "charge",
		charge_params, charge_names, charge_types, 3,
		stvVisOnly | stvCpOnly );

	/* Other Parameters */

	sptr = "appear prob";
	vptr = (STV_VALUE) &APPEAR_PROB;
	type = stvDouble;

	stv_paramDefine( sptr, vptr, type, stvVisOnly | stvCpOnly );

	sptr = "disappear prob";
	vptr = (STV_VALUE) &DISAPPEAR_PROB;
	type = stvDouble;

	stv_paramDefine( sptr, vptr, type, stvVisOnly | stvCpOnly );

	/* Checkpointed Only Parameters */

	sptr = "next charge id";
	vptr = (STV_VALUE) &NEXT_CHARGE_ID;
	type = stvInt;

	stv_paramDefine( sptr, vptr, type, stvCpOnly );

	sptr = "vflag";
	vptr = (STV_VALUE) &vflag;
	type = stvInt;

	stv_paramDefine( sptr, vptr, type, stvCpOnly );

	sptr = "qflag";
	vptr = (STV_VALUE) &qflag;
	type = stvInt;

	stv_paramDefine( sptr, vptr, type, stvCpOnly );

	sptr = "dflag";
	vptr = (STV_VALUE) &dflag;
	type = stvInt;

	stv_paramDefine( sptr, vptr, type, stvCpOnly );

	sptr = "cflag";
	vptr = (STV_VALUE) &cflag;
	type = stvInt;

	stv_paramDefine( sptr, vptr, type, stvCpOnly );
}


REGION create_region()
{
	REGION tmp;

	tmp = (REGION) malloc( sizeof( struct region_struct ) );
	memcheck( tmp, "Region Structure" );

	tmp->tid = -1;

	tmp->xmin = -1;
	tmp->xmax = -1;

	tmp->ymin = -1;
	tmp->ymax = -1;

	tmp->nneigh = -1;

	tmp->neigh = (REGION) NULL;

	return( tmp );
}


void
free_region( ptr )
REGION *ptr;
{
	REGION R;

	if ( ptr == NULL || *ptr == NULL )
		return;

	R = *ptr;

	R->tid = -1;

	R->xmin = -1;
	R->xmax = -1;

	R->ymin = -1;
	R->ymax = -1;

	R->nneigh = -1;

	R->neigh = (REGION) NULL;

	free( R );

	*ptr = (REGION) NULL;
}


void
free_region_list( ptr )
REGION *ptr;
{
	REGION Rneigh;
	REGION R;

	if ( ptr == NULL || *ptr == NULL )
		return;

	R = *ptr;

	*ptr = (REGION) NULL;

	while ( R != NULL )
	{
		Rneigh = R->neigh;

		free_region( &R );

		R = Rneigh;
	}
}


CHARGE create_charge()
{
	CHARGE tmp;

	tmp = (CHARGE) malloc( sizeof( struct charge_struct ) );
	memcheck( tmp, "Charge Structure" );

	tmp->cid = -1;

	tmp->loc_x = -1.0;
	tmp->loc_y = -1.0;

	tmp->bounce = -1;

	tmp->value = -1.0;
	tmp->radius = -1.0;
	tmp->mass = -1.0;

	tmp->vel_angle = -1.0;
	tmp->vel_magn = -1.0;

	tmp->next = (CHARGE) NULL;

	return( tmp );
}


void
free_charge( ptr )
CHARGE *ptr;
{
	CHARGE C;

	if ( ptr == NULL || *ptr == NULL )
		return;

	C = *ptr;

	C->cid = -1;

	C->loc_x = -1.0;
	C->loc_y = -1.0;

	C->bounce = -1;

	C->value = -1.0;
	C->radius = -1.0;
	C->mass = -1.0;

	C->vel_angle = -1.0;
	C->vel_magn = -1.0;

	C->next = (CHARGE) NULL;

	free( C );

	*ptr = (CHARGE) NULL;
}


CHARGE add_charge( cid, x, y, value, radius, mass, vel_angle, vel_magn )
int cid;
double x, y;
double value;
double radius;
double mass;
double vel_angle;
double vel_magn;
{
	CHARGE C;

	if ( CHARGES == NULL )
		C = CHARGES = create_charge();

	else
	{
		C = CHARGES;

		while ( C->next != NULL )
			C = C->next;

		C = C->next = create_charge();
	}

	NCHARGES++;

	DUMP_CHARGE( "adding", cid, x, y, value, radius, mass,
		vel_angle, vel_magn );

	C->cid = cid;

	C->loc_x = x;
	C->loc_y = y;

	C->value = value;
	C->radius = radius;
	C->mass = mass;

	C->vel_angle = vel_angle;
	C->vel_magn = vel_magn;

	C->bounce = stvFalse;

	return( C );
}


void
add_rand_charge()
{
	double value;
	double radius;
	double mass;

	double vel_angle;
	double vel_magn;

	int x, y;

	int cid;

	/* Generate One Random Charge */

	if ( vflag )
	{
		printf( "\nadding random charge:\n" );
		fflush( stdout );
	}

	else if ( !qflag )
	{
		printf( "C" );
		fflush( stdout );
	}

	cid = CHARGE_ID( MYINST );

	rand_point( MYREGION, &x, &y );

	value = 2.0 * ( rand_double() - 0.5 );

	radius = 10.0 * rand_double();

	mass = 10.0 * rand_double();

	vel_angle = rand_angle();

	vel_magn = 10.0 * rand_double();

	add_charge( cid, (double) x, (double) y, value, radius, mass,
		vel_angle, vel_magn );
}


CHARGE get_charge_by_cid( cid )
int cid;
{
	CHARGE C;

	C = CHARGES;

	while ( C != NULL )
	{
		if ( C->cid == cid )
			return( C );

		C = C->next;
	}

	return( (CHARGE) NULL );
}


BLIP create_blip()
{
	BLIP tmp;

	tmp = (BLIP) malloc( sizeof( struct blip_struct ) );
	memcheck( tmp, "Blip Structure" );

	tmp->loc_x = -1.0;
	tmp->loc_y = -1.0;

	tmp->value = -1.0;
	tmp->radius = -1.0;
	tmp->mass = -1.0;

	tmp->next = (BLIP) NULL;

	return( tmp );
}


void
free_blip( ptr )
BLIP *ptr;
{
	BLIP B;

	if ( ptr == NULL || *ptr == NULL )
		return;

	B = *ptr;

	B->loc_x = -1.0;
	B->loc_y = -1.0;

	B->value = -1.0;
	B->radius = -1.0;
	B->mass = -1.0;

	B->next = (BLIP) NULL;

	free( B );

	*ptr = (BLIP) NULL;
}


void
free_blip_list( ptr )
BLIP *ptr;
{
	BLIP Bnext;
	BLIP B;

	if ( ptr == NULL || *ptr == NULL )
		return;

	B = *ptr;

	*ptr = (BLIP) NULL;

	while ( B != NULL )
	{
		Bnext = B->next;

		free_blip( &B );

		B = Bnext;
	}
}


BLIP add_blip( x, y, value, radius, mass )
double x, y;
double value;
double radius;
double mass;
{
	BLIP B;

	if ( BLIPS == NULL )
		B = BLIPS = create_blip();

	else
	{
		B = BLIPS;

		while ( B->next != NULL )
			B = B->next;

		B = B->next = create_blip();
	}

	B->loc_x = x;
	B->loc_y = y;

	B->value = value;
	B->radius = radius;
	B->mass = mass;

	return( B );
}


void
cleanup_region()
{
	BLIP Bnext;
	BLIP B;

	CHARGE Cnext;
	CHARGE C;

	/* Free Any Charge */

	C = CHARGES;

	CHARGES = (CHARGE) NULL;

	NCHARGES = 0;

	while ( C != NULL )
	{
		Cnext = C->next;

		free_charge( &C );

		C = Cnext;
	}

	/* Free Any Blips */

	B = BLIPS;

	BLIPS = (BLIP) NULL;

	while ( B != NULL )
	{
		Bnext = B->next;

		free_blip( &B );

		B = Bnext;
	}
}


void
get_charge( hnum, R, id, coords )
int hnum;
STV_REGION R;
STV_PARTICLE_ID *id;
int *coords;
{
	static CHARGE C_save = (CHARGE) NULL;
	static int count_save = -1;

	CHARGE C;

	int hcoords[2];

	int count;

	/* Reset */

	if ( hnum == -1 )
	{
		*id = (STV_PARTICLE_ID) NULL;

		C_save = (CHARGE) NULL;

		count_save = -1;
	}

	/* Restore */

	if ( hnum >= count_save && count_save != -1 )
	{
		C = C_save;

		count = count_save;
	}

	/* Fresh Search */

	else
	{
		C = CHARGES;

		count = 0;
	}

	/* Search for Charge */

	while ( C != NULL )
	{
		hcoords[0] = (int) C->loc_x;
		hcoords[1] = (int) C->loc_y;

		if ( stv_particle_in_region( R, hcoords, 2 ) )
		{
			count++;

			if ( count == hnum )
			{
				*id = (STV_PARTICLE_ID) C;

				coords[0] = hcoords[0];
				coords[1] = hcoords[1];

				C_save = C->next;

				count_save = count;

				return;
			}
		}

		C = C->next;
	}

	/* Charge Not Found */

	*id = (STV_PARTICLE_ID) NULL;

	C_save = (CHARGE) NULL;

	count_save = -1;
}


void
get_charge_pfield( id, field, value )
STV_PARTICLE_ID id;
STV_VALUE field;
STV_VALUE *value;
{
	CHARGE C;

	long fid;

	C = (CHARGE) id;

	fid = (long) field;

	switch ( fid )
	{
		case PF_CID:
			*value = (STV_VALUE) &(C->cid);
			break;

		case PF_VALUE:
			*value = (STV_VALUE) &(C->value);
			break;

		case PF_RADIUS:
			*value = (STV_VALUE) &(C->radius);
			break;

		case PF_MASS:
			*value = (STV_VALUE) &(C->mass);
			break;

		case PF_VEL_ANGLE:
			*value = (STV_VALUE) &(C->vel_angle);
			break;

		case PF_VEL_MAGN:
			*value = (STV_VALUE) &(C->vel_magn);
			break;

		case PF_X_COORD:
			*value = (STV_VALUE) &(C->loc_x);
			break;

		case PF_Y_COORD:
			*value = (STV_VALUE) &(C->loc_y);
			break;

		default:
			*value = (STV_VALUE) NULL;
			break;
	}
}


void
make_charge( coords, id )
int *coords;
STV_PARTICLE_ID *id;
{
	CHARGE C;

	if ( CHARGES == NULL )
		C = CHARGES = create_charge();

	else
	{
		C = CHARGES;

		while ( C->next != NULL )
			C = C->next;

		C = C->next = create_charge();
	}

	NCHARGES++;

	C->loc_x = (double) coords[0];
	C->loc_y = (double) coords[1];

	C->bounce = stvFalse;

	*id = (STV_PARTICLE_ID) C;
}


void
set_charge_pfield( id, field, value )
STV_PARTICLE_ID id;
STV_VALUE field;
STV_VALUE value;
{
	CHARGE C;

	long fid;

	C = (CHARGE) id;

	fid = (long) field;

	switch ( fid )
	{
		case PF_CID:
			C->cid = STV_VALUE_OF( value, int );
			break;

		case PF_VALUE:
			C->value = STV_VALUE_OF( value, double );
			break;

		case PF_RADIUS:
			C->radius = STV_VALUE_OF( value, double );
			break;

		case PF_MASS:
			C->mass = STV_VALUE_OF( value, double );
			break;

		case PF_VEL_ANGLE:
			C->vel_angle = STV_VALUE_OF( value, double );
			break;

		case PF_VEL_MAGN:
			C->vel_magn = STV_VALUE_OF( value, double );
			break;

		case PF_X_COORD:
			C->loc_x = STV_VALUE_OF( value, double );
			break;

		case PF_Y_COORD:
			C->loc_y = STV_VALUE_OF( value, double );
			break;

		default:
			break;
	}
}


void
normalize_angle( angle )
double *angle;
{
	while ( *angle > PI )
		*angle -= 2.0 * PI;

	while ( *angle < -1.0 * PI )
		*angle += 2.0 * PI;
}


void
rand_point( R, x, y )
REGION R;
int *x, *y;
{
	*x = R->xmin + ( (int) random() % ( R->xmax - R->xmin ) );

	*y = R->ymin + ( (int) random() % ( R->ymax - R->ymin ) );
}


double
rand_double()
{
	double r;

	r = (double) ( (int) random() % 1000000 ) / 1000000.0;

	return( r );
}


double
rand_angle()
{
	double angle;

	angle = ( -1.0 * PI ) + ( 2.0 * PI * rand_double() );

	return( angle );
}


double
distance( x1, y1, x2, y2 )
double x1, y1, x2, y2;
{
	return( sqrt(((x2 - x1) * (x2 - x1)) + ((y2 - y1) * (y2 - y1))) );
}


double
angle_of( x1, y1, x2, y2 )
double x1, y1, x2, y2;
{
	double slope;
	double at;

	/* North is heading angle 0.0 */
	/* y's grow downward from the top of the screen */

	if ( x2 == x1 )
	{
		if ( y2 > y1 )
			at = -1.0 * PI;

		else
			at = 0.0;
	}

	else
	{
		slope = (y1 - y2) / (x2 - x1);

		at = atan( slope );

		if ( x1 > x2 )
			at += PI / 2.0;

		else
			at -= PI / 2.0;
	}

	return( at );
}


void
add_to_vector( a, m, a2, m2 )
double *a, *m;
double a2, m2;
{
	double x2, y2;
	double x, y;

	x = (*m) * cos( (*a) );
	y = (*m) * sin( (*a) );

	x2 = m2 * cos( a2 );
	y2 = m2 * sin( a2 );

	x += x2;
	y += y2;

	*m = sqrt( sq( x ) + sq( y ) );

	*a = atan( y / x );

	if ( x < 0 )
	{
		if ( y > 0 )
			*a += PI;

		else
			*a -= PI;
	}
}


double
pwr3( x )
double x;
{
	return( x * x * x );
}


double
sq( x )
double x;
{
	return( x * x );
}


void
memcheck( ptr, name )
char *ptr;
char *name;
{
	if ( ptr == NULL )
	{
		fprintf( stderr, "\nError Allocating Memory for \"%s\"\n\n",
			name );

		exit( -1 );
	}
}

