
C $Id: fcloud_nocp.f,v 3.4 1999/09/30 20:15:07 kohl Exp $


C
C                       CUMULVS version 1.0:
C       Computational Steering and Interactive Visualization
C                    in Distributed Applications
C            Oak Ridge National Laboratory, Oak Ridge TN.
C       Authors:  James Arthur Kohl and Philip M. Papadopoulos
C                    (C) 1996 All Rights Reserved
C
C                               NOTICE
C
C  Permission to use, copy, modify, and distribute this software and
C  its documentation for any purpose and without fee is hereby granted
C  provided that the above copyright notice appear in all copies and
C  that both the copyright notice and this permission notice appear
C  in supporting documentation.
C
C  Neither the Institution, Oak Ridge National Laboratory, nor the
C  Authors make any representations about the suitability of this
C  software for any purpose.  This software is provided ``as is''
C  without express or implied warranty.
C
C  CUMULVS was funded by the U.S. Department of Energy.
C


C fcloud_nocp.f - A small test program to test the communication
C                  setup with avs.

       program fcloud_nocp
       implicit none
       include 'fpvm3.h'
       include 'fstv.h'

C ---- Parameter declarations
       character*6 FCLOUDGRP
       integer rows, columns
       integer I_DIM, J_DIM, K_DIM, NTASKS, DEPTH, STVTAG
       integer NFIELDS, CYCLESZ,CYCLEDEC,NPARAMS,MSGCTXT
       integer RINGMSG,CONFIGMSG
       integer SPMD
       parameter (FCLOUDGRP='fcloud_nocp')
       parameter(ROWS=2, COLUMNS=2)
       parameter(I_DIM=30, J_DIM=30, K_DIM=120)
       parameter(NTASKS=ROWS*COLUMNS, DEPTH=K_DIM)
       parameter(STVTAG=100)
       parameter(MSGCTXT = 101)
       parameter(CONFIGMSG = 1000)
       parameter(RINGMSG = 1001)
       parameter(NFIELDS=2)
       parameter(CYCLESZ=2)
       parameter(NPARAMS=4)
       parameter(CYCLEDEC= CYCLESZ*(1+(J_DIM*COLUMNS)/
     +             (CYCLESZ*COLUMNS)) )
       parameter(SPMD=1)

C ----- Variable declarations
       integer mytid, myrow, mycol, myinum
       integer tids(NTASKS), info
       integer i,j,k, declare(3), sequence
       real*4  density(I_DIM,J_DIM,K_DIM), value
       real*4  pressure(I_DIM,CYCLEDEC,K_DIM)
       integer rowsize, colsize
       integer centeri,centerj,centerk,globali,globalj,globalk
       integer cglobalj
       integer decompId(NFIELDS), fieldID(NFIELDS)
       integer paramId(NPARAMS)
       real*4  radius, rate, vradius, dist, dist1, vrate
       real*4 params(NPARAMS)
       integer glb(3), gub(3), prank, pshape(2), paddr(2)
       integer axisType(3), axisInfo(3)
       integer restart,np_restart
       integer plist(0:NTASKS-1),onum,otid,dest,src
       integer size
       integer gen
       integer tval

C ---- Start of code
       do i=1,I_DIM
          do j = 1, CYCLEDEC
              do k = 1, K_DIM
                   pressure(i,j,k) = 0.0
              enddo
          enddo
       enddo
       call pvmfmytid(mytid)
       print *,'Fcloud started at tid ', mytid
       call pvmfjoingroup(FCLOUDGRP, myinum)
       print *,'Joined Group ', FCLOUDGRP, ' as instance ', myinum

C - Initialize the steering and visualization stuff
C -  NTASKS can be any number if it is unknown. stvinit will gets it
C    copy of NTASKS set with the call to stvfcpinit, if the task is
C    restarting.

       print *, 'Initializing CUMULVS...'
C      call stvfsetopt( STVOPTNULL, STVOPTVERBOSE, STVDEBUGALL, info )
       call stvfinit( FCLOUDGRP, STVTAG, NTASKS, myinum, info)
       print *, '    info = ', info

       if (myinum .eq.  0) then
           if (SPMD .ne. 1) then
               print *, 'Master Fcloud Task.'
               if (NTASKS - 1 .gt. 0) then
                   print *, 'Spawn Remaining Tasks...'
C                  call pvmfcatchout(1,info)
                   call pvmfspawn('fcloud_nocp',
     +                 0,
C    +                 PVMTASKHOST+PVMTASKDEBUG,
C    +                 PVMTASKHOST+PVMHOSTCOMPL,
C    +                 PVMTASKHOST+PVMHOSTCOMPL+PVMTASKDEBUG,
     +                 ' ', NTASKS - 1, tids, info)
                   if (info .lt. NTASKS - 1)  then
                       write(6,*) 'spawn cacked with info = ', info
                   else
                       print *, 'Spawned ', NTASKS - 1, ' tasks.'
                       do i = 1, NTASKS - 1
                           print *, '    tid=', tids(i)
                       enddo
                   endif
               else
                   print *,'Master Task:  All ',NTASKS,' Tasks Spawned.'
               endif
           else
               print *, 'SPMD Fcloud Task Started (one of ',NTASKS,').'
           endif
       endif
       myrow = myinum/COLUMNS
       mycol = mod(myinum,COLUMNS)


C -- Record the size of the computational region
       glb(1) = 1
       glb(2) = 1
       glb(3) = 1
       gub(1) = ROWS*I_DIM
       gub(2) = COLUMNS*J_DIM
       gub(3) = K_DIM

C -- Initialize the parameters
       radius = 0
       rate = .2
       vrate = 1.0
       sequence = 0
       params(1) = rate
       params(2) = vrate

       centeri = (I_DIM*ROWS+1)/2
       centerj = (J_DIM*COLUMNS+1)/2
       centerk = (K_DIM+1)/2
       write(6,*) 'center:', centeri, centerj, centerk

C -- Define the steering parameters
       call stvfparamdefine( rate, 'Radius Rate', STVFLOAT,STVVISCP,
     +                          paramId(1))
       call stvfparamdefine( vrate,'Volume Rate',STVFLOAT,STVVISCP,
     +                          paramId(2))

C - Declare decomposition of density -- it is block, block, collapsed
       axisType(1) = STVBLOCK
       axisType(2) = STVBLOCK
       axisType(3) = STVCOLLAPSE
       prank = 2
       pshape(1) = ROWS
       pshape(2) = COLUMNS
       call stvfdecompdefine(3,axisType,STVDEFAULTINFO,STVDEFAULTINFO,
     +              glb,gub,prank,pshape,decompId(1))

C - declare the density data  field with the above decomposition
       paddr(1) = myrow
       paddr(2) = mycol
       declare(1) = I_DIM
       declare(2) = J_DIM
       declare(3) = K_DIM
       call stvffielddefine(density,'density',decompId(1),
     +              STVNOFIELDOFFSETS, declare,
     +              STVFLOAT, paddr, STVVISONLY, fieldId(1))
       call stvftimeof( 100, STVTIMEMICROSEC, tval )
       call stvfsetopt( fieldId(1), STVOPTFIELDTIMESCALE, tval, info )

C - Declare decomposition of pressure -- it is block, cyclic, collapsed
       axisType(1) = STVBLOCK
       axisType(2) = STVCYCLIC
       axisType(3) = STVCOLLAPSE
       axisInfo(1) = STVDEFAULT
       axisInfo(2) = CYCLESZ
       axisInfo(3) = STVDEFAULT
       prank = 2
       pshape(1) = ROWS
       pshape(2) = COLUMNS
       call stvfdecompdefine(3,axisType,axisInfo,STVDEFAULTINFO,
     +              glb,gub,prank,pshape,decompId(2))

C - declare the pressure field with the above decomposition
       paddr(1) = myrow
       paddr(2) = mycol
       declare(1) = I_DIM
       declare(2) = CYCLEDEC
       declare(3) = K_DIM
       call stvffielddefine(pressure,'pressure',decompId(2),
     +              STVNOFIELDOFFSETS, declare,
     +              STVFLOAT, paddr, STVVISONLY, fieldId(2))
       call stvftimeof( 50, STVTIMEMICROSEC, tval )
       call stvfsetopt( fieldId(2), STVOPTFIELDTIMESCALE, tval, info )

C ---
       colsize = I_DIM * ROWS
       rowsize = J_DIM * COLUMNS

5       size = 0
        call globlistsetup(FCLOUDGRP,plist,NTASKS,myinum,
     >                mytid,MSGCTXT,CONFIGMSG,info)
10     Continue

       radius = radius + rate
       sequence = sequence + 1
       vradius = sequence*vrate
        print *, 'Sequence ', sequence

       do k = 1, K_DIM
           globalk = k
           do j=1,CYCLEDEC
C Determine from the center of the compgrid, which points should
C Should be colored
                globalj = J + mycol*J_DIM
                cglobalj = (J-1)/CYCLESZ
                cglobalj = (cglobalj*(CYCLESZ*COLUMNS))+MOD(J-1,CYCLESZ)
                cglobalj = cglobalj + mycol*CYCLESZ +1
                do I=1,I_DIM
                   globali = I + myrow*I_DIM
                   dist = (globali - centeri)**2+(globalk - centerk)**2
                   dist1 = sqrt(dist + (cglobalj -centerj)**2 )
                   dist = sqrt(dist + (globalj -centerj)**2 )
                   if (dist .lt. radius) then
                       value = (1.0 - dist/radius)*rate*vradius
                   else
                       value = 0.0
                   endif
               if (j .le. J_DIM) then
                       density(i,j,k) = value
               endif
                   if (dist1 .lt. radius) then
                       value = (1.0 - dist1/radius)*rate*vradius
                   else
                       value = 0.0
                   endif
                   if (cglobalj  .le. (J_DIM*COLUMNS)) then
                       pressure(i,j,k) = value
                   endif
                enddo
           enddo
       enddo

C --    send a message to mod(myinum+1,ntasks)
        call pvmfinitsend(PvmDataDefault,info)
        call pvmfpack(INTEGER4,myinum,1,1,info)
        call pvmfpack(INTEGER4,mytid,1,1,info)
        dest = myinum + 1
        if (dest .eq. NTASKS) dest = 0
        dest = plist(dest)
        src = myinum - 1
        if (src .lt. 0) src = NTASKS - 1
        src = plist(src)
        call pvmfsend(dest,RINGMSG,info)

        call pvmfrecv(src,RINGMSG,info)
        call pvmfunpack(INTEGER4,onum,1,1,info)
        call pvmfunpack(INTEGER4,otid,1,1,info)

C -- send messages to front end viewer
       call stvfsendtofe(info)
       call stvfsleep(1)

C      /* End of the computation */
       if ( sequence .lt. 10000 ) then
           goto 10
       endif
       print *, 'fcloud_nocp exiting...'
       end

C - Gets the globlist
       subroutine globlistsetup(grp,list,ntsks,myinum,
     >                mytid,ctxt,tag,info)
       character*6 grp
       integer list(0:ntsks-1), ntsks
       integer myinum,mytid,ctxt,tag
       include 'fpvm3.h'
       include 'fstv.h'

       integer size,info,onum,otid

       size = 0
       print *, 'setup:'
       print *, ' grp=',grp
       print *, ' size=',size
       print *, ' ntask=',ntsks
       print *, ' myinum=',myinum
       print *, ' mytid=',mytid
       print *, ' ctxt=',ctxt
       print *, ' tag=',tag
       do while (size .lt. ntsks)
           call pvmfgsize(grp, size)
           if (size .lt. ntsks) call stvfsleep(1)
            print *, 'group size', size
       enddo
        print *, 'group formed'

C --    Bcast a message to everybody in the group
        call pvmfinitsend(PvmDataDefault,info)
        call pvmfpack(INTEGER4,myinum,1,1,info)
        call pvmfpack(INTEGER4,mytid,1,1,info)
        call pvmfbcast(grp,ctxt,info)

        list(myinum) = mytid
        do i = 1, ntsks-1
            call pvmfrecv(-1,tag,info)
            call pvmfunpack(INTEGER4,onum,1,1,info)
            call pvmfunpack(INTEGER4,otid,1,1,info)
            list(onum) = otid
        enddo
        info = 0
        return
        end

